@@ -35,7 +35,7 @@ Prototype void FreeControl(Connection *conn);
 Prototype void ExecuteSupersedes(Connection *conn, const char *supers, const char *art, int artLen);
 
 Control *ParseControl(Connection *conn, const char *cmd, const char *from, const char *newsgroup);
-void SendMail(FILE *logf, long lpos, const char *cmd, const char *arg);
+void SendMail(FILE *logf, long lpos, char *hostname, const char *cmd, const char *arg);
 const char *PerformControl(Connection *conn, Control *ctl, const char **cmd, const char *art, const char *body, int bodyLen, char *groups);
 const char *CtlCancel(Connection *conn, Control *ctl, const char **cmd, const char *art, const char *body, int bodyLen, char *groups);
 const char *CtlNewGroup(Connection *conn, Control *ctl, const char **cmd, const char *body, int bodyLen);
@@ -160,7 +160,7 @@ ExecuteControl(Connection *conn, const char *ctlMsg, const char *art, int artLen
 	if (cmd[1] == NULL) {
 	    badCmd = "no argument to cancel\n";
 	} else {
-	    cmd[1] = MsgId(cmd[1]);
+	    cmd[1] = MsgId(cmd[1], NULL);
 	    if (strcmp(cmd[1], "<>") == 0)
 		badCmd = "bad message-id in cancel\n";
 	}
@@ -251,7 +251,7 @@ ExecuteControl(Connection *conn, const char *ctlMsg, const char *art, int artLen
 	if ((ctl->ct_Flags & CTF_EXECUTE)) {
 	    if (ctl->ct_LogFo) {
 		if (badCmd) {
-		    fprintf(ctl->ct_LogFo, ">> unable to execute control\n");
+		    fprintf(ctl->ct_LogFo, ">> unable to execute control (%s)\n", badCmd);
 		} else {
 		    fprintf(ctl->ct_LogFo, ">> executing control:\n");
 		}
@@ -290,7 +290,8 @@ ExecuteControl(Connection *conn, const char *ctlMsg, const char *art, int artLen
 	if (ctl->ct_Flags & CTF_MAIL) {
 	    fprintf(ctl->ct_LogFo, ">> mail requested\n");
 	    fflush(ctl->ct_LogFo);
-	    SendMail(ctl->ct_LogFo, ctl->ct_LogSeekPos, cmd[0], cmd[1]);
+	    SendMail(ctl->ct_LogFo, ctl->ct_LogSeekPos,
+		conn->co_Auth.dr_VServerDef->vs_HostName, cmd[0], cmd[1]);
 	} else {
 	    fprintf(ctl->ct_LogFo, ">> mail not requested\n");
 	}
@@ -526,18 +527,18 @@ FreeControl(Connection *conn)
  */
 
 void
-SendMail(FILE *logf, long lpos, const char *cmd, const char *arg)
+SendMail(FILE *logf, long lpos, char *hostname, const char *cmd, const char *arg)
 {
     pid_t pid;
     char *argv[] = { SENDMAIL_PATH, SENDMAIL_ARG0, "-t", "-odq", NULL };
     int fds[3];
 
-    if ((pid = RunProgramPipe(fds, RPF_STDOUT, argv)) > 0) {
+    if ((pid = RunProgramPipe(fds, RPF_STDOUT, argv, NULL)) > 0) {
 	FILE *fo = fdopen(fds[1], "w");
 	char buf[128];
 
-	fprintf(fo, "To: %s\n", NEWSMASTER);
-	fprintf(fo, "From: news@%s\n", ReportedHostName);
+	fprintf(fo, "To: %s\n", safestr(DOpts.NewsMaster, "news"));
+	fprintf(fo, "From: news@%s\n", hostname);
 	fprintf(fo, "Subject: Control message %s %s\n", cmd, (arg ? arg : ""));
 	fprintf(fo, "\n");
 	fseek(logf, lpos, 0);
@@ -560,7 +561,9 @@ PerformControl(Connection *conn, Control *ctl, const char **cmd, const char *art
     const char *badCmd = NULL;
 
     if (strcasecmp(cmd[0], "cancel") == 0 && cmd[1]) {
+#ifndef	DIRTYSPOOL
 	badCmd = CtlCancel(conn, ctl, cmd, art, body, bodyLen, groups);
+#endif
     } else if (strcasecmp(cmd[0], "newgroup") == 0 && cmd[1]) {
 	badCmd = CtlNewGroup(conn, ctl, cmd, body, bodyLen);
     } else if (strcasecmp(cmd[0], "rmgroup") == 0 && cmd[1]) {
@@ -582,6 +585,12 @@ PerformControl(Connection *conn, Control *ctl, const char **cmd, const char *art
  *
  * What we do is take the newsgroup list and scan the overview data for the
  * message-id.
+ *
+ * 1) Find OverInfo (index) for group (no locking)
+ * 2) Lock index
+ * 3) Lookup article info (OverArt)
+ * 4) Write cancelled article
+ * 5) Unlock index
  */
 
 const char *
@@ -629,7 +638,7 @@ CtlCancel(Connection *conn, Control *ctl, const char **cmd, const char *art, con
 		 * Get article contents, set isValidCancel to 1 or -1
 		 */
 		int hlen;
-		const char *data = GetOverRecord(ov, artNo, &hlen, NULL);
+		const char *data = GetOverRecord(ov, artNo, &hlen, NULL, NULL);
 		if (data != NULL) {
 		    int cxLen;
 		    const char *cxtrace = locateXTrace(data, hlen, &cxLen);
@@ -729,8 +738,8 @@ CtlNewGroup(Connection *conn, Control *ctl, const char **cmd, const char *body,
      */
 
     if ((rec = KPDBReadRecord(KDBActive, cmd[1], KP_LOCK, &recLen)) == NULL) {
-	KPDBWrite(KDBActive, cmd[1], "NB", "1", KP_LOCK);
-	KPDBWrite(KDBActive, cmd[1], "NE", "0", KP_LOCK_CONTINUE);
+	KPDBWrite(KDBActive, cmd[1], "NB", "0000000001", KP_LOCK);
+	KPDBWrite(KDBActive, cmd[1], "NE", "0000000000", KP_LOCK_CONTINUE);
 	isNew = 1;
     } else {
 	if (KPDBGetField(rec, recLen, "CTS", NULL, NULL) != NULL)
@@ -773,7 +782,7 @@ CtlNewGroup(Connection *conn, Control *ctl, const char **cmd, const char *body,
 	zfreeStr(&conn->co_MemPool, &moderator);
     }
     if (description) {
-	SanitizeString(description);
+	SanitizeDescString(description);
 	KPDBWriteEncode(KDBActive, cmd[1], "GD", description, KP_LOCK_CONTINUE);
 	zfreeStr(&conn->co_MemPool, &description);
     }
@@ -877,7 +886,7 @@ CtlCheckGroups(Connection *conn, Control *ctl, const char **cmd, const char *bod
 	/*
 	 * clean up the description
 	 */
-	SanitizeString(description);
+	SanitizeDescString(description);
 	{
 	    int l = strlen(description);
 	    while (*description == ' ' || *description == '\t') {
