@@ -5,7 +5,22 @@
  *	Group/Overview routines.  Group control and overview information is
  *	arranged in a two-level directory scheme by hash code under
  *	/news/spool/group.  The first level consists of 256 subdirectories.
- *	The group is stored as a KPDB database file in the second level.
+ *	The group is stored as a KPDB database file (dactive.kp)
+ *
+ *	The header information is stored in 2 files:
+ *		over.0.HASH		Overview index
+ *		data.nnnn.HASH		Overview data
+ *
+ *	The overview data file is a list of NUL terminated article
+ *	headers in no particular order.
+ *
+ *	The overview index file is binary file containing a header
+ *	(struct OverInfo) and fixed length records (struct OverArt)
+ *	with each article number placed in a slot based on the
+ *	maximum number of possible articles in the group (ov_MaxArts).
+ *
+ *	Each header is temporarily stored in a memory reference list
+ *	(struct OverData) linked from OverInfo->ov_HData.
  *
  * (c)Copyright 1998, Matthew Dillon, All Rights Reserved.  Refer to
  *    the COPYRIGHT file in the base directory of this distribution
@@ -24,7 +39,7 @@ Prototype void NNFeedOverview(Connection *conn);
 Prototype void FlushOverCache(void);
 Prototype OverInfo *GetOverInfo(const char *group);
 Prototype void PutOverInfo(OverInfo *ov);
-Prototype const char *GetOverRecord(OverInfo *ov, int artno, int *plen, int *alen);
+Prototype const char *GetOverRecord(OverInfo *ov, int artno, int *plen, int *aleno, TimeRestrict *tr);
 Prototype OverInfo *FindCanceledMsg(const char *group, const char *msgid, int *partNo, int *pvalidGroups);
 Prototype int CancelOverArt(OverInfo *ov, int artNo);
 
@@ -32,8 +47,8 @@ Prototype int NNTestOverview(Connection *conn);
 Prototype const char *NNRetrieveHead(Connection *conn, int *povlen, const char **msgid);
 
 const OverArt *GetOverArt(OverInfo *ov, int artNo, off_t *ppos);
-void AssignArticleNo(Connection *conn, ArtNumAss **pan, const char *group, const char *xref, int approved, const char *art, int artLen);
-void WriteOverview(Connection *conn, ArtNumAss *an, const char *group, const char *xref, const char *art, int artLen, const char *msgid);
+void AssignArticleNo(Connection *conn, ArtNumAss **pan, const char *group, const char *xref, int approved, const char *art, int artLen, const char *msgid);
+int WriteOverview(Connection *conn, ArtNumAss *an, const char *group, const char *xref, const char *art, int artLen, const char *msgid);
 OverData *MakeOverHFile(OverInfo *ov, int artNo, int create);
 void FreeOverInfo(OverInfo *ov);
 void FreeOverData(OverData *od);
@@ -49,6 +64,8 @@ NNFeedOverview(Connection *conn)
     char *art = MBNormalize(&conn->co_ArtBuf, &artLen);
     char *line;
     int appr = 0;
+    char ch;
+    int err = 0;
 
     char *newsgroups = NULL;
     char *msgid = NULL;
@@ -81,24 +98,26 @@ NNFeedOverview(Connection *conn)
 	}
 	l -= line - art;
 
+	ch = tolower(*line);
+
 	if (l == 0 || (l == 1 && line[0] == '\r')) {
 	    /* out of headers */
 	    break;
-	} else if (strncasecmp(line, "Newsgroups:", 11) == 0) {
+	} else if (ch == 'n' && strncasecmp(line, "Newsgroups:", 11) == 0) {
 	    newsgroups = zallocStrTrim(&conn->co_MemPool, line + 11, l - 11);
-	} else if (strncasecmp(line, "Message-ID:", 11) == 0) {
+	} else if (ch == 'm' && strncasecmp(line, "Message-ID:", 11) == 0) {
 	    msgid = zallocStrTrim(&conn->co_MemPool, line + 11, l - 11);
-	} else if (strncasecmp(line, "Subject:", 8) == 0) {
+	} else if (ch == 's' && strncasecmp(line, "Subject:", 8) == 0) {
 	    subject = zallocStrTrim(&conn->co_MemPool, line + 8, l - 8);
-	} else if (strncasecmp(line, "Date:", 5) == 0) {
+	} else if (ch == 'd' && strncasecmp(line, "Date:", 5) == 0) {
 	    date = zallocStrTrim(&conn->co_MemPool, line + 5, l - 5);
-	} else if (strncasecmp(line, "Xref:", 5) == 0) {
+	} else if (ch == 'x' && strncasecmp(line, "Xref:", 5) == 0) {
 	    xref = zallocStrTrim(&conn->co_MemPool, line + 5, l - 5);
-	} else if (strncasecmp(line, "Control:", 8) == 0) {
+	} else if (ch == 'c' && strncasecmp(line, "Control:", 8) == 0) {
 	    control = zallocStrTrim(&conn->co_MemPool, line + 8, l - 8);
-	} else if (strncasecmp(line, "Supersedes:", 11) == 0) {
+	} else if (ch == 's' && strncasecmp(line, "Supersedes:", 11) == 0) {
 	    supers = zallocStrTrim(&conn->co_MemPool, line + 11, l - 11);
-	} else if (strncasecmp(line, "Approved:", 9) == 0) {
+	} else if (ch == 'a' && strncasecmp(line, "Approved:", 9) == 0) {
 	    appr = 1;
 	}
     }
@@ -109,22 +128,30 @@ NNFeedOverview(Connection *conn)
 	/*
 	 * failure
 	 */
+	conn->co_Auth.dr_PostFailCount++;
 	if (conn->co_Flags & COF_IHAVE) {
-	    MBPrintf(&conn->co_TMBuf, "437 Rejected, headers missing\r\n");
+	    MBLogPrintf(conn, &conn->co_TMBuf, "437 Rejected, headers missing\r\n");
 	} else {
-	    MBPrintf(&conn->co_TMBuf, "439 %s\r\n",  conn->co_IHaveMsgId);
+	    MBLogPrintf(conn, &conn->co_TMBuf, "439 %s\r\n",  conn->co_IHaveMsgId);
 	}
-    } else if (conn->co_ByteCounter == 0 && conn->co_BytesHeader == 0) {
+    } else if (conn->co_ByteCounter == 0.0 && conn->co_BytesHeader == 0) {
+	conn->co_Auth.dr_PostFailCount++;
 	if (conn->co_Flags & COF_IHAVE) {
-	    MBPrintf(&conn->co_TMBuf, "437 Rejected, Bytes header missing for header-only feed\r\n");
+	    MBLogPrintf(conn, &conn->co_TMBuf, "437 Rejected, Bytes header missing for header-only feed\r\n");
 	} else {
-	    MBPrintf(&conn->co_TMBuf, "439 %s headerOnlyFeed requires Bytes header\r\n",  conn->co_IHaveMsgId);
+	    MBLogPrintf(conn, &conn->co_TMBuf, "439 %s headerOnlyFeed requires Bytes header\r\n",  conn->co_IHaveMsgId);
 	}
     } else if (FindCancelCache(msgid) == 0) {
+	char logbuf[1024];
+	conn->co_Auth.dr_PostFailCount++;
+	snprintf(logbuf, sizeof(logbuf), "%s cancel cache", msgid);
+	LogCmd(conn, '-', logbuf);
+	if (DRIncomingLogPat != NULL)
+	    LogIncoming("%s - %s%s\n", conn->co_Auth.dr_Host, "", logbuf);
 	if (conn->co_Flags & COF_IHAVE) {
-	    MBPrintf(&conn->co_TMBuf, "437 Article Already Canceled\r\n");
+	    MBLogPrintf(conn, &conn->co_TMBuf, "437 Article Already Cancelled\r\n");
 	} else {
-	    MBPrintf(&conn->co_TMBuf, "439 %s Article Already Canceled\r\n",  conn->co_IHaveMsgId);
+	    MBLogPrintf(conn, &conn->co_TMBuf, "439 %s Article Already Cancelled\r\n",  conn->co_IHaveMsgId);
 	}
     } else {
 	/*
@@ -132,73 +159,137 @@ NNFeedOverview(Connection *conn)
 	 */
 	char *group;
 	char *ngroup = NULL;
+	ArtNumAss	*ANABase = NULL;
 
-	if ((conn->co_Flags & COF_WASCONTROL) == 0) {
-	    ArtNumAss	*ANABase = NULL;
 
+	/*
+	 * if it is a control message, we don't really care what the newsgroups
+	 * line says.  instead, we cobble up "control.<type>" or just "control"
+	 */
+
+	if (conn->co_Flags & COF_WASCONTROL) {
+	    char cmsgtype[64];
+	    char *cptr;
+
+	    cptr = control;
+	    while (*cptr == ' ' || *cptr == '\t') {
+		cptr++;
+	    }
+	    snprintf(cmsgtype, sizeof(cmsgtype), "control%s%s", *cptr ? "." : "", cptr);
+	    if (((cptr = strchr(cmsgtype, ' '))) || ((cptr = strchr(cmsgtype, '\t')))) {
+		*cptr = '\0';
+	    }
+	    zfreeStr(&conn->co_MemPool, &newsgroups);
+	    newsgroups = zallocStr(&conn->co_MemPool, cmsgtype);
+	}
+	if (DebugOpt)
+	    printf("Feed overview %s %s\n", msgid, newsgroups);
+
+	/*
+	 * pass 1 - assign article numbers
+	 */
+
+	for (group = newsgroups; *group; group = ngroup) {
+	    char c;
+	    char whspc = 0;
+	    char *whspptr;
+
+	    /* Strip leading spaces */
+	    while (*group == ' ' || *group == '\t') {
+		group++;
+	    }
+	    if (! *group) {
+		break;
+	    }
+
+	    for (ngroup = group; *ngroup && *ngroup != ','; ++ngroup)
+		;
+	    c = *ngroup;
+	    *ngroup = 0;
+
+	    /* Strip trailing space or tab from group name */
+	    whspptr = strpbrk(group, " \t");
+	    if (whspptr) {
+		whspc = *whspptr;
+		*whspptr = 0;
+	    }
+
+	    AssignArticleNo(conn, &ANABase, group, xref, appr, art, artLen, msgid);
+
+	    /* Repair string back to its former state */
+	    if (whspptr) {
+		*whspptr = whspc;
+	    }
+
+	    *ngroup = c;
+
+	    if (*ngroup == ',')
+		++ngroup;
+	}
+
+	/*
+	 * Supersedes is allowed on non-control messages.  We execute the
+	 * cancel AND post the article.  Note: we do not allow supersedes
+	 * on Control: messages. (XXX is this still true with the new logic? JG)
+	 */
+
+	if (supers) {
 	    if (DebugOpt)
-		printf("Feed overview %s %s\n", msgid, newsgroups);
+		printf("has Supersedes: %s %s\n", msgid, newsgroups);
+	    ExecuteSupersedes(conn, supers, art, artLen);
+	}
 
-	    /*
-	     * pass 1 - assign article numbers
-	     */
+	err = 0;
+	for (group = newsgroups; *group; group = ngroup) {
+	    char c;
+	    char whspc = 0;
+	    char *whspptr;
 
-	    for (group = newsgroups; *group; group = ngroup) {
-		char c;
+	    for (ngroup = group; *ngroup && *ngroup != ','; ++ngroup)
+		;
+	    c = *ngroup;
+	    *ngroup = 0;
 
-		for (ngroup = group; *ngroup && *ngroup != ','; ++ngroup)
-		   ;
-		c = *ngroup;
-		*ngroup = 0;
-
-		AssignArticleNo(conn, &ANABase, group, xref, appr, art, artLen);
-
-		*ngroup = c;
-		if (*ngroup == ',')
-		    ++ngroup;
+	    /* Strip trailing space or tab from group name */
+	    whspptr = strpbrk(group, " \t");
+	    if (whspptr) {
+		whspc = *whspptr;
+		*whspptr = 0;
 	    }
 
-	    /*
-	     * Supersedes is allowed on non-control messages.  We execute the
-	     * cancel AND post the article.  Note: we do not allow supersedes
-	     * on Control: messages.
-	     */
+	    err += WriteOverview(conn, ANABase, group, xref, art, artLen, msgid);
 
-	    if (supers) {
-		if (DebugOpt)
-		    printf("has Supersedes: %s %s\n", msgid, newsgroups);
-		ExecuteSupersedes(conn, supers, art, artLen);
+	    /* Repair string back to its former state */
+	    if (whspptr) {
+		*whspptr = whspc;
 	    }
 
-	    for (group = newsgroups; *group; group = ngroup) {
-		char c;
-
-		for (ngroup = group; *ngroup && *ngroup != ','; ++ngroup)
-		   ;
-		c = *ngroup;
-		*ngroup = 0;
-
-		WriteOverview(conn, ANABase, group, xref, art, artLen, msgid);
-
-		*ngroup = c;
-		if (*ngroup == ',')
-		    ++ngroup;
-	    }
-	    while (ANABase) {
-		ArtNumAss *an = ANABase;
-		ANABase = an->an_Next;
-		zfree(&conn->co_MemPool, an, sizeof(ArtNumAss));
-	    }
-	} else {
+	    *ngroup = c;
+	    if (*ngroup == ',')
+		++ngroup;
+	}
+	while (ANABase) {
+	    ArtNumAss *an = ANABase;
+	    ANABase = an->an_Next;
+	    zfree(&conn->co_MemPool, an, sizeof(ArtNumAss));
+	}
+	if (conn->co_Flags & COF_WASCONTROL) {
 	    if (DebugOpt)
 		printf("Control message: %s %s\n", msgid, newsgroups);
+	    LogCmd(conn, 'c', control);
+	    if (DRIncomingLogPat != NULL)
+		LogIncoming("%s c %s %s\n", conn->co_Auth.dr_Host,
+							msgid, control);
 	    ExecuteControl(conn, control, art, artLen);
 	}
 
-	if (conn->co_Flags & COF_IHAVE) {
-	    MBPrintf(&conn->co_TMBuf, "235\r\n");
-	} else {
-	    MBPrintf(&conn->co_TMBuf, "239 %s\r\n",  conn->co_IHaveMsgId);
+	if (!err) {
+	    conn->co_Auth.dr_PostCount++;
+	    if (conn->co_Flags & COF_IHAVE) {
+		MBLogPrintf(conn, &conn->co_TMBuf, "235\r\n");
+	    } else {
+		MBLogPrintf(conn, &conn->co_TMBuf, "239 %s\r\n",  conn->co_IHaveMsgId);
+	    }
 	}
     }
 
@@ -215,7 +306,7 @@ NNFeedOverview(Connection *conn)
 }
 
 void
-AssignArticleNo(Connection *conn, ArtNumAss **pan, const char *group, const char *xref, int approved, const char *art, int artLen)
+AssignArticleNo(Connection *conn, ArtNumAss **pan, const char *group, const char *xref, int approved, const char *art, int artLen, const char *msgid)
 {
     const char *rec;
     int recLen;
@@ -228,13 +319,48 @@ AssignArticleNo(Connection *conn, ArtNumAss **pan, const char *group, const char
     char aabegbuf[16];
     char aaendbuf[16];
     int artNo;
+    char logbuf[1024];
 
-    /*
-     * locate group in active file and lock
-     */
+    if (DOpts.ReaderAutoAddToActive) {
+	/*
+	 * locate group in active file and lock, create if it does not exist
+	 * You will have to manually add a GD, M, and fix S as appropriate
+	 * through some external process, if you use this.
+	 */
+	if ((rec = KPDBReadRecord(KDBActive, group, KP_LOCK, &recLen)) == NULL) {
+	    if (ValidGroupName(group) < 0) {
+		/* syslog(LOG_ERR, "group %s illegal", group); */
+	    } else {
+		char tsBuf[64];
 
-    if ((rec = KPDBReadRecord(KDBActive, group, KP_LOCK, &recLen)) == NULL)
-	return;
+		KPDBWrite(KDBActive, group, "NB", "0000000001", KP_LOCK);
+		KPDBWrite(KDBActive, group, "NE", "0000000000", KP_LOCK_CONTINUE);
+		sprintf(tsBuf, "%08x", (int)time(NULL));
+		KPDBWrite(KDBActive, group, "CTS", tsBuf, KP_LOCK_CONTINUE);
+		KPDBWrite(KDBActive, group, "LMTS", tsBuf, KP_LOCK_CONTINUE);   
+		KPDBWrite(KDBActive, group, "S", "y", KP_UNLOCK);
+	}
+            if ((rec = KPDBReadRecord(KDBActive, group, KP_LOCK, &recLen)) == NULL) {
+		snprintf(logbuf, sizeof(logbuf), "%s %s not in dactive", msgid, group);
+		LogCmd(conn, '-', logbuf);
+		if (DRIncomingLogPat != NULL)
+		    LogIncoming("%s - %s%s\n", conn->co_Auth.dr_Host,
+							"", logbuf);
+		return;
+	    }
+	}
+    } else {
+	/*
+	 * locate group in active file and lock
+	 */
+	if ((rec = KPDBReadRecord(KDBActive, group, KP_LOCK, &recLen)) == NULL) {
+	    snprintf(logbuf, sizeof(logbuf), "%s %s not in dactive", msgid, group);
+	    LogCmd(conn, '-', logbuf);
+	    if (DRIncomingLogPat != NULL)
+		LogIncoming("%s - %s%s\n", conn->co_Auth.dr_Host, "", logbuf);
+	    return;
+	}
+    }
 
     /*
      * silently drop postings to moderated groups that do not have an
@@ -248,6 +374,11 @@ AssignArticleNo(Connection *conn, ArtNumAss **pan, const char *group, const char
 	while (flagsLen > 0) {
 	    if (*flags == 'm') {
 		KPDBUnlock(KDBActive, rec);
+		snprintf(logbuf, sizeof(logbuf), "%s %s unapproved", msgid, group);
+		LogCmd(conn, '-', logbuf);
+		if (DRIncomingLogPat != NULL)
+		    LogIncoming("%s - %s%s\n", conn->co_Auth.dr_Host,
+							"", logbuf);
 		return;
 	    }
 	    --flagsLen;
@@ -282,8 +413,12 @@ AssignArticleNo(Connection *conn, ArtNumAss **pan, const char *group, const char
      * drop the newsgroup on the floor.
      */
 
-    if (foundXRef == 0 && XRefHost) {
+    if (foundXRef == 0 && DOpts.ReaderXRefSlaveHost != NULL) {
 	KPDBUnlock(KDBActive, rec);
+	snprintf(logbuf, sizeof(logbuf), "%s %s no valid xref present in slave mode", msgid, group);
+	LogCmd(conn, '-', logbuf);
+	if (DRIncomingLogPat != NULL)
+	    LogIncoming("%s - %s%s\n", conn->co_Auth.dr_Host, "", logbuf);
 	return;
     }
 
@@ -332,13 +467,15 @@ AssignArticleNo(Connection *conn, ArtNumAss **pan, const char *group, const char
     }
 }
 
-void
+int
 WriteOverview(Connection *conn, ArtNumAss *an, const char *group, const char *xref, const char *art, int artLen, const char *msgid)
 {
     int artNo = 0;
     const char *body;
     char *xtmp = NULL;
-    int xtmpLen = 16 + strlen(ReportedHostName);
+    int xtmpLen = 16 + strlen(conn->co_Auth.dr_VServerDef->vs_ClusterName);
+    char logbuf[1024];
+    int err = 0;
 
     /*
      * Locate article number assignment
@@ -355,7 +492,13 @@ WriteOverview(Connection *conn, ArtNumAss *an, const char *group, const char *xr
     }
 
     if (artNo == 0)
-	return;
+	return(0);
+
+    /*
+     * XXX We should find some way to aggregate these into one log entry.
+     */
+
+    snprintf(logbuf, sizeof(logbuf), "%s %s:%d", msgid, group, artNo);
 
     /*
      * Locate start of body (we may have to append our own Xref: line)
@@ -393,11 +536,25 @@ WriteOverview(Connection *conn, ArtNumAss *an, const char *group, const char *xr
 	int iovLen = 0;
 	struct iovec iov[3];
 
-	if ((ov = GetOverInfo(group)) == NULL)
-	    return;
+	if ((ov = GetOverInfo(group)) == NULL) {
+	    syslog(LOG_ERR, "Error in GetOverInfo(%s) msgid=%s", group, msgid);
+	    LogCmd(conn, '-', logbuf);
+	    if (DRIncomingLogPat != NULL)
+		LogIncoming("%s - %s%s\n", conn->co_Auth.dr_Host, "", logbuf);
+	    MBLogPrintf(conn, &conn->co_TMBuf, "400 Error writing data\r\n");
+	    NNTerminate(conn);
+	    return(1);
+	}
 
-	if (MakeOverHFile(ov, artNo, 1) == NULL)
-	    return;
+	if (MakeOverHFile(ov, artNo, 1) == NULL) {
+	    syslog(LOG_ERR, "Error in MakeOverHFile() msgid=%s", msgid);
+	    LogCmd(conn, '-', logbuf);
+	    if (DRIncomingLogPat != NULL)
+		LogIncoming("%s - %s%s\n", conn->co_Auth.dr_Host, "", logbuf);
+	    MBLogPrintf(conn, &conn->co_TMBuf, "400 Error writing data\r\n");
+	    NNTerminate(conn);
+	    return(1);
+	}
 
 	hflock(ov->ov_HCache->od_HFd, 0, XLOCK_EX);
 	pos = lseek(ov->ov_HCache->od_HFd, 0L, 2);
@@ -413,7 +570,7 @@ WriteOverview(Connection *conn, ArtNumAss *an, const char *group, const char *xr
 	    int soff;
 	    
 	    xtmp = zalloc(&conn->co_MemPool, xtmpLen);
-	    sprintf(xtmp, "Xref: %s", ReportedHostName);
+	    sprintf(xtmp, "Xref: %s", DOpts.ReaderXRefHost);
 	    soff = strlen(xtmp);
 		
 	    for (scan = an; scan; scan = scan->an_Next) {
@@ -443,7 +600,7 @@ WriteOverview(Connection *conn, ArtNumAss *an, const char *group, const char *xr
 		actLen += iov[i].iov_len;
 	}
 
-	if (XRefHost && 
+	if (DOpts.ReaderXRefSlaveHost && 
 	    (oa = GetOverArt(ov, artNo, NULL)) != NULL &&
 	    oa->oa_MsgHash.h1 == hv.h1 &&
 	    oa->oa_MsgHash.h2 == hv.h2 &&
@@ -453,16 +610,23 @@ WriteOverview(Connection *conn, ArtNumAss *an, const char *group, const char *xr
 	     * We can detect duplicate articles in XRef slave mode.  If 
 	     * we see one, do not do anything.
 	     */
+	    LogCmd(conn, 'd', logbuf);
+	    if (DRIncomingLogPat != NULL)
+		LogIncoming("%s d %s%s\n", conn->co_Auth.dr_Host, "", logbuf);
 	    ; /* EMPTY */
-	} else if (DiabloReaderXOverMode == 0) {
+	} else if (DOpts.ReaderXOverMode == 0) {
 	    /*
 	     * Do not write xover info at all.  This mode is not really
 	     * supported by the reader but may eventually be supported
 	     * in 100% nntp-cache mode if/when we develop it.
 	     */
+	    syslog(LOG_INFO, "ReaderXOverMode0 (%d:%s)", artNo, msgid);
+	    LogCmd(conn, 'm', logbuf);
+	    if (DRIncomingLogPat != NULL)
+		LogIncoming("%s m %s%s\n", conn->co_Auth.dr_Host, "", logbuf);
 	    ; /* EMPTY */
 	} else if (
-	    DiabloReaderXOverMode == 2 ||
+	    DOpts.ReaderXOverMode == 2 ||
 	    writev(ov->ov_HCache->od_HFd, iov, iovLen) == actLen
 	) {
 	    /*
@@ -473,29 +637,42 @@ WriteOverview(Connection *conn, ArtNumAss *an, const char *group, const char *xr
 	    OverArt ovart = { 0 };
 	    off_t ovpos = 0;
 
+	    hflock(ov->ov_OFd, 0, XLOCK_EX);
+
 	    (void)GetOverArt(ov, artNo, &ovpos);
 
 	    ovart.oa_ArtNo = artNo;
-	    if (DiabloReaderXOverMode == 2)
+	    if (DOpts.ReaderXOverMode == 2)
 		ovart.oa_SeekPos = -1;
 	    else
 		ovart.oa_SeekPos = pos;
 	    ovart.oa_Bytes = actLen - 1;	/* do not include \0 */
 	    ovart.oa_MsgHash = hv;
 	    ovart.oa_TimeRcvd = (int)time(NULL);
-	    ovart.oa_ArtSize = (conn->co_ByteCounter ? conn->co_ByteCounter : conn->co_BytesHeader);
+	    ovart.oa_ArtSize = (conn->co_ByteCounter > 0.0 ? conn->co_ByteCounter : conn->co_BytesHeader);
 
 	    lseek(ov->ov_OFd, ovpos, 0);
 	    write(ov->ov_OFd, &ovart, sizeof(ovart));
+	    hflock(ov->ov_OFd, 0, XLOCK_UN);
+	    LogCmd(conn, '+', logbuf);
+	    if (DRIncomingLogPat != NULL)
+		LogIncoming("%s + %s%s\n", conn->co_Auth.dr_Host, "", logbuf);
 	} else {
 	    ftruncate(ov->ov_HCache->od_HFd, pos);
 	    logit(LOG_ERR, "error writing overview data file for %s", group);
+	    LogCmd(conn, '-', logbuf);
+	    if (DRIncomingLogPat != NULL)
+		LogIncoming("%s - %s%s\n", conn->co_Auth.dr_Host, "", logbuf);
+	    MBLogPrintf(conn, &conn->co_TMBuf, "400 Error writing data\r\n");
+	    NNTerminate(conn);
+	    err = 1;
 	}
 	hflock(ov->ov_HCache->od_HFd, 0, XLOCK_UN);
 	PutOverInfo(ov);
     }
     if (xtmp)
 	zfree(&conn->co_MemPool, xtmp, xtmpLen);
+    return(err);
 }
 
 void FlushOverCache(void)
@@ -536,6 +713,7 @@ GetOverInfo(const char *group)
     if (ov == NULL) {
 	struct stat st;
 	char *path;
+	int iter = 0;
 
 	bzero(&st, sizeof(st));
 
@@ -570,27 +748,38 @@ GetOverInfo(const char *group)
 	ov = zalloc(&SysMemPool, sizeof(OverInfo));
 	ov->ov_Group = zallocStr(&SysMemPool, group);
 
-	path = zalloc(&SysMemPool, strlen(MyGroupHome) + 32);
+	{
+	    const char *rec;
+	    int recLen;
+	    if ((rec = KPDBReadRecord(KDBActive, group, KP_LOCK, &recLen)) == NULL) {
+		return(NULL);
+	    }
+	    iter = strtol(KPDBGetField(rec, recLen, "ITER", NULL, "0"), NULL, 10);
+	    KPDBUnlock(KDBActive, rec);
+	}
+
+	path = zalloc(&SysMemPool, strlen(MyGroupHome) + 48);
 again:
 	{
-	    const char *gfname = GFName(group, "over", 0);
-	    const char *slash  = strchr(gfname, '/');
+	    const char *gfname = GFName(group, GRPFTYPE_OVER, 0, 1, iter,
+						&DOpts.ReaderGroupHashMethod);
 
-	    sprintf(path, "%s/%*.*s", 
-		MyGroupHome,
-		slash - gfname, slash - gfname, gfname
-	    );
-	    if (stat(path, &st) < 0)
-		mkdir(path, 0755);
-	    errno = 0;
-	    ov->ov_OFd = xopen(O_RDWR|O_CREAT, 0644, "%s%s", path, slash);
+	    sprintf(path, "%s/%s", MyGroupHome, gfname);
+	    ov->ov_OFd = -1;
+	    if (MakeGroupDirectory(path) == -1)
+		logit(LOG_ERR, "Error on overview dir open/create: %s (%s)",
+						path, strerror(errno));
+	    else
+		ov->ov_OFd = xopen(O_RDWR|O_CREAT, 0644, "%s", path);
 	}
 	if (ov->ov_OFd < 0) {
-	    logit(LOG_ERR, "Error on overview open/create for group %s", group);
+	    logit(LOG_ERR, "Error on overview open/create for group %s: %s (%s)",
+						group, path, strerror(errno));
 	    FreeOverInfo(ov);
 	    ov = NULL;
 	} else {
 	    OverHead oh;
+	    int r;
 
 	    /*
 	     * Leave a shared lock on the over.* file so expireover knows when
@@ -611,39 +800,43 @@ again:
 	     * check if new overview file or illegal overview file and size 
 	     * accordingly
 	     */
-	    if (read(ov->ov_OFd, &oh, sizeof(oh)) != sizeof(oh) ||
-		oh.oh_Version != OH_VERSION ||
-		oh.oh_ByteOrder != OH_BYTEORDER
-	    ) {
+	    r = (read(ov->ov_OFd, &oh, sizeof(oh)) != sizeof(oh));
+	    if (r == 0 && oh.oh_ByteOrder != OH_BYTEORDER) {
+		syslog(LOG_CRIT, "Incorrect overview byte order for %s (%s)",
+								group, path);
+		r = -1;
+	    }
+	    if (r == 0 && oh.oh_Version > OH_VERSION) {
+		syslog(LOG_CRIT, "Incorrect overview version for %s (%s)",
+								group, path);
+		r = -1;
+	    }
+	    if (r != 0) {
 		hflock(ov->ov_OFd, 0, XLOCK_EX);
-
 		/*
 		 * we have to test again after we got the lock in case
 		 * another process had a lock and adjusted the file.
 		 */
 		lseek(ov->ov_OFd, 0L, 0);
 		if (read(ov->ov_OFd, &oh, sizeof(oh)) != sizeof(oh) ||
-		    oh.oh_Version != OH_VERSION ||
 		    oh.oh_ByteOrder != OH_BYTEORDER
 		) {
-		    ExpireCtl save;
+		    OverExpire save;
 
-		    FindExpireCtl(group, &save);
+		    GetOverExpire(group, &save);
 
 		    /*
-		     * If 'aMaxArts' option not given in expireCtl
+		     * If 'aInitArts' option not given in expireCtl
 		     */
 
-		    if (save.ex_MaxArts == 0)
-			save.ex_MaxArts = DEFARTSINGROUP;
+		    if (save.oe_InitArts == 0)
+			save.oe_InitArts = DEFARTSINGROUP;
 
-		    if (save.ex_MaxArts < 32)
-			save.ex_MaxArts = 32;
-		    if (save.ex_MaxArts > 65536)
-			save.ex_MaxArts = 65536;
+		    if (save.oe_InitArts > 65536)
+			save.oe_InitArts = 65536;
 
 		    ftruncate(ov->ov_OFd, 0);
-		    st.st_size = sizeof(oh) + sizeof(OverArt) * save.ex_MaxArts;
+		    st.st_size = sizeof(oh) + sizeof(OverArt) * save.oe_InitArts;
 		    ftruncate(ov->ov_OFd, st.st_size);
 		    fsync(ov->ov_OFd);
 		    lseek(ov->ov_OFd, 0L, 0);
@@ -651,17 +844,32 @@ again:
 		    oh.oh_Version = OH_VERSION;
 		    oh.oh_ByteOrder = OH_BYTEORDER;
 		    oh.oh_HeadSize = sizeof(oh);
-		    oh.oh_MaxArts = save.ex_MaxArts;
+		    oh.oh_MaxArts = save.oe_InitArts;
+		    strncpy(oh.oh_Gname, group, sizeof(oh.oh_Gname));
 		    write(ov->ov_OFd, &oh, sizeof(oh));
 		    fsync(ov->ov_OFd);
 		}
 		hflock(ov->ov_OFd, 0, XLOCK_UN);
 	    }
+	    if (oh.oh_Version > 1 && strcmp(oh.oh_Gname, group) != 0) {
+		hflock(ov->ov_OFd, 4, XLOCK_UN);
+		close(ov->ov_OFd);
+		ov->ov_OFd = -1;
+		iter++;
+		goto again;
+	    }
+	    if (iter > 0) {
+		char tsBuf[64];
+		sprintf(tsBuf,"%06d", iter);
+		KPDBWrite(KDBActive, group, "ITER", tsBuf, 0);   
+	    }
+	    ov->ov_Iter = iter;
 	    ov->ov_Size = st.st_size;
 	    ov->ov_MaxArts = (st.st_size - oh.oh_HeadSize) / sizeof(OverArt);
 	    ov->ov_Head = xmap(NULL, ov->ov_Size, PROT_READ, MAP_SHARED, ov->ov_OFd, 0);
 	    if (ov->ov_Head == NULL) {
-		logit(LOG_ERR, "Error on overview mmap for group %s", group);
+		logit(LOG_ERR, "Error on overview mmap for group %s (%s)",
+						group, strerror(errno));
 		FreeOverInfo(ov);
 		ov = NULL;
 	    } else {
@@ -672,7 +880,7 @@ again:
 		*pov = ov;
 	    }
 	}
-	zfree(&SysMemPool, path, strlen(MyGroupHome) + 32);
+	zfree(&SysMemPool, path, strlen(MyGroupHome) + 48);
     }
     if (ov)
 	++ov->ov_Refs;
@@ -702,7 +910,7 @@ GetOverArt(OverInfo *ov, int artno, off_t *ppos)
 }
 
 const char *
-GetOverRecord(OverInfo *ov, int artno, int *plen, int *alen)
+GetOverRecord(OverInfo *ov, int artno, int *plen, int *alen, TimeRestrict *tr)
 {
     int hvpos;
     int xpos;
@@ -712,12 +920,15 @@ GetOverRecord(OverInfo *ov, int artno, int *plen, int *alen)
 
     oa = GetOverArt(ov, artno, NULL);
 
-    if (oa->oa_ArtNo != artno) {
+    if (oa == NULL || oa->oa_ArtNo != artno) {
 	if (plen)
 	    *plen = 0;
 	return(NULL);
     }
 
+    if (tr && tr->tr_Time > oa->oa_TimeRcvd)
+	return(NULL);
+
     if (alen)
 	*alen = oa->oa_ArtSize;
 
@@ -882,8 +1093,10 @@ CancelOverArt(OverInfo *ov, int artNo)
 
     if (ov != NULL) { 
 	off_t ovpos = 0;
-	const OverArt *oa = GetOverArt(ov, artNo, &ovpos);
+	const OverArt *oa;
 
+	hflock(ov->ov_OFd, 0, XLOCK_EX);
+	oa = GetOverArt(ov, artNo, &ovpos);
 	if (oa->oa_ArtNo == artNo) {
 	    OverArt copy = *oa;
 
@@ -892,6 +1105,7 @@ CancelOverArt(OverInfo *ov, int artNo)
 	    write(ov->ov_OFd, &copy, sizeof(copy));
 	    r = 1;
 	}
+	hflock(ov->ov_OFd, 0, XLOCK_UN);
     }
     return(r);
 }
@@ -915,7 +1129,8 @@ MakeOverHFile(OverInfo *ov, int artNo, int create)
 	++count;
     }
     if (od == NULL) {
-	const char *gfname = GFName(ov->ov_Group, "data", artBase);
+	const char *gfname = GFName(ov->ov_Group, GRPFTYPE_DATA, artBase, 1,
+				ov->ov_Iter, &DOpts.ReaderGroupHashMethod);
 
 	*pod = od = zalloc(&SysMemPool, sizeof(OverData));
         errno = 0;
@@ -932,7 +1147,7 @@ MakeOverHFile(OverInfo *ov, int artNo, int create)
 	    *pod = od = NULL;
 	} else {
 	    od->od_ArtBase = artBase;
-	    if (count > MAX_OVERVIEW_CACHE_REGIONS) {
+	    if (count > DOpts.ReaderThreads) {
 		OverData *t = ov->ov_HData;
 		ov->ov_HData = t->od_Next;
 		FreeOverData(t);
@@ -1000,7 +1215,7 @@ NNTestOverview(Connection *conn)
     int r = -1;
 
     if ((ov = GetOverInfo(conn->co_GroupName)) != NULL) {
-	if (GetOverRecord(ov, conn->co_ArtNo, NULL, NULL) != NULL)
+	if (GetOverRecord(ov, conn->co_ArtNo, NULL, NULL, NULL) != NULL)
 	    r = 0;
 	PutOverInfo(ov);
     }
@@ -1014,25 +1229,30 @@ NNRetrieveHead(Connection *conn, int *povlen, const char **pmsgid)
     const char *res = NULL;
 
     *povlen = 0;
-    *pmsgid = "<>";
+    if (pmsgid != NULL)
+	*pmsgid = "<>";
 
     if (conn->co_GroupName == NULL) {
 	return(NULL);
     }
 
     if ((ov = GetOverInfo(conn->co_GroupName)) != NULL) {
-	if ((res = GetOverRecord(ov, conn->co_ArtNo, povlen, NULL)) != NULL) {
-	    /*
-	     * Locate the Message-ID
-	     */
+	if ((res = GetOverRecord(ov, conn->co_ArtNo, povlen, NULL, NULL)) != NULL) {
+
 	    const char *scan = res;
 	    int scanLen = *povlen;
 
+	    /*
+	     * Locate and extract the Message-ID
+	     */
+
 	    while (scanLen > 0) {
 		int i;
+		char ch = tolower(scan[0]);
 		for (i = 0; i < scanLen && scan[i] != '\n'; ++i)
 		    ;
-		if (strncasecmp(scan, "Message-ID:", 11) == 0) {
+		if (pmsgid != NULL && ch == 'm' &&
+				strncasecmp(scan, "Message-ID:", 11) == 0) {
 		    int b = 11;
 		    int e;
 		    char buf[MAXMSGIDLEN];
@@ -1047,12 +1267,18 @@ NNRetrieveHead(Connection *conn, int *povlen, const char **pmsgid)
 		    if (e - b < MAXMSGIDLEN) {
 			bcopy(scan + b, buf, e - b);
 			buf[e-b] = 0;
-			*pmsgid = MsgId(buf);
+			*pmsgid = MsgId(buf, NULL);
 		    }
+		    /*
+		     * Stop scanning through headers
+		     */
 		    break;
 		}
-		if (i == 1 && scan[0] == '\r')
+
+		if (i == 1 && scan[0] == '\r') {
+		    *povlen -= scanLen;
 		    break;
+		}
 		if (i < scanLen)
 		    ++i;
 		scanLen -= i;
@@ -1061,7 +1287,7 @@ NNRetrieveHead(Connection *conn, int *povlen, const char **pmsgid)
 	} 
 	PutOverInfo(ov);
     }
-    if (strcmp(*pmsgid, "<>") == 0)
+    if (pmsgid != NULL && strcmp(*pmsgid, "<>") == 0)
 	res = NULL;
     return(res);
 }
