@@ -17,25 +17,70 @@ Prototype void NNCommand2(Connection *conn);
 Prototype void NNBadCommandUse(Connection *conn);
 Prototype void NNUnknownCommand(Connection *conn);
 Prototype void NNTerminate(Connection *conn);
+Prototype void NNAuthDone(Connection *conn);
 Prototype void NNWriteHello(Connection *conn);
 Prototype void NNWaitThread(Connection *conn);
 Prototype void NNTPHelp(Connection *conn, char **pptr);
 Prototype Connection *InitConnection(ForkDesc *desc, DnsRes *dres);
 Prototype void DeleteConnection(Connection *conn);
 Prototype void StatusUpdate(Connection *conn, const char *ctl, ...);
+Prototype void LogCmd(Connection *conn, char dir, char *msg);
+Prototype void GroupStats(Connection *conn);
 
 Prototype KPDB *KDBActive;
 
 void HandleReaderMsg(ForkDesc *desc);
 int makeReaderSlot(void);
 void freeReaderSlot(int slot);
+void sigHup(int SigNo);
 void sigSegVReader(int sigNo);
 
 KPDB *KDBActive;
 int   TFd = -1;		/* thread's return fd to the parent */
 int   NumReaders;
 int   *ReaderSlotAry;
-int   TerminatePending;
+int   TerminatePending = 0;
+int   CanTerminate = 0;
+time_t   TerminateTime = 0;
+
+void LogCmd(Connection *conn, char dir, char *msg)
+{
+    char *ptr, tmp = '\0';
+    int doLog;
+
+    doLog = DOpts.ReaderDetailLog;
+    if (conn->co_Auth.dr_ReaderDef &&
+	conn->co_Auth.dr_ReaderDef->rd_LogCmd >= 0)
+	doLog = conn->co_Auth.dr_ReaderDef->rd_LogCmd;
+    if (doLog == 0)
+	return;
+
+    if (conn->co_Desc->d_Slot < 0) {
+	/*
+	 * Oh, lame.  Diablo generally doesn't strip \r from inputted text
+	 * lines, and seeing caret-M in syslog drives me nuts.  I'm not
+	 * sure which is more lame, the problem or the hack/workaround
+	 *
+	 * JG200012210208
+	 */
+	if ((ptr = strpbrk(msg, "\r\n"))) {
+	    tmp = *ptr;
+	    *ptr = '\0';
+	}
+	logit(LOG_DEBUG, "%s [%d] %c %s", conn->co_Desc->d_Id, getpid(), dir, msg);
+	if (ptr) {
+	    *ptr = tmp;
+	}
+    } else {
+	logit(LOG_DEBUG, "%s%s%s%s%s [%d/%d] %c %s", 
+		*conn->co_Auth.dr_AuthUser ? conn->co_Auth.dr_AuthUser : "",
+		*conn->co_Auth.dr_AuthUser ? "/" : "",
+		*conn->co_Auth.dr_IdentUser ? conn->co_Auth.dr_IdentUser : "",
+		*conn->co_Auth.dr_IdentUser ? "@" : "",
+		conn->co_Auth.dr_Host, getpid(), conn->co_Desc->d_Slot,
+		dir, msg);
+    }
+}
 
 void
 ReaderTask(int fd, const char *id)
@@ -44,20 +89,30 @@ ReaderTask(int fd, const char *id)
     time_t ltime = 0;
     time_t itime = 0;
     time_t ftime = 0;
+    time_t atime = 0;
     int counter = 0;
+    int forceallcheck = 0;
+    int check_disconn_counter = 0;
 
     TFd = fd;
 
     /*
      * [re]open RTStatus
      */
-    RTStatusOpen(RTStatus, ThisReaderFork * DiabloReaderThreads + 1, DiabloReaderThreads);
+    RTStatusOpen(RTStatus, ThisReaderFork * DOpts.ReaderThreads + 1, DOpts.ReaderThreads);
 
     /*
      * Since we setuid(), we won't core.  This is for debugging
      */
-    if (CoreDebugOpt)
+    if (CoreDebugOpt || (DOpts.ReaderCrashHandler != NULL &&
+			strcasecmp(DOpts.ReaderCrashHandler, "none") != 0)) {
 	signal(SIGSEGV, sigSegVReader);
+	signal(SIGBUS, sigSegVReader);
+	signal(SIGFPE, sigSegVReader);
+	signal(SIGILL, sigSegVReader);
+    }
+
+    signal(SIGHUP, sigHup);
 
     /*
      * Setup thread for passed pipe
@@ -77,13 +132,19 @@ ReaderTask(int fd, const char *id)
 	exit(1);
     }
 
+    LoadExpireCtl(1);
+
+    /*
+     * Only startup connections to backend spools for reader threads
+     */
+    if (!FeedOnlyServer)
+	CheckServerConfig(time(NULL), 1);
+
     /*
      * Selection core
      */
 
-    CheckServerConfig(time(NULL), 1);
-
-    while (TerminatePending == 0 || NReadServAct || NumReaders) {
+    while (!TerminatePending || NReadServAct || NumReaders) {
 	/*
 	 *  select core
 	 */
@@ -93,13 +154,27 @@ ReaderTask(int fd, const char *id)
 	fd_set read_only_to_find_eof_fds;
 	int i, sel_r;
 
+	if (TerminatePending) {
+	    if (TerminateTime == 0)
+		TerminateTime = time(NULL) + 2;
+	    if (TerminateTime < time(NULL) || !NumReaders)
+		CanTerminate = 1;
+	}
+
 	/*
 	 *  Get next scheduled timeout, no more then 2 seconds
 	 *  (x 10 counter counts = 20 seconds max for {d,i,f}time 
 	 *  check)
+	 *
+	 * If we are terminating, then speed up the select to clear
+	 * out the connections.
+	 *
 	 */
 
-	NextTimeout(&tv, 2 * 1000);
+	if (TerminatePending)
+	    NextTimeout(&tv, 50);
+	else
+	    NextTimeout(&tv, 2 * 1000);
 
 	stprintf("%s readers=%02d spoolsrv=%d/%d postsrv=%d/%d",
 	    id,
@@ -108,18 +183,29 @@ ReaderTask(int fd, const char *id)
 	    NWriteServAct, NWriteServers
 	);
 
+	/* Check for disconnected clients every 50 times through the loop */
 	FD_ZERO(&read_only_to_find_eof_fds);
-	for (i = 0; i < MaxFds; ++i) {
-	    if (FD_ISSET(i, &wfds) && (!(FD_ISSET(i, &rfds)))) {
-		FD_SET(i, &rfds);
-		FD_SET(i, &read_only_to_find_eof_fds);
+	if (++check_disconn_counter == 50) {
+	    for (i = 0; i < MaxFds; ++i) {
+		if (FD_ISSET(i, &wfds) && (!(FD_ISSET(i, &rfds)))) {
+		    FD_SET(i, &rfds);
+		    FD_SET(i, &read_only_to_find_eof_fds);
+		}
 	    }
+	    check_disconn_counter = 0;
 	}
 
+#if USE_AIO
+	AIOUnblockSignal();
+#endif
 	sel_r = select(MaxFds, &rfds, &wfds, NULL, &tv);
+#if USE_AIO
+	AIOBlockSignal();
+#endif
+
 	gettimeofday(&CurTime, NULL);
 
-	if(sel_r < 0)
+	if(sel_r < 0 && errno != EINTR)
 	    logit(LOG_CRIT,
 		  "select error: %s (rfds=0x%x, wfds=0x%x)",
 		  strerror(errno),
@@ -145,13 +231,15 @@ ReaderTask(int fd, const char *id)
 		    dtime += t - ltime;
 		    itime += t - ltime;
 		    ftime += t - ltime;
+		    atime += t - ltime;
 		}
 
 		/*
 		 * Check for server config change once a minute
 		 */
-		if (dtime < -5 || dtime >= 60) {
-		    CheckServerConfig(t, ServersTerminated);
+		if (dtime < -5 || dtime >= 5) {
+		    if (!TerminatePending)
+			CheckServerConfig(t, ServersTerminated);
 		    dtime = 0;
 		}
 
@@ -160,6 +248,7 @@ ReaderTask(int fd, const char *id)
 		 */
 		if (ftime < -5 || ftime >= 30) {
 		    FlushOverCache();
+		    LoadExpireCtl(0);
 		    ftime = 0;
 		}
 
@@ -173,6 +262,15 @@ ReaderTask(int fd, const char *id)
 		    rfds = RFds;
 		    itime = 0;
 		}
+
+		/*
+		 * Force a check all of FD's every 30 seconds to handle
+		 * idle and session timeouts
+		 */
+		if (atime < -5 || atime >= 30) {
+		    forceallcheck = 1;
+		    atime = 0;
+		}
 		ltime = t;
 		counter = 0;
 	    }
@@ -182,9 +280,10 @@ ReaderTask(int fd, const char *id)
 	     * cache every 5 minutes, and with a greater granularity.
 	     * It should cycle much faster than that normally, and this
 	     * is to prevent idle feed-only forks from keeping locks.
+	     *
 	     */
 
-	    if (++counter == 50) {
+	    if (++counter == 10) {
 		time_t t = CurTime.tv_sec;
 		if (ltime) {
 		    ftime += t - ltime;
@@ -194,6 +293,8 @@ ReaderTask(int fd, const char *id)
 		    FlushOverCache();
 		    ftime = 0;
 		}
+		ltime = t;
+		counter = 0;
 	    }
 	}
 
@@ -221,29 +322,20 @@ ReaderTask(int fd, const char *id)
 		    ForkDesc *desc;
 		    if((desc = FindThread(i, -1)) != NULL) {
 			Connection *conn = desc->d_Data;
-			int still_in_use = 0;
 
 			if(conn) {
-			    /*
-			     * If this client has an outstanding server request,
-			     * avoid killing it, because this conn is referenced
-			     * elsewhere.
-			     */
-			    if(conn->co_SReq != NULL)
-				still_in_use = 1;
-
-			    if(!still_in_use)
-				DeleteConnection(conn);
+			    NNTerminate(conn);
+			    DeleteConnection(conn);
 			}
-			if(!still_in_use)
-			    DelThread(desc);
+			DelThread(desc);
 		    }
 		}
 	    }
 	}
 
 	for (i = 0; i < MaxFds; ++i) {
-	    if (FD_ISSET(i, &rfds) || FD_ISSET(i, &wfds)) {
+	    if (forceallcheck || TerminatePending ||
+		FD_ISSET(i, &rfds) || FD_ISSET(i, &wfds)) {
 		ForkDesc *desc;
 
 		if ((desc = FindThread(i, -1)) != NULL) {
@@ -256,7 +348,7 @@ ReaderTask(int fd, const char *id)
 
 			MBFlush(conn, &conn->co_TMBuf);
 			conn->co_FCounter = 0;
-		    }
+		    } 
 
 		    /*
 		     * Function dispatch
@@ -264,12 +356,21 @@ ReaderTask(int fd, const char *id)
 
 		    switch(desc->d_Type) {
 		    case THREAD_READER:
-			HandleReaderMsg(desc);
+			if (FD_ISSET(i, &rfds) || FD_ISSET(i, &wfds))
+			    HandleReaderMsg(desc);
 			break;
 		    case THREAD_NNTP:		/* client	  */
+			conn->co_Func(conn);
+			if (conn->co_Auth.dr_ResultFlags & DR_REQUIRE_DNS) {
+			    /* Go back to parent for DNS check */
+			    conn->co_Auth.dr_Code = 0;
+			    conn->co_TMBuf.mh_WEof = 1;
+			}
+			break;
 		    case THREAD_SPOOL:		/* spool server	  */
 		    case THREAD_POST:		/* posting server */
 			conn->co_Func(conn);
+			LogServerInfo(conn, TFd);
 			break;
 		    default:
 			/* panic */
@@ -286,9 +387,45 @@ ReaderTask(int fd, const char *id)
 		     */
 
 		    if (conn) {
-			if (conn->co_RMBuf.mh_REof && 
+			int idleTimeout = 0;
+			if (conn->co_Auth.dr_ReaderDef) {
+			    if (conn->co_Auth.dr_ReaderDef->rd_IdleTimeout &&
+				    conn->co_LastActiveTime +
+				    conn->co_Auth.dr_ReaderDef->rd_IdleTimeout <=
+				    CurTime.tv_sec) {
+				logit(LOG_INFO, "timeout idle %s",
+				    conn->co_Auth.dr_Host);
+				MBLogPrintf(conn,
+				    &conn->co_TMBuf,
+				    "400 %s: Idle timeout.\r\n",
+				    conn->co_Auth.dr_VServerDef->vs_HostName
+				);
+				idleTimeout = 1;
+				NNTerminate(conn);
+			    }
+			    if (conn->co_Auth.dr_ReaderDef->rd_SessionTimeout &&
+				    conn->co_SessionStartTime +
+				    conn->co_Auth.dr_ReaderDef->rd_SessionTimeout <=
+				    CurTime.tv_sec) {
+				logit(LOG_INFO, "timeout session %s",
+				    conn->co_Auth.dr_Host);
+				MBLogPrintf(conn,
+				    &conn->co_TMBuf,
+				    "400 %s: Session timeout.\r\n",
+				    conn->co_Auth.dr_VServerDef->vs_HostName
+				);
+				idleTimeout = 1;
+				NNTerminate(conn);
+			    }
+			}
+			if ((!conn->co_Auth.dr_Code &&
+				desc->d_Type == THREAD_NNTP) ||
+			    idleTimeout ||
+			    (conn->co_RMBuf.mh_REof && 
 			    conn->co_TMBuf.mh_WEof &&
-			    conn->co_TMBuf.mh_MBuf == NULL
+			    conn->co_TMBuf.mh_MBuf == NULL) ||
+			    (TerminatePending &&
+					!(conn->co_Flags & COF_MAYNOTCLOSE))
 			) {
 			    DeleteConnection(conn);
 			    DelThread(desc);
@@ -297,7 +434,10 @@ ReaderTask(int fd, const char *id)
 		}
 	    }
 	}
+	forceallcheck = 0;
 	(void)ScanTimers(1, 0);
+	if (CanTerminate)
+	    break;
     }
 }
 
@@ -305,76 +445,66 @@ void
 HandleReaderMsg(ForkDesc *desc)
 {
     int r;
+    int recv_fd;
     DnsRes  dres;
-    struct msghdr msg;
-    struct iovec  iov;
-    struct {
-#if FDPASS_USES_CMSG
-	struct cmsghdr cmsg;
-#endif
-	int fd;
-    } cmsg;
 
-    bzero(&msg, sizeof(msg));
-
-    iov.iov_base = (void *)&dres;
-    iov.iov_len = sizeof(dres);
-
-    msg.msg_iov = &iov;
-    msg.msg_iovlen = 1;
-#if FDPASS_USES_ACC
-    msg.msg_accrights = (caddr_t)&cmsg;
-    msg.msg_accrightslen = sizeof(cmsg);
-#else
-    msg.msg_control = (caddr_t)&cmsg;
-    msg.msg_controllen = sizeof(cmsg);
-#endif
-#if FDPASS_USES_CMSG
-    msg.msg_flags = 0;
-    cmsg.cmsg.cmsg_len = sizeof(cmsg);
-#endif
-    cmsg.fd = -1;
-    errno = 0;
-
-    /*
-     * receive message w/ file descriptor
-     */
-
-    if ((r = recvmsg(desc->d_Fd, &msg, MSG_EOR|MSG_WAITALL)) == sizeof(dres)) {
-	if (DebugOpt)
-	    printf("RECV MSG\n");
-
-	if (cmsg.fd >= MAXFDS) {
-	    logit(LOG_WARNING, "fd too large %d/%d, increase MAXFDS for select. Closing fd", cmsg.fd, MAXFDS);
+    if ((r = RecvMsg(desc->d_Fd, &recv_fd, &dres)) == sizeof(DnsRes)) {
+	if (recv_fd >= MAXFDS) {
+	    logit(LOG_WARNING, "fd too large %d/%d, increase MAXFDS for select. Closing fd", recv_fd, MAXFDS);
 	    /*
 	     * Tell the main server that we are done with the connection
 	     */
 	    fcntl(TFd, F_SETFL, 0);
-	    write(TFd, &dres, sizeof(dres));
+	    SendMsg(TFd, recv_fd, &dres);
 	    fcntl(TFd, F_SETFL, O_NONBLOCK);
-	} else if (cmsg.fd >= 0) {
+	} else if (recv_fd >= 0) {
 	    ForkDesc *ndesc;
 	    Connection *conn;
+	    char vsbuf[11];
+	    char hsbuf[31];
 
-	    ndesc = AddThread("client", cmsg.fd, -1, THREAD_NNTP, makeReaderSlot(), 0);
+	    if (ReadAccessCache() == 1) {
+		ScanThreads(THREAD_NNTP, UpdateAuthDetails);
+		ScanThreads(THREAD_SPOOL, UpdateAuthDetails);
+		ScanThreads(THREAD_READER, UpdateAuthDetails);
+		ScanThreads(THREAD_SPOOL, UpdateAuthDetails);
+		ScanThreads(THREAD_FEEDER, UpdateAuthDetails);
+		ClearOldAccessMap();
+	    }
+	    SetAuthDetails(&dres, dres.dr_ReaderName);
+
+	    ndesc = AddThread("client", recv_fd, -1, THREAD_NNTP, makeReaderSlot(), 0);
 	    ++NumReaders;
 	    if (DebugOpt)
-		printf("add thread fd=%d\n", cmsg.fd);
+		printf("add thread fd=%d\n", recv_fd);
 
 	    FD_SET(ndesc->d_Fd, &WFds);	/* will cause immediate effect */
 	    conn = InitConnection(ndesc, &dres);
 	    if (conn->co_Auth.dr_Flags & DF_FEED)
 		conn->co_Flags |= COF_SERVER;
-	    RTStatusBase(ndesc->d_Slot, "ACTV %-30s", dres.dr_Host);
+
+	    snprintf(vsbuf, sizeof(vsbuf), "%s", conn->co_Auth.dr_VServerDef->vs_Name);
+
+	    snprintf(hsbuf, sizeof(hsbuf), "%s%s%s%s%s",
+		*conn->co_Auth.dr_AuthUser ? conn->co_Auth.dr_AuthUser : "",
+		*conn->co_Auth.dr_AuthUser ? "/" : "",
+		*conn->co_Auth.dr_IdentUser ? conn->co_Auth.dr_IdentUser : "",
+		*conn->co_Auth.dr_IdentUser ? "@" : "",
+		conn->co_Auth.dr_Host);
+	    RTStatusBase(conn->co_Desc->d_Slot, "ACTV %-10s %-30s", vsbuf, hsbuf);
+
 	    StatusUpdate(conn, "(startup)");
-	    NNWriteHello(conn);
+	    if (conn->co_Auth.dr_ResultFlags & DR_REQUIRE_DNS)
+		NNAuthDone(conn);
+	    else
+		NNWriteHello(conn);
 	} else {
 	    if (DebugOpt)
 		printf("recvmsg(): EOF1\n");
 	    DelThread(desc);
 	    TerminatePending = 1;
 	}
-    } 
+    }
 
     /*
      * If recv error, check errno.  If temporary error,
@@ -382,6 +512,10 @@ HandleReaderMsg(ForkDesc *desc)
      * terminate.
      */
 
+    if (r != sizeof(DnsRes)) {
+	if (DebugOpt)
+	    printf("recvmsg(): Bad size read from RecvMsg\n");
+    }
     if (r < 0) {
 	if (errno != EINTR &&
 	    errno != EWOULDBLOCK &&
@@ -417,10 +551,19 @@ InitConnection(ForkDesc *desc, DnsRes *dres)
     conn->co_Desc = desc;
     conn->co_MemPool = pool;
     gettimeofday(&conn->co_RateTv, NULL);
+    conn->co_LastActiveTime = conn->co_SessionStartTime = conn->co_RateTv.tv_sec;
     MBInit(&conn->co_TMBuf, desc->d_Fd, &conn->co_MemPool, &conn->co_BufPool);
     MBInit(&conn->co_RMBuf, desc->d_Fd, &conn->co_MemPool, &conn->co_BufPool);
     MBInit(&conn->co_ArtBuf, -1, &conn->co_MemPool, &conn->co_BufPool);
 
+    if (conn->co_Desc->d_Type == THREAD_NNTP) {
+	conn->co_ClientTotalByteCount = 0.0;
+	conn->co_ClientTotalArticleCount = 0;
+	conn->co_ClientGroupByteCount = 0.0;
+	conn->co_ClientGroupArticleCount = 0;
+	conn->co_ClientGroupCount = conn->co_ClientPostCount = 0;
+    }
+
     return(conn);
 }
 
@@ -429,8 +572,29 @@ DeleteConnection(Connection *conn)
 {
     MemPool *mpool = conn->co_MemPool;
 
-    if (conn->co_Desc->d_Type == THREAD_NNTP) {
-	RTStatusBase(conn->co_Desc->d_Slot, "CLSD %-30s", conn->co_Auth.dr_Host);
+    if (conn->co_Desc->d_Type == THREAD_SPOOL) {
+	conn->co_LastServerLog = 1;
+	LogServerInfo(conn, TFd);
+    } else if (conn->co_Desc->d_Type == THREAD_NNTP) {
+	char statbuf[1024];
+	char vsbuf[11];
+	char hsbuf[31];
+
+	snprintf(vsbuf, sizeof(vsbuf), "%s", conn->co_Auth.dr_VServerDef->vs_Name);
+
+	snprintf(hsbuf, sizeof(hsbuf), "%s%s%s%s%s",
+		*conn->co_Auth.dr_AuthUser ? conn->co_Auth.dr_AuthUser : "",
+		*conn->co_Auth.dr_AuthUser ? "/" : "",
+		*conn->co_Auth.dr_IdentUser ? conn->co_Auth.dr_IdentUser : "",
+		*conn->co_Auth.dr_IdentUser ? "@" : "",
+		conn->co_Auth.dr_Host);
+	RTStatusBase(conn->co_Desc->d_Slot, "CLSD %-10s %-30s", vsbuf, hsbuf);
+
+	GroupStats(conn);
+
+	snprintf(statbuf, sizeof(statbuf), "exit articles %lu groups %lu posts %lu bytes %.0f", conn->co_ClientTotalArticleCount, conn->co_ClientGroupCount, conn->co_ClientPostCount, conn->co_ClientTotalByteCount);
+	LogCmd(conn, '$', statbuf);
+
 	StatusUpdate(conn, "(closed)");
 	freeReaderSlot(conn->co_Desc->d_Slot);
 	--NumReaders;
@@ -441,9 +605,7 @@ DeleteConnection(Connection *conn)
 	 * connection terminates.
 	 */
 	conn->co_Auth.dr_ByteCount = conn->co_TMBuf.mh_TotalBytes;
-	fcntl(TFd, F_SETFL, 0);
-	write(TFd, &conn->co_Auth, sizeof(conn->co_Auth));
-	fcntl(TFd, F_SETFL, O_NONBLOCK);
+	SendMsg(TFd, conn->co_Desc->d_Fd, &conn->co_Auth);
     }
 
     FreeControl(conn);
@@ -480,6 +642,73 @@ NNTerminate(Connection *conn)
 	conn->co_SReq->sr_CConn = NULL;
 	conn->co_SReq = NULL;
     }
+
+    /*
+     * If this connection was performing a list operation and had the
+     * active cache locked, unlock it. If it was building the cache,
+     * blast the cache because it's half-built.
+     */
+    if (conn->co_ListCacheMode == ACMODE_READ)
+	ActiveCacheReadUnlock();
+    if (conn->co_ListCacheMode == ACMODE_WRITE) {
+	ActiveCacheFreeMain();
+	ActiveCacheWriteUnlock();
+    }
+    MBFlush(conn, &conn->co_TMBuf);
+}
+
+void
+NNAuthDone(Connection *conn)
+{
+    conn->co_Auth.dr_ResultFlags = 0;
+    if (conn->co_Auth.dr_Code == 1 && (conn->co_Auth.dr_Flags & DF_AUTHREQUIRED) == 0) {
+	conn->co_Auth.dr_Flags &= ~DF_AUTHREQUIRED;
+	MBLogPrintf(conn, &conn->co_TMBuf, "281 Ok\r\n");
+	syslog(LOG_DEBUG,
+		"Access granted to %s via %s", conn->co_Auth.dr_Host,
+		conn->co_Auth.dr_ReaderName);
+	NNCommand(conn);
+    } else {
+	MBLogPrintf(conn, &conn->co_TMBuf, "502 Authentication error\r\n");
+	NNTerminate(conn);
+    }
+}
+
+char *
+WelcomeString(char *base, Vserver *vs, const char *servertype)
+{
+    static char r[512];
+    int l = 0;
+
+    r[0] = 0;
+    while (l < sizeof(r) && *base) {
+	if (*base == '%') {
+	    switch (*(++base)) {
+		case 'a':
+		    l += snprintf(&r[l], 512 - l, "%s", vs->vs_NewsAdm);
+		    break;
+		case 'c':
+		    l += snprintf(&r[l], 512 - l, "%s", vs->vs_ClusterName);
+		    break;
+		case 'h':
+		    l += snprintf(&r[l], 512 - l, "%s", vs->vs_HostName);
+		    break;
+		case 'o':
+		    l += snprintf(&r[l], 512 - l, "%s", vs->vs_Org);
+		    break;
+		case 't':
+		    l += snprintf(&r[l], 512 - l, "%s", servertype);
+		    break;
+		case 'v':
+		    l += snprintf(&r[l], 512 - l, "Diablo %s-%s", VERS, SUBREV);
+		    break;
+	    }
+	    base++;
+	} else {
+	    r[l++] = *base++;
+	}
+    }
+    return(r);
 }
 
 void
@@ -487,33 +716,50 @@ NNWriteHello(Connection *conn)
 {
     const char *postingOk = "(no posting)";
     const char *noReading = "";
-    const char *serverType = "NNTP";
+    const char *serverType = "NNRP";
 
     if (conn->co_Auth.dr_Flags & DF_POST)
 	postingOk = "(posting ok)";
     if (conn->co_Flags & COF_SERVER) {
-	serverType = "FEED";
+	serverType = "NNTP-FEED";
     } else if ((conn->co_Auth.dr_Flags & DF_READ) == 0) {
 	noReading = "(no reading)";
     }
-    if ((conn->co_Auth.dr_Flags & (DF_FEED|DF_READ|DF_POST)) == 0) {
-	MBPrintf(
+    if ((conn->co_Auth.dr_Flags & (DF_FEED|DF_READ|DF_POST|DF_AUTHREQUIRED)) == 0) {
+	MBLogPrintf(conn,
 	    &conn->co_TMBuf, 
-	    "500 %s Diablo Server, you have no permissions.\r\n",
-	    ReportedHostName
+	    "502 %s: Access denied to your node%s%s.\r\n",
+	    conn->co_Auth.dr_VServerDef->vs_HostName,
+	    *conn->co_Auth.dr_VServerDef->vs_NewsAdm ? " - " : "",
+	    conn->co_Auth.dr_VServerDef->vs_NewsAdm
 	);
 	NNTerminate(conn);
 	return;
     }
 
-    MBPrintf(
-	&conn->co_TMBuf, 
-	"200 %s Diablo %s Server ready %s%s.\r\n",
-	ReportedHostName,
-	serverType,
-	postingOk,
-	noReading
-    );
+    if (*conn->co_Auth.dr_VServerDef->vs_Welcome)
+	MBLogPrintf(conn,
+		&conn->co_TMBuf, 
+		"%d %s %s%s\r\n",
+		(conn->co_Auth.dr_Flags & DF_POST) ? 200 : 201,
+		WelcomeString(conn->co_Auth.dr_VServerDef->vs_Welcome,
+					conn->co_Auth.dr_VServerDef,
+					serverType),
+		postingOk,
+		noReading
+	);
+    else
+	MBLogPrintf(conn,
+		&conn->co_TMBuf, 
+		"%d %s %s Service Ready%s%s %s%s.\r\n",
+		(conn->co_Auth.dr_Flags & DF_POST) ? 200 : 201,
+		conn->co_Auth.dr_VServerDef->vs_HostName,
+		serverType,
+		*conn->co_Auth.dr_VServerDef->vs_NewsAdm ? " - " : "",
+		conn->co_Auth.dr_VServerDef->vs_NewsAdm,
+		postingOk,
+		noReading
+	);
     NNCommand(conn);
 }
 
@@ -629,7 +875,7 @@ Command Cmds[] = {
 	"list",
 	CMDF_AUTH|0	     |CMDF_READER,
 	NNTPList,
-	"[active|active.times|newsgroups|distributions|distrib.pats|moderators|overview.fmt|subscriptions]"
+	"[active|active.times|newsgroups|extensions|distributions|distrib.pats|moderators|overview.fmt|subscriptions]"
     },
     { 
 	"listgroup",
@@ -647,7 +893,7 @@ Command Cmds[] = {
 	"newnews",
 	CMDF_AUTH|0	     |CMDF_READER,
 	NNTPNewNews,
-	"(not implemented)"
+	"newsgroups yymmdd hhmmss [\"GMT\"] [<distribution>]"
     },
     { 
 	"post",	
@@ -673,6 +919,18 @@ Command Cmds[] = {
 	NNTPXHdr,
 	"header [range|MessageID]"
     },
+    { 
+	"hdr",
+	CMDF_AUTH|0	     |CMDF_READER,
+	NNTPXHdr,
+	"header [range|MessageID]"
+    },
+    { 
+	"over",
+	CMDF_AUTH|0	     |CMDF_READER,
+	NNTPXOver,
+	"[range]"
+    },
     { 
 	"xover",
 	CMDF_AUTH|0	     |CMDF_READER,
@@ -702,8 +960,36 @@ Command Cmds[] = {
 void 
 NNCommand(Connection *conn)
 {
+
+    /*
+     * Check for changes to the access cache before every command
+     */
+
+    if (ReadAccessCache() == 1) {
+	ScanThreads(THREAD_NNTP, UpdateAuthDetails);
+	ScanThreads(THREAD_SPOOL, UpdateAuthDetails);
+	ScanThreads(THREAD_READER, UpdateAuthDetails);
+	ScanThreads(THREAD_SPOOL, UpdateAuthDetails);
+	ScanThreads(THREAD_FEEDER, UpdateAuthDetails);
+	ClearOldAccessMap();
+    }
+
     MBFlush(conn, &conn->co_TMBuf);
+
+    if (conn->co_Auth.dr_ReaderDef->rd_ByteLimit &&
+	    conn->co_Auth.dr_ReaderDef->rd_ByteLimit <
+	    conn->co_ClientTotalByteCount) {
+	logit(LOG_INFO, "bytelimit session %s",
+	    conn->co_Auth.dr_Host);
+	MBLogPrintf(conn,
+	    &conn->co_TMBuf,
+	    "400 %s: Session byte limit reached.\r\n",
+	    conn->co_Auth.dr_VServerDef->vs_HostName
+	);
+	NNTerminate(conn);
+    }
     NNCommand2(conn);
+    conn->co_LastActiveTime = CurTime.tv_sec;
 }
 
 void
@@ -762,6 +1048,12 @@ NNCommand2(Connection *conn)
     if (DebugOpt)
 	printf("command: %s\n", ptr);
 
+    if (strncasecmp(ptr, "authinfo pass ", 14)) {
+        LogCmd(conn, '<', ptr);
+    } else {
+        LogCmd(conn, '<', "authinfo pass **unlogged**");
+    }
+
     /*
      * extract command (note: StatusUpdate() will limit the line length)
      */
@@ -802,7 +1094,7 @@ NNCommand2(Connection *conn)
 	if ((scan->cmd_Flags & CMDF_AUTH) &&
 	    (conn->co_Auth.dr_Flags & DF_AUTHREQUIRED)
 	) {
-	    MBPrintf(&conn->co_TMBuf, "480 Authentication required for command\r\n");
+	    MBLogPrintf(conn, &conn->co_TMBuf, "480 Authentication required for command\r\n");
 	    NNCommand(conn);
 	} else {
 	    scan->cmd_Func(conn, &ptr);
@@ -815,14 +1107,14 @@ NNCommand2(Connection *conn)
 void
 NNBadCommandUse(Connection *conn)
 {
-    MBPrintf(&conn->co_TMBuf, "501 Bad command use\r\n");
+    MBLogPrintf(conn, &conn->co_TMBuf, "501 Bad command use\r\n");
     NNCommand(conn);
 }
 
 void
 NNUnknownCommand(Connection *conn)
 {
-    MBPrintf(&conn->co_TMBuf, "500 What?\r\n");
+    MBLogPrintf(conn, &conn->co_TMBuf, "500 What?\r\n");
     NNCommand(conn);
 }
 
@@ -848,7 +1140,7 @@ NNTPHelp(Connection *conn, char **pptr)
 {
     Command *scan;
 
-    MBPrintf(&conn->co_TMBuf, "100 Legal commands\r\n");
+    MBLogPrintf(conn, &conn->co_TMBuf, "100 Legal commands\r\n");
     for (scan = &Cmds[0]; scan < &Cmds[arysize(Cmds)]; ++scan) {
 	int ok = 0;
 
@@ -864,6 +1156,9 @@ NNTPHelp(Connection *conn, char **pptr)
 	} else {
 	    if (scan->cmd_Flags & CMDF_READER)
 		ok = 1;
+	    if (!conn->co_Auth.dr_ReaderDef->rd_AllowNewnews &&
+				(strcmp(scan->cmd_Name, "newnews") == 0))
+		ok = 0;
 	}
 	if (ok)
 	    MBPrintf(&conn->co_TMBuf, "  %s %s\r\n", scan->cmd_Name, scan->cmd_Help);
@@ -882,8 +1177,12 @@ StatusUpdate(Connection *conn, const char *ctl, ...)
 	((conn->co_Auth.dr_Flags & DF_FEED) ? 'f' : '-'),
 	((conn->co_Auth.dr_Flags & DF_READ) ? 'r' : '-'),
 	((conn->co_Auth.dr_Flags & DF_POST) ? 'p' : '-'),
-	(conn->co_GroupName ? conn->co_GroupName : "(none)")
+	(conn->co_Auth.dr_Flags & DF_GROUPLOG) ?
+	    (conn->co_GroupName ? conn->co_GroupName : "(none)") :
+	    "(none)"
     );
+    buf[34] = ' ';
+    buf[35] = '\0';
     va_start(va, ctl);
     vsnprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), ctl, va);
     va_end(va);
@@ -897,10 +1196,10 @@ makeReaderSlot(void)
     int i;
 
     if (ReaderSlotAry == NULL) {
-	ReaderSlotAry = zalloc(&SysMemPool, sizeof(*ReaderSlotAry) * DiabloReaderThreads);
+	ReaderSlotAry = zalloc(&SysMemPool, sizeof(*ReaderSlotAry) * DOpts.ReaderThreads);
     }
-    for (i = 0; i < DiabloReaderThreads; ++i) {
-	SlotNo = (SlotNo + 1) % DiabloReaderThreads;
+    for (i = 0; i < DOpts.ReaderThreads; ++i) {
+	SlotNo = (SlotNo + 1) % DOpts.ReaderThreads;
 	if (ReaderSlotAry[SlotNo] == 0) {
 	    ReaderSlotAry[SlotNo] = 1;
 	    return(SlotNo);
@@ -918,12 +1217,38 @@ freeReaderSlot(int slot)
 }
 
 void
-sigSegVReader(int sigNo)
+sigHup(int SigNo)
 {
+    TerminatePending++;
+}
+
+void
+sigSegVReader(int sigNo)
+{
+    int i;
+
     if (TFd >= 0)
 	close(TFd);
+    for (i = 0; i < MaxFds; ++i)
+	if (FD_ISSET(i, &RFds) || FD_ISSET(i, &WFds))
+	    close(i);
+
     nice(20);
-    for (;;)
-	;
+    if (DOpts.ReaderCrashHandler == NULL ||
+			strcasecmp(DOpts.ReaderCrashHandler, "none") == 0) {
+	/* Chew spare CPU cycles and try to be noticed? :-) */
+	for (;;)
+	    ;
+    } else {
+	char cmdbuf[256];
+
+	if (fork())
+	   return;
+	snprintf(cmdbuf, sizeof(cmdbuf), "%s %d", DOpts.ReaderCrashHandler, getpid());
+	system(cmdbuf);
+
+	for (;;)
+	    sleep(60);
+    }
 }
 
