@@ -10,7 +10,8 @@
 #include "defs.h"
 
 Prototype void CheckServerConfig(time_t t, int force);
-Prototype void NNArticleRetrieveByMessageId(Connection *conn, const char *msgid);
+Prototype void LogServerInfo(Connection *conn, int fd);
+Prototype void NNArticleRetrieveByMessageId(Connection *conn, const char *msgid, int grouphint);
 Prototype void NNServerIdle(Connection *conn);
 Prototype void NNServerTerminate(Connection *conn);
 Prototype void NNFinishSReq(Connection *conn, const char *ctl, int requeue);
@@ -27,10 +28,11 @@ void QueueServerRequests(void);
 int queueServerRequest(ServReq *sreq, int type, int maxq);
 void queueDesc(ServReq *sreq, ForkDesc *desc);
 int requeueServerRequest(Connection *conn, ServReq *sreq, int type, int maxq);
-int AddServer(const char *host, int type, int port, int flags, int pri);
+int AddServer(char *bindinfo, int txbufsize, int rxbufsize, const char *host, int type, int port, int flags, int pri, char *groups, const char *localspool, HashFeed *hashFeed);
 void NNServerPrime1(Connection *conn);
 void NNServerPrime2(Connection *conn);
 void NNServerPrime3(Connection *conn);
+void NNServerPrime4(Connection *conn);
 void NNServerConnect(Connection *conn);
 
 ServReq	*SReadBase;
@@ -76,12 +78,10 @@ setServerClosedRemovals(ForkDesc *desc)
 void 
 CheckServerConfig(time_t t, int force)
 {
-    static struct stat St;
+    static struct stat St = { 0 };
     struct stat st;
     const char *path = PatLibExpand(DServerHostsPat);
 
-    printf("CheckServerConfig %d %d\n", force, ServersTerminated);
-
     /*
      * Assuming we can stat the file, if we haven't checked the server config
      * before or the server config time is different from when we last checked
@@ -98,10 +98,15 @@ CheckServerConfig(time_t t, int force)
 	((long)(t - st.st_mtime) > 2 || (long)(t - st.st_mtime) < 0))
     ) {
 	FILE *fi;
-	if ((fi = fopen(path, "r")) != NULL) {
-	    char buf[256];
 
-	    St = st;
+	if (force)
+	    stat(path, &st);
+	memcpy(&St, &st, sizeof(St));
+
+	if ((fi = fopen(path, "r")) != NULL) {
+	    char buf[PATH_MAX];
+
+	    ServersTerminated = 0;
 
 	    ScanThreads(THREAD_SPOOL, setServerMaybeCloseFlag);
 	    ScanThreads(THREAD_POST, setServerMaybeCloseFlag);
@@ -115,6 +120,13 @@ CheckServerConfig(time_t t, int force)
 		int port = 119;
 		int nflags = 0;
 		int npri = 0;
+		char *option;
+		char *bindinfo = NULL;
+		char *groups = NULL;
+		char *localspool = NULL;
+		HashFeed hash = { 0 };
+		int txbufsize = 0;
+		int rxbufsize = 0;
 
 		if (host == NULL || host[0] == '#')
 		    continue;
@@ -134,18 +146,23 @@ CheckServerConfig(time_t t, int force)
 			if ((desc = FindThreadId(THREAD_SPOOL, host)) != NULL) {
 			    Connection *conn = desc->d_Data;
 			    conn->co_Flags &= ~COF_MAYCLOSESRV;
+			    desc->d_Pri = npri;
 			} else {
 			    addAsServer = 1;
 			}
 			break;
 		    case 'M':
-			nflags = COF_MODEREADER;
+			nflags |= COF_MODEREADER;
+			break;
+		    case 'R':
+			nflags |= COF_READONLY;
 			break;
 		    case 'o':	/* outgoing (post)	*/
 			npri = strtol(flags + 1, NULL, 10);
 			if ((desc = FindThreadId(THREAD_POST, host)) != NULL) {
 			    Connection *conn = desc->d_Data;
 			    conn->co_Flags &= ~COF_MAYCLOSESRV;
+			    desc->d_Pri = npri;
 			} else {
 			    addAsPoster = 1;
 			}
@@ -158,8 +175,38 @@ CheckServerConfig(time_t t, int force)
 		    }
 		    ++flags;
 		}
+		while ((option = strtok(NULL, " \t\n")) != NULL) {
+		    if (strncmp(option, "bind=", 5) == 0) {
+			bindinfo = option + 5;
+		    } else if (strncmp(option, "txbufsize=", 10) == 0) {
+			txbufsize = strtol(option + 10, NULL, 0);
+		    } else if (strncmp(option, "rxbufsize=", 10) == 0) {
+			rxbufsize = strtol(option + 10, NULL, 0);
+		    } else if (strncmp(option, "groups=", 7) == 0) {
+			groups = option + 7;
+		    } else if (strncmp(option, "localspool=", 11) == 0) {
+			localspool = option + 11;
+		    } else if (strncmp(option, "hash=", 5) == 0) {
+			char *p;
+			option += 5;
+			if ((p = strchr(option, '-')) != NULL) {
+			    hash.hf_Begin = strtol(option, NULL, 0);
+			    hash.hf_End = strtol(++p, NULL, 0);
+			    if ((p = strchr(option, '/')) != NULL)
+				hash.hf_Mod = strtol(++p, NULL, 0);
+			} else {
+			    hash.hf_Begin = hash.hf_End = strtol(option, NULL, 0);
+			    if ((p = strchr(option, '/')) != NULL)
+				hash.hf_Mod = strtol(++p, NULL, 0);
+			}
+		    } else {
+			logit(LOG_ERR, "Unknown option '%s' - ignoring", option);
+		    }
+		}
 		if (addAsServer) {
-		    if (AddServer(host, THREAD_SPOOL, port, nflags, npri) == 0) {
+		    if (AddServer(bindinfo, txbufsize, rxbufsize, host,
+				THREAD_SPOOL, port, nflags, npri,
+				groups, localspool, &hash) == 0) {
 			++NReadServers;
 			++NReadServAct;
 		    } else {
@@ -167,7 +214,9 @@ CheckServerConfig(time_t t, int force)
 		    }
 		}
 		if (addAsPoster) {
-		    if (AddServer(host, THREAD_POST, port, nflags, npri) == 0) {
+		    if (AddServer(bindinfo, txbufsize, rxbufsize, host,
+				THREAD_POST, port, nflags, npri,
+				groups, localspool, &hash) == 0) {
 			++NWriteServers;
 			++NWriteServAct;
 		    } else {
@@ -182,8 +231,60 @@ CheckServerConfig(time_t t, int force)
     }
 }
 
+/*
+ * Log some spool server stats on an hourly basis
+ */
+void
+LogServerInfo(Connection *conn, int fd)
+{
+    time_t now = time(NULL);
+
+    if (!conn->co_LastServerLog)
+	conn->co_LastServerLog = now;
+    if (!conn->co_ServerByteCount ||
+			conn->co_LastServerLog + 300 > now)
+	return;
+    syslog(LOG_INFO, "info server %s articles=%ld bytes=%ld",
+		(conn->co_Desc->d_Id ? conn->co_Desc->d_Id : "UNKNOWN"),
+		conn->co_ServerArticleCount,
+		conn->co_ServerByteCount
+    );
+    {
+	DnsRes dr;
+	dr.dr_ResultFlags = DR_SERVER_STATS;
+	dr.dr_ArtCount = conn->co_ServerArticleCount;
+	dr.dr_ByteCount = conn->co_ServerByteCount;
+	SendMsg(fd, -1, &dr);
+    }
+    conn->co_ServerByteCount = 0;
+    conn->co_ServerArticleCount = 0;
+    conn->co_LastServerLog = now;
+}
+
+GroupList *
+makeGroupList(Connection *conn, char *groups)
+{
+    GroupList *gr = NULL;
+    GroupList *grStart = NULL;
+    char *p;
+
+    if (groups == NULL)
+	return(NULL);
+    for (p = strtok(groups, ","); p != NULL; p = strtok(NULL, ",")) {
+	GroupList *g = zalloc(&conn->co_MemPool, sizeof(GroupList));
+	g->group = zallocStr(&conn->co_MemPool, p);
+	g->next = NULL;
+	if (gr != NULL)
+	    gr->next = g;
+	gr = g;
+	if (grStart == NULL)
+	    grStart = gr;
+    }
+    return(grStart);
+}
+
 int
-AddServer(const char *host, int type, int port, int flags, int pri)
+AddServer(char *bindinfo, int txbufsize, int rxbufsize, const char *host, int type, int port, int flags, int pri, char *groups, const char *localspool, HashFeed *hashFeed)
 {
     ForkDesc *desc;
     int fd;
@@ -214,13 +315,37 @@ AddServer(const char *host, int type, int port, int flags, int pri)
 	logit(LOG_ERR, "socket() call failed on host %s\n", host);
 	return(-1);
     }
+
+    if (bindinfo != NULL) {
+	struct hostent *he;
+	struct sockaddr_in lsin;
+	bzero(&lsin, sizeof(lsin));
+	if ((he = gethostbyname(bindinfo)) != NULL) {
+	    lsin.sin_addr = *(struct in_addr *)he->h_addr;
+	} else {
+	    lsin.sin_addr.s_addr = inet_addr(bindinfo);
+	    if (lsin.sin_addr.s_addr == INADDR_NONE) {
+		logit(LOG_ERR, "Unknown host for bindhost option: %s\n",
+								bindinfo);
+		return(-1);
+	    }
+	}
+	lsin.sin_family = AF_INET;
+	lsin.sin_port = 0;
+	if (bind(fd, (struct sockaddr *) &lsin, sizeof(lsin)) < 0) {
+	    logit(LOG_ERR, "failed to bind source address %s (%s)",
+						bindinfo, strerror(errno));
+	    return(-1);
+	}
+
+    }
     {
-	int on;
+	int on = 1;
 	setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (void *)&on, sizeof(on));
-	/* 
-	 * XXX set TxBufSize, RxBufSize, but different from bufsize options
-	 * for client connections.. make these bigger.
-	 */
+	if (txbufsize > 0)
+	    setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (void *)&txbufsize, sizeof(int));
+	if (rxbufsize > 0)
+	    setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (void *)&rxbufsize, sizeof(int));
     }
     fcntl(fd, F_SETFL, O_NONBLOCK);	/* asynchronous connect() */
     errno = 0;
@@ -244,7 +369,14 @@ AddServer(const char *host, int type, int port, int flags, int pri)
     conn->co_Flags |= flags;
     conn->co_Flags |= COF_INPROGRESS | COF_ININIT;
     desc->d_Count = THREAD_LIMSIZE;
+    conn->co_Auth.dr_ReaderDef = NULL;
+    conn->co_ListCacheGroups = makeGroupList(conn, groups);
+    bcopy(hashFeed, &conn->co_RequestHash, sizeof(HashFeed));
+    if (localspool != NULL)
+	desc->d_LocalSpool = zallocStr(&conn->co_MemPool, localspool);
     NNServerConnect(conn);
+    if (DebugOpt)
+	printf("Added Server (type=%d fd=%d pid=%d)\n", type, fd, (int)getpid());
     return(0);
 }
 
@@ -289,6 +421,24 @@ QueueServerRequests(void)
     }
 }
 
+int
+cbQueueReqFindThread(void *cbData, void *data)
+{
+    int result;
+    ServReq *sreq = cbData;
+    Connection *conn = data;
+
+    if (conn->co_RequestHash.hf_Mod != 0 &&
+	    !HashFeedMatch(&conn->co_RequestHash, quickhash(sreq->sr_MsgId)))
+	return(0);
+
+    if (sreq->sr_Group == NULL || conn->co_ListCacheGroups == NULL)
+	result = 1;
+    else
+	result = GroupFindWild(sreq->sr_Group, conn->co_ListCacheGroups);
+    return(result);
+}
+
 /*
  * queueServerRequest() - find best server to queue request to, return 0
  *			  on success, -1 if the request could not be queued.
@@ -301,17 +451,17 @@ int
 queueServerRequest(ServReq *sreq, int type, int maxq)
 {
     ForkDesc *desc;
-    static int Randex1;
+    static int Randex1 = -1;
 
-    desc = FindLeastUsedThread(type, maxq, 0, &Randex1, -1);
+    desc = FindLeastUsedThread(type, maxq, 0, &Randex1, -1,
+						cbQueueReqFindThread, sreq);
 
     if (desc != NULL) {
 	sreq->sr_NoPass = desc->d_Fd;
 	queueDesc(sreq, desc);
 	return(0);
     } 
-    if (DebugOpt)
-	printf("Unable to find any servers for request %s\n", sreq->sr_MsgId);
+    logit(LOG_ERR, "Unable to find any spools (or spool threads too busy) for request %s ", sreq->sr_MsgId);
     return(-1);
 }
 
@@ -335,10 +485,12 @@ queueDesc(ServReq *sreq, ForkDesc *desc)
 
     sreq->sr_Rolodex = desc->d_Fd;
 
+    if (DebugOpt)
+	printf("Request %s queued to fd %d\n", sreq->sr_MsgId, desc->d_Fd);
+
     /*
      * If server was idle, kick it.
      */
-
     if (psreq == &conn->co_SReq)
 	NNServerIdle(conn);
 }
@@ -372,7 +524,9 @@ requeueServerRequest(Connection *conn, ServReq *sreq, int type, int maxq)
 	maxq * 2,
 	conn->co_Desc->d_Pri,
 	&sreq->sr_Rolodex, 
-	sreq->sr_NoPass
+	sreq->sr_NoPass,
+	cbQueueReqFindThread,
+	sreq
     );
 
     if (desc == NULL) {
@@ -381,7 +535,9 @@ requeueServerRequest(Connection *conn, ServReq *sreq, int type, int maxq)
 	    maxq * 2, 
 	    conn->co_Desc->d_Pri + 1, 
 	    &Randex2, 
-	    -1
+	    -1,
+	    cbQueueReqFindThread,
+	    sreq
 	);
 	if (desc) {
 	    sreq->sr_NoPass = desc->d_Fd;
@@ -516,8 +672,21 @@ NNFinishSReq(Connection *conn, const char *ctl, int requeue)
 	     * can be NULL if client was terminated
 	     */
 	    if (sreq->sr_CConn) {
-		if (ctl)
-		    MBWrite(&sreq->sr_CConn->co_TMBuf, ctl, strlen(ctl));
+		if (ctl) {
+		    if (ctl[1] == '4') {
+			/*
+			 * x4x codes are POST results of some sort -
+			 * this result only appears here via NNPostResponseX.
+			 * we probably want to log this response to the
+			 * client as it may be an informative error message
+			 * unlike the average spool server transaction
+			 * which we probably don't care about.
+			 */
+		        MBLogPrintf(sreq->sr_CConn, &sreq->sr_CConn->co_TMBuf, "%s", ctl);
+		    } else {
+		        MBWrite(&sreq->sr_CConn->co_TMBuf, ctl, strlen(ctl));
+		    }
+		}
 		/*
 		 * set FCounter to 1 to prevent further recursion, which might
 		 * feed back and screw up the state machine for this
@@ -550,6 +719,9 @@ NNServerPrime1(Connection *conn)
     if (conn->co_Flags & COF_MODEREADER) {
 	MBPrintf(&conn->co_TMBuf, "mode reader\r\n");
 	NNServerPrime2(conn);
+    } else if (conn->co_Flags & COF_READONLY) {
+	MBPrintf(&conn->co_TMBuf, "mode readonly\r\n");
+	NNServerPrime4(conn);
     } else {
 	MBPrintf(&conn->co_TMBuf, "mode thisbetterfail\r\n");
 	NNServerPrime3(conn);
@@ -568,7 +740,7 @@ NNServerPrime2(Connection *conn)
     if ((len = MBReadLine(&conn->co_RMBuf, &ptr)) != 0) {
 	int code = strtol(ptr, NULL, 10);
 	if (code >= 200 && code <= 299) {
-	    logit(LOG_ERR, "mode-reader(%s) %s", conn->co_Desc->d_Id, ptr);
+	    logit(LOG_INFO, "mode-reader(%s) %s", conn->co_Desc->d_Id, ptr);
 	    NNServerStart(conn);
 	} else {
 	    logit(LOG_ERR, "mode-reader(%s) failed: %s", conn->co_Desc->d_Id, ptr);
@@ -596,6 +768,25 @@ NNServerPrime3(Connection *conn)
     }
 }
 
+void
+NNServerPrime4(Connection *conn)
+{
+    int len;
+    char *ptr;
+
+    conn->co_Func = NNServerPrime4;
+    conn->co_State = "prime4";
+
+    if ((len = MBReadLine(&conn->co_RMBuf, &ptr)) != 0) {
+	int code = strtol(ptr, NULL, 10);
+	if (code >= 200 && code <= 299)
+	    logit(LOG_INFO, "mode-readonly(%s) %s", conn->co_Desc->d_Id, ptr);
+	else
+	    logit(LOG_ERR, "mode-readonly(%s) failed: %s", conn->co_Desc->d_Id, ptr);
+	NNServerStart(conn);
+    }
+}
+
 /*
  * NNSERVERSTART() - start normal server operations.  Clean up from connect
  *		     code, make server available to clients.
@@ -645,10 +836,14 @@ NNServerIdle(Connection *conn)
 	char *buf;
 
 	if ((len = MBReadLine(&conn->co_RMBuf, &buf)) != 0) {
-	    if (len > 0)
-		logit(LOG_ERR, "Server closed connection: %s", buf);
-	    else
-		logit(LOG_ERR, "Server closed connection");
+	    if (len > 1) {
+		buf[len - 2] = 0;
+		logit(LOG_ERR, "Server closed connection: %s: %s",
+						conn->co_Desc->d_Id, buf);
+	    } else {
+		logit(LOG_ERR, "Server closed connection: %s",
+						conn->co_Desc->d_Id);
+	    }
 	    conn->co_Flags |= COF_CLOSESERVER;
 	}
     }
@@ -753,7 +948,7 @@ NNServerConnect(Connection *conn)
 
     if ((len = MBReadLine(&conn->co_RMBuf, &buf)) <= 0) {
 	if (len < 0) {
-	    logit(LOG_ERR, "connect() failed");
+	    logit(LOG_ERR, "connect(%s) failed", conn->co_Desc->d_Id);
 	    NNServerTerminate(conn);
 	}
 	return;
@@ -801,7 +996,7 @@ NNServerConnect(Connection *conn)
  */
 
 void
-NNArticleRetrieveByMessageId(Connection *conn, const char *msgid)
+NNArticleRetrieveByMessageId(Connection *conn, const char *msgid, int grouphint)
 {   
     FILE *cache = NULL;
 
@@ -809,7 +1004,7 @@ NNArticleRetrieveByMessageId(Connection *conn, const char *msgid)
      * (a) retrieve from cache if caching is enabled
      */
 
-    if (DiabloReaderCacheMode) {
+    if (DOpts.ReaderCacheMode) {
 	int valid;
 	int size;
 	int cfd;
@@ -821,11 +1016,12 @@ NNArticleRetrieveByMessageId(Connection *conn, const char *msgid)
 	     * good cache
 	     */
 	    const char *map;
-	    printf("good cache\n");
+	    if (DebugOpt)
+		printf("good cache\n");
 	    if ((map = xmap(NULL, size, PROT_READ, MAP_SHARED, cfd, 0)) != NULL) {
 		xadvise(map, ((size < 65536) ? size : 65536), XADV_WILLNEED);
 		if (conn->co_ArtMode != COM_BODYNOSTAT) {
-		    MBPrintf(&conn->co_TMBuf, "%03d %d %s %s\r\n", 
+		    MBLogPrintf(conn, &conn->co_TMBuf, "%03d %d %s %s\r\n", 
 			GoodRC(conn),
 			((conn->co_ArtMode==COM_ARTICLEWVF)?conn->co_ArtNo:0),
 			msgid,
@@ -839,9 +1035,11 @@ NNArticleRetrieveByMessageId(Connection *conn, const char *msgid)
 		xunmap((void *)map, size);
 	    } else {
 		if (conn->co_ArtMode == COM_BODYNOSTAT)
-		    MBPrintf(&conn->co_TMBuf, "(article not available)\r\n.\r\n");
+		    MBLogPrintf(conn, &conn->co_TMBuf, "(article not available)\r\n.\r\n");
+		else if (conn->co_RequestFlags == ARTFETCH_ARTNO)
+		    MBLogPrintf(conn, &conn->co_TMBuf, "423 No such article number in this group\r\n");
 		else
-		    MBPrintf(&conn->co_TMBuf, "430 No such article\r\n");
+		    MBLogPrintf(conn, &conn->co_TMBuf, "430 No such article\r\n");
 	    }
 	    close(cfd);
 	    NNCommand(conn);
@@ -850,7 +1048,8 @@ NNArticleRetrieveByMessageId(Connection *conn, const char *msgid)
 	    /*
 	     * not in cache, cache file created if cfd >= 0
 	     */
-	    printf("bad cache\n");
+	    if (DebugOpt)
+		printf("bad cache\n");
 	    if (cfd >= 0) {
 		cache = fdopen(cfd, "w");
 	    }
@@ -859,11 +1058,15 @@ NNArticleRetrieveByMessageId(Connection *conn, const char *msgid)
 	    /*
 	     * negatively cached
 	     */
-	    printf("neg cache\n");
+	    if (DebugOpt)
+		printf("neg cache\n");
+
 	    if (conn->co_ArtMode == COM_BODYNOSTAT)
-		MBPrintf(&conn->co_TMBuf, "(article not available)\r\n.\r\n");
+		MBLogPrintf(conn, &conn->co_TMBuf, "(article not available)\r\n.\r\n");
+	    else if (conn->co_RequestFlags == ARTFETCH_ARTNO)
+		MBLogPrintf(conn, &conn->co_TMBuf, "423 No such article number in this group\r\n");
 	    else
-		MBPrintf(&conn->co_TMBuf, "430 No such article\r\n");
+		MBLogPrintf(conn, &conn->co_TMBuf, "430 No such article\r\n");
 	    NNCommand(conn);
 	    return;
 	}
@@ -872,7 +1075,8 @@ NNArticleRetrieveByMessageId(Connection *conn, const char *msgid)
     /*
      * (b)(c)(d)
      */
-    NNServerRequest(conn, conn->co_GroupName, msgid, cache, SREQ_RETRIEVE);
+    NNServerRequest(conn, grouphint ? conn->co_GroupName : NULL, msgid,
+							cache, SREQ_RETRIEVE);
     NNWaitThread(conn);
 }
 
