@@ -9,6 +9,8 @@ typedef struct ExpireCtl {
     int                 ex_Slot;
     int                 ex_MaxSize;
     int                 ex_ExpireLines;
+    int                 ex_InitArts;
+    int                 ex_MinArts;
     int                 ex_MaxArts;
     double              ex_ExpireDays;
     double              ex_CrossFactor;
