@@ -80,6 +80,7 @@ Prototype int KPDBWrite(KPDB *kpdb, const char *key, const char *tok, const char
 Prototype void KPDBLock(KPDB *kpdb, const char *rec);
 Prototype void KPDBUnlock(KPDB *kpdb, const char *rec);
 Prototype int KPDBDelete(KPDB *kpdb, const char *key);
+Prototype int KPDBAppendCount(KPDB *kpdb);
 
 void KPDBReSize(KPDB *kpdb);
 int KPDBValidate(KPDB *kpdb);
@@ -1150,4 +1151,9 @@ KPDBSort(KPDB *kpdb)
     return(0);
 }
 
+int
+KPDBAppendCount(KPDB *kpdb)
+{
+    return strtol(kpdb->kp_MapBase + A_OFF, NULL, 16);
+}
 
