@@ -10,16 +10,18 @@
 #include "defs.h"
 
 Prototype void rsignal(int sigNo, void (*func)(int sigNo));
+Prototype void nrsignal(int sigNo, void (*func)(int sigNo));
 
 void 
-rsignal(int sigNo, void (*func)(int sigNo))
+internal_signal(int sigNo, void (*func)(int sigNo), int restart)
 {
     struct sigaction sa;
 
     bzero(&sa, sizeof(sa));
 
 #ifndef SUNOS
-    sa.sa_flags = SA_RESTART;	/* do not set SA_RESETHAND */
+    if (restart)
+	sa.sa_flags = SA_RESTART;	/* do not set SA_RESETHAND */
 #endif
     sa.sa_handler = func;
     sigaddset(&sa.sa_mask, sigNo);
@@ -30,3 +32,15 @@ rsignal(int sigNo, void (*func)(int sigNo))
     }
 }
 
+void 
+rsignal(int sigNo, void (*func)(int sigNo))
+{
+    internal_signal(sigNo, func, 1);
+}
+
+void 
+nrsignal(int sigNo, void (*func)(int sigNo))
+{
+    internal_signal(sigNo, func, 0);
+}
+
