@@ -6,9 +6,6 @@
  *    the COPYRIGHT file in the base directory of this distribution
  *    for specific rights granted.
  *
- * dexpireover [-a] [-v[N]] [-w grpwildcard] [-f dactive.kp] [-NB] [-n] [-O[n]]
- *             [-s] [-R] [-e] [-o] [-pN]
- *
  * 	In this incarnation, dexpireover cleans up overview information as
  *	specified in the 'x' fields in dexpire.ctl (see the sample dexpire.ctl)
  *
@@ -27,6 +24,17 @@
  *	Specifying the -pN option will fork N dexpireover processes and
  *	perform the expiration process in parallel. This is useful to speed
  *	up ExpireBySpool where dhistory lookups take a long time.
+ *
+ *	LOCKING INFO
+ *
+ *	The data file will have an advisory lock at offset 4 from dreaderd
+ *	when it has the file open (can't rewrite data)
+ *	The info file will have an advisory lock at offset 4 from dreaderd
+ *	when it has the file open (can't resize)
+ *
+ *	od_HFd = data. file
+ *	ov_OFd = over. file
+ *
  */
 
 #include <dreaderd/defs.h>
@@ -38,9 +46,10 @@ typedef struct Group {
     int		gr_EndNo;
     int		gr_CTS;
     int		gr_LMTS;
+    int		gr_Iter;
     char	*gr_GroupName;
     char	*gr_Flags;
-    hash_t	gr_Hv;
+    char	*gr_Hash;
     int		gr_UpdateFlag;
 } Group;
 
@@ -84,17 +93,20 @@ typedef struct bucket_t {
 KPDB  *KDBActive;
 Group *GHash[GHSIZE];
 
+void ScanDirectories(void);
+void scanDirectory(const char *dirpath, char *dirname, int *level);
 void DeleteJunkFile(const char *dirPath, const char *name);
-void ProcessOverviewFile(const char *dirPath, const char *name);
+void ProcessOverviewFile(const char *dirPath, const char *name, int type);
+int getFirstArtAge(Group *group, int fd, OverHead *oh);
 char *allocTmpCopy(const char *buf, int bufLen);
-Group *EnterGroup(const char *groupName, int begNo, int endNo, int lmts, int cts, const char *flags);
-Group *FindGroupByHash(hash_t *hv);
+Group *EnterGroup(const char *groupName, int begNo, int endNo, int lmts, int cts, int iter, const char *flags);
+Group *FindGroupByHash(char *Hash, int iter);
 int SetField(char **pptr, const char *str);
-void ExpireByDays(Group *group, int fd, OverHead *oh);
+void ExpireByDays(Group *group, int fd, OverHead *oh, int expireSecs);
 void ExpireBySpool(Group *group, int fd, OverHead *oh);
-void ExpireFromFile(Group *group, int fd, OverHead *oh);
-void RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int h2);
-void rewriteDataFile(Group *group, ReplaceFile **prf, const char *cacheBase, int cacheSize, const OverArt *oa, OverArt *ob, const char *dirPath, int h1, int h2);
+void ExpireFromFile(Group *group, int fd, OverHead *oh, int expireSecs);
+void RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath);
+void rewriteDataFile(Group *group, ReplaceFile **prf, const char *cacheBase, int cacheSize, const OverArt *oa, OverArt *ob, const char *dirPath);
 void ResizeGroup(Group *group, int fd, OverHead *oh, int maxArts);
 int nearestPower(int n);
 void ReadDExpOverList(void);
@@ -113,21 +125,67 @@ int VerboseOpt = -1;
 int ResizeOpt = -1;
 int ForReal = 1;
 int OldGroups = 0;
+int UseExpireByDays = 0;
 int UseExpireBySpool = 0;
 int UseExpireFromFile = 0;
 int ParallelCount = 0;
+int FactorExpire = 0;
+int LockWaitTime= 1;
+int MustExit = 0;
 char *Wild;
-
 bucket_t *dexpover_msgid_hash;
+int ParallelIdx = 0;
+int ParallelPid[MAX_PAR_COUNT];
+
+void
+sigInt(int sigNo)
+{
+    printf("Exit signal caught - exiting\n");
+    ++MustExit;
+    if (MustExit > 3)
+	exit(1);
+}
+
+/*
+ * We use alarms to cancel a lock wait, so just ignore the alarm
+ */
+void
+sigAlarm(int sigNo)
+{
+ ;;
+}
+
+void Usage(char *progname)
+{
+    fprintf(stderr, "Expire the reader header database\n");
+    fprintf(stderr, "dexpireover [-a] [-e] [-f active] [-l#] [-NB] [-n] [-O#] [-o] [-p#] [-R] [-s] [-U] [-v#] [-w wildmat] [-x] [-y] [-C diablo.config] [-d[n]] [-V]\n");
+    fprintf(stderr, "\t-a\t\tDo a standard header expire run (-NB -U -s -y)\n");
+    fprintf(stderr, "\t-e\t\tExpire by checking a local spool\n");
+    fprintf(stderr, "\t-f file\t\tSpecify the name of the active file\n");
+    fprintf(stderr, "\t-l#\t\tWait N seconds for lock files\n");
+    fprintf(stderr, "\t-NB\t\tUpdate the begining article number (NB) in active\n");
+    fprintf(stderr, "\t-n\t\tDon't actually make any changes (dry run)\n");
+    fprintf(stderr, "\t-O#\t\tRemove groups not used within # days\n");
+    fprintf(stderr, "\t-o\t\tExpire from file of msgid hashes\n");
+    fprintf(stderr, "\t-p#\t\tRun # parallel expire runs (speed up)\n");
+    fprintf(stderr, "\t-R\t\tRewrite header data files\n");
+    fprintf(stderr, "\t-s\t\tResize group indexes\n");
+    fprintf(stderr, "\t-U\t\tAdd CTS (group create time) in active if not present\n");
+    fprintf(stderr, "\t-v[#]\t\tVerbose mode\n");
+    fprintf(stderr, "\t-w wildmat\tSpecify a wildmat of groups to expire\n");
+    fprintf(stderr, "\t-x\t\tUse factoring for expire (see man page)\n");
+    fprintf(stderr, "\t-y\t\tActually expire headers based on dexpire.ctl\n");
+    fprintf(stderr, "\t-C file\tspecify diablo.config to use\n");
+    fprintf(stderr, "\t-d[n]\tset debug [with optional level]\n");
+    fprintf(stderr, "\t-V\tprint version and exit\n");
+    exit(1);
+}
 
 int
 main(int ac, char **av)
 {
     int i;
-    int freeSpaceTarget = 0;
     char *dbfile = NULL;
-    int ParallelIdx = 0;
-    int ParallelPid[MAX_PAR_COUNT];
 
     LoadDiabloConfig(ac, av);
 
@@ -135,44 +193,47 @@ main(int ac, char **av)
 	char *ptr = av[i];
 	if (*ptr != '-') {
 	    fprintf(stderr, "Unexpected argument: %s\n", ptr);
-	    exit(1);
+	    Usage(av[0]);
 	}
 	ptr += 2;
 	switch(ptr[-1]) {
-	case 'C':           /* parsed by LoadDiabloConfig */
-	    if (*ptr == 0)
-		++i;
-	    break;
 	case 'a':
+	    UseExpireByDays = 1;
 	    UpdateBegArtNoOpt = 1;
 	    UpdateCTSOpt = 1;
 	    ResizeOpt = 1;
-	    if (freeSpaceTarget == 0)
-		freeSpaceTarget = 512;
 	    break;
-	case 's':
-	    ResizeOpt = 1;
+	case 'e':
+	    UseExpireBySpool = 1;
 	    break;
-	case 'v':
-	    VerboseOpt = (*ptr) ? strtol(ptr, NULL, 0) : 1;
+	case 'f':
+	    dbfile = (*ptr) ? ptr : av[++i];
 	    break;
-	case 'r':
-	    ptr = (*ptr) ? ptr : av[++i];
-	    freeSpaceTarget = strtol(ptr, NULL, 0);
+	case 'l':
+	    LockWaitTime = strtol((*ptr) ? ptr : av[++i], NULL, 0);
 	    break;
-	case 'w':
-	    Wild = (*ptr) ? ptr : av[++i];
+	case 'N':
+	    while (*ptr) {
+		switch(*ptr) {
+		case 'B':
+		    UpdateBegArtNoOpt = 1;
+		    break;
+		default:
+		    break;
+		}
+		++ptr;
+	    }
 	    break;
 	case 'n':
 	    ForReal = 0;
 	    if (VerboseOpt < 0)
 		VerboseOpt = 1;
 	    break;
-	case 'f':
-	    dbfile = (*ptr) ? ptr : av[++i];
-	    break;
-	case 'e':
-	    UseExpireBySpool = 1;
+	case 'O':
+	    if (*ptr)
+		OldGroups = strtol(ptr, NULL, 0);
+	    else
+		OldGroups = 30 * 3;	/* 3 months by default */
 	    break;
 	case 'o':
 	    UseExpireFromFile = 1;
@@ -189,33 +250,44 @@ main(int ac, char **av)
 	case 'R':
 	    RewriteDataOpt = 1;
 	    break;
-	case 'O':
-	    if (*ptr)
-		OldGroups = strtol(ptr, NULL, 0);
-	    else
-		OldGroups = 30 * 3;	/* 3 months by default */
+	case 's':
+	    ResizeOpt = 1;
 	    break;
 	case 'U':
 	    UpdateCTSOpt = 1;
 	    break;
-	case 'N':
-	    while (*ptr) {
-		switch(*ptr) {
-		case 'B':
-		    UpdateBegArtNoOpt = 1;
-		    break;
-		default:
-		    break;
-		}
-		++ptr;
-	    }
+	case 'v':
+	    VerboseOpt = (*ptr) ? strtol(ptr, NULL, 0) : 1;
+	    break;
+	case 'w':
+	    Wild = (*ptr) ? ptr : av[++i];
+	    break;
+	case 'x':
+	    FactorExpire = 1;
+	    break;
+	case 'y':
+	    UseExpireByDays = (*ptr) ? strtol(ptr, NULL, 0) : 1;
+	    break;
+	/* Common options */
+	case 'C':           /* parsed by LoadDiabloConfig */
+	    if (*ptr == 0)
+		++i;
+	    break;
+	case 'd':
+	    DebugOpt = (*ptr) ? strtol(ptr, NULL, 0) : 1;
+	    break;
+	case 'V':
+	    PrintVersion();
 	    break;
 	default:
 	    fprintf(stderr, "Unknown option: %s\n", ptr - 2);
-	    exit(1);
+	    Usage(av[0]);
 	}
     }
 
+    if (!UseExpireByDays && ResizeOpt <= 0 && !RewriteDataOpt &&
+			!UseExpireFromFile && !UseExpireBySpool && !OldGroups)
+	Usage(av[0]);
     /*
      * Read in the list of expired msgid hashes, if we are using it
      */
@@ -266,6 +338,8 @@ main(int ac, char **av)
      * Open active file database
      */
 
+    if (VerboseOpt)
+	printf("Loading active file\n");
     if (dbfile) {
 	KDBActive = KPDBOpen(dbfile, O_RDWR);
     } else {
@@ -277,7 +351,7 @@ main(int ac, char **av)
     }
     if (OldGroups && Wild == NULL) {
 	fprintf(stderr, "group wildcard must be specified if -O option used\n");
-	exit(1);
+	Usage(av[0]);
     }
 
     LoadExpireCtl(1);
@@ -293,6 +367,8 @@ main(int ac, char **av)
      * scan dactive.kp
      */
 
+    if (VerboseOpt)
+	printf("Hashing newsgroups\n");
     {
 	int recLen;
 	int recOff;
@@ -311,6 +387,7 @@ main(int ac, char **av)
 	    int endNo = strtol(KPDBGetField(rec, recLen, "NE", NULL, "-1"), NULL, 10);
 	    int lmts = (int)strtoul(KPDBGetField(rec, recLen, "LMTS", NULL, "0"), NULL, 16);
 	    int cts = (int)strtoul(KPDBGetField(rec, recLen, "CTS", NULL, "0"), NULL, 16);
+	    int iter = (int)strtoul(KPDBGetField(rec, recLen, "ITER", NULL, "0"), NULL, 16);
 	    Group *grp;
 
 	    if (cts == 0)	/* enter non-zero cts only if group has no CTS field */
@@ -338,74 +415,25 @@ main(int ac, char **av)
 		endNo,
 		lmts,
 		cts,
+		iter,
 		flags
 	    );
 	    grp->gr_State &= ~(GRF_NEW|GRF_MODIFIED);
 	}
     }
 
+    rsignal(SIGINT, sigInt);
+    rsignal(SIGHUP, sigInt);
+    rsignal(SIGTERM, sigInt);
+    rsignal(SIGALRM, SIG_IGN);
+
     /*
-     * scan /news/spool/group/
+     * Scan /news/spool/group/ and do the actual expire, over resize
+     * and data rewrite
+     *
      */
 
-    {
-	DIR *dir;
-
-	if ((dir = opendir(PatExpand(GroupHomePat))) != NULL) {
-	    den_t *den;
-	    char *tmp = malloc(strlen(PatExpand(GroupHomePat)) + 32);
-
-	    while ((den = readdir(dir)) != NULL) {
-		if (strlen(den->d_name) == 2 && 
-		    isalnum(den->d_name[0]) &&
-		    isalnum(den->d_name[1]) &&
-		    /* We explicitly use the first char, because overview
-		     * sizes appear to be not evenly distributed wrt second
-		     * char.
-		     */
-		    (ParallelCount ? ((hexCharToInt(den->d_name[0]) % ParallelCount) ==
-				      ParallelIdx) : 1)
-		) {
-		    DIR *dir2;
-
-		    sprintf(tmp, "%s/%s", PatExpand(GroupHomePat), den->d_name);
-		    if ((dir2 = opendir(tmp)) != NULL) {
-			den_t *den2;
-
-			while ((den2 = readdir(dir2)) != NULL) {
-			    if (strncmp(den2->d_name, "over.", 5) == 0)
-				ProcessOverviewFile(tmp, den2->d_name);
-			}
-			rewinddir(dir2);
-			while ((den2 = readdir(dir2)) != NULL) {
-			    /*
-			     * delete junk files from previously interrupted
-			     * dexpireover -R
-			     */
-			    if (strncmp(den2->d_name, ".data.", 6) == 0)
-				DeleteJunkFile(tmp, den2->d_name);
-
-			    /*
-			     * process data. files
-			     */
-			    if (strncmp(den2->d_name, "data.", 5) == 0)
-				ProcessOverviewFile(tmp, den2->d_name);
-			}
-			closedir(dir2);
-		    }
-		}
-	    }
-	    closedir(dir);
-	    free(tmp);
-	}
-    }
-
-    printf("Scanned %d groups, %d were bad, %d/%d resized\n",
-	GoodGroups + BadGroups, 
-	BadGroups, 
-	ResizedGroups, 
-	NoResizedGroups + ResizedGroups
-    );
+    ScanDirectories();
 
     /*
      * Writeback active file
@@ -495,13 +523,143 @@ main(int ac, char **av)
     return(0);
 }
 
+/*
+ * Scan /news/spool/group/ and do the actual expire, over resize
+ * and data rewrite
+ *
+ * This works by scanning all files in the directories, mapping the
+ * filename (hash) to a newsgroup name and performing the following,
+ * depending on file type:
+ *
+ * over.* :
+ * *.o.* :
+ *		1) resize over (if requested)
+ *		2) expire articles (by days, spool or file)
+ *		3) rewrite data (if requested)
+ *
+ * data.*:
+ * *.d.*:
+ *		1) if the range of articles for the file falls below NB,
+ *		   then delete the file
+ *
+ * If a newsgroup for a file cannot be found, the file is removed
+ */
+
+void
+ScanDirectories(void)
+{
+    DIR *dir;
+    int level = 0;
+
+    if (VerboseOpt)
+	printf("Scanning group directories\n");
+
+    chdir(PatExpand(GroupHomePat));
+    if ((dir = opendir(".")) != NULL) {
+	den_t *den;
+	struct stat st;
+
+	while ((den = readdir(dir)) != NULL) {
+	    if (isalnum((int)den->d_name[0]) &&
+	    	stat(den->d_name, &st) == 0 && S_ISDIR(st.st_mode) &&
+		/*
+		 * We explicitly use the first char, because overview
+		 * sizes appear to be not evenly distributed wrt second
+		 * char.
+		 */
+		(ParallelCount ? ((GFIndex(den->d_name[0]) % ParallelCount) ==
+				      ParallelIdx) : 1)
+	    )
+		scanDirectory(PatExpand(GroupHomePat), den->d_name, &level);
+	}
+	closedir(dir);
+    }
+
+    printf("Scanned %d files, %d were bad, %d/%d indexes resized successfully\n",
+	GoodGroups + BadGroups, 
+	BadGroups, 
+	ResizedGroups, 
+	NoResizedGroups + ResizedGroups
+    );
+
+}
+
+void
+scanDirectory(const char *dirpath, char *dirname, int *level)
+{
+    DIR *dir2;
+    char path[PATH_MAX];
+    char origpath[PATH_MAX];
+
+    sprintf(path, "%s/%s", dirpath, dirname);
+
+    if (getcwd(origpath, sizeof(origpath)) == NULL) {
+	printf("Unable to getcwd(%s): %s\n", dirname, strerror(errno));
+	return;
+    }
+    if (chdir(dirname) != 0) {
+	printf("Unable to chdir(%s)\n", dirname);
+	return;
+    }
+
+    if ((dir2 = opendir(".")) != NULL) {
+	den_t *den2;
+	char tbuf[24];
+	int tint;
+	struct stat st;
+
+	while ((den2 = readdir(dir2)) != NULL) {
+	    if (strcmp(den2->d_name, ".") == 0 || strcmp(den2->d_name, "..") == 0)
+		continue;
+	    if (stat(den2->d_name, &st) == 0 && S_ISDIR(st.st_mode)) {
+		if (*level <= 2)
+		    scanDirectory(path, den2->d_name, level);
+	    } else if (strncmp(den2->d_name, "over.", 5) == 0 ||
+		strncmp(den2->d_name, "o.", 2) == 0 ||
+		strstr(den2->d_name, ".o.") != NULL)
+		ProcessOverviewFile(path, den2->d_name, 1);
+	    if (MustExit)
+		exit(1);
+	}
+	rewinddir(dir2);
+	while ((den2 = readdir(dir2)) != NULL) {
+	    /*
+	     * delete junk files from previously interrupted
+	     * dexpireover -R
+	     */
+	    if (strncmp(den2->d_name, ".data.", 6) == 0 ||
+		strncmp(den2->d_name, ".d.", 3) == 0 ||
+		sscanf(den2->d_name, ".%[^.]s.%d.d.", tbuf, &tint) == 2) {
+		DeleteJunkFile(path, den2->d_name);
+		continue;
+	    }
+
+	    /*
+	     * process data. files
+	     */
+	    if (strncmp(den2->d_name, "data.", 5) == 0 ||
+		strstr(den2->d_name, ".d.") != NULL)
+		ProcessOverviewFile(path, den2->d_name, 2);
+	    if (MustExit)
+		exit(1);
+	}
+	closedir(dir2);
+    }
+    if (chdir(origpath) != 0) {
+	printf("FATAL: Unable to return with chdir(%s)\n", dirname);
+	exit(1);
+    }
+}
+
 void
 DeleteJunkFile(const char *dirPath, const char *name)
 {
-    if (ForReal) {
-	char path[1024];
+    char path[PATH_MAX];
 
-	snprintf(path, sizeof(path), "%s/%s", dirPath, name);
+    snprintf(path, sizeof(path), "%s/%s", dirPath, name);
+    if (VerboseOpt)
+	printf("Removing old temp file: %s\n", path);
+    if (ForReal) {
 	remove(path);
     }
 }
@@ -520,32 +678,29 @@ DeleteJunkFile(const char *dirPath, const char *name)
  */
 
 void
-ProcessOverviewFile(const char *dirPath, const char *name)
+ProcessOverviewFile(const char *dirPath, const char *name, int type)
 {
-    long h1;
-    long h2;
-    long artBase;
-    int type = 0;
+    long artBase = -1;
     Group *group;
-    hash_t hv;
-    char path[1024];
+    char path[PATH_MAX];
+    char Hash[PATH_MAX];
+    int iter = 0;
 
     snprintf(path, sizeof(path), "%s/%s", dirPath, name);
 
-    if (strncmp(name, "over.", 5) == 0)
-	type = 1;
-    if (strncmp(name, "data.", 5) == 0)
-	type = 2;
+    if (DebugOpt > 0)
+	printf("ProcessOverviewFile(%s)\n", path);
 
-    if (type == 0)
+    bzero(Hash, sizeof(Hash));
+
+    if (ExtractGroupHashInfo(name, Hash, &artBase, &iter) == HASHGRP_NONE)
 	return;
 
-    if (sscanf(name + 5, "%ld.%lx.%lx", &artBase, &h1, &h2) != 3)
-	return;
-    hv.h1 = h1;
-    hv.h2 = h2;
+    if (DebugOpt > 2)
+	printf("File: %s  type=%d  artBase=%ld  iter=%d  Hash=%s\n",
+						name, type, artBase, iter, Hash);
 
-    if ((group = FindGroupByHash(&hv)) == NULL) {
+    if ((group = FindGroupByHash(Hash, iter)) == NULL) {
 	/*
 	 * If we gave a wildcard, we can't remove stale groups because
 	 * we do not have a full group list.
@@ -553,13 +708,17 @@ ProcessOverviewFile(const char *dirPath, const char *name)
 
 	if (Wild == NULL) {
 	    ++BadGroups;
-	    printf("Group not found, removing %s\n", path);
-	    remove(path);
+	    printf("Group for over file not found, removing %s\n", path);
+	    if (ForReal)
+		remove(path);
 	}
 	return;
     }
     ++GoodGroups;
 
+    if (DebugOpt > 2)
+	printf("Maps to group: %s\n", group->gr_GroupName);
+
     if (type == 1) {
 	/*
 	 * over. file	(fixed length file)
@@ -568,15 +727,38 @@ ProcessOverviewFile(const char *dirPath, const char *name)
 	OverHead oh;
 
 	if ((fd = open(path, O_RDWR)) >= 0) {
-	    if (read(fd, &oh, sizeof(oh)) == sizeof(oh) &&
-		oh.oh_Version == OH_VERSION &&
-		oh.oh_ByteOrder == OH_BYTEORDER
-	    ) {
+	    int r;
+	    r = (read(fd, &oh, sizeof(oh)) == sizeof(oh));
+	    if (oh.oh_Version > OH_VERSION ||
+		oh.oh_ByteOrder != OH_BYTEORDER)
+		r = 0;
+	    if (r && oh.oh_Version > 1 &&
+				strcmp(oh.oh_Gname, group->gr_GroupName) != 0)
+		r = 0;
+	    if (r) {
+		int expireSecs;
+		OverExpire save;
+		int recLen;
+		const char *rec = KPDBReadRecord(KDBActive, group->gr_GroupName,
+                                                        KP_LOCK, &recLen);
+
+		/*
+		 * Refresh the EndNo
+		 */
+		group->gr_EndNo = strtol(KPDBGetField(rec, recLen, "NE",
+							NULL, "-1"), NULL, 10);
+		KPDBUnlock(KDBActive, rec);
+
 		if (group->gr_EndNo - group->gr_StartNo >= oh.oh_MaxArts) {
 		    group->gr_StartNo = group->gr_EndNo - oh.oh_MaxArts + 1;
 		    group->gr_State |= GRF_EDITEDBEG | GRF_MODIFIED;
 		}
 
+		/*
+		 * Figure out expire values
+		 */
+		expireSecs = GetOverExpire(group->gr_GroupName, &save);
+
 		/*
 		 * Force regeneration of over. file if RewriteDataOpt, else
 		 * only regenerate if ResizeOpt and a size differential.  Valid
@@ -587,8 +769,42 @@ ProcessOverviewFile(const char *dirPath, const char *name)
 		if (ResizeOpt > 0 || RewriteDataOpt > 0) {
 		    int numArts = group->gr_EndNo - group->gr_StartNo + 1;
 		    int maxArts = oh.oh_MaxArts;
+		    int artAge;
 
-		    if (numArts < maxArts / 2)
+		    /*
+		     * Find out the age of the first article in the
+		     * group and calculate an age factor based on this
+		     * and the expire value in dexpire.ctl.
+		     *
+		     * We don't let the age factor get too large
+		     * if the age of the first articles is less than an
+		     * hour because this can lead to some bloated files.
+		     * We also adjust the age factor if there are a
+		     * large number of articles in the group and the
+		     * age factor is low as we don't need a huge
+		     * number of spare slots.
+		     */
+		    artAge = getFirstArtAge(group, fd, &oh);
+		    if (FactorExpire && artAge > 0 && expireSecs > 0) {
+			float ageFactor;
+
+			if (artAge <= 0)
+			    artAge = 1;
+			ageFactor = (expireSecs * 1.0) / (artAge * 1.0);
+			if (ageFactor < 1.5)
+			    ageFactor = 1.5;
+			if (numArts > 20000 && ageFactor == 1.5)
+			    ageFactor = 1.3;
+			if (artAge < 3600 && ageFactor > 50.0)
+			    ageFactor = 50.0;
+			if (numArts > 500 && ageFactor > 20.0)
+			    ageFactor = 20.0;
+			if (numArts > 5000 && ageFactor > 5.0)
+			    ageFactor = 5.0;
+			if (ageFactor > 400.0)
+			    ageFactor = 400.0;
+			numArts = ageFactor * numArts;
+		    } else if (numArts < maxArts / 2)
 			numArts = maxArts - nearestPower(maxArts) / 3;
 		    else if (numArts > maxArts * 2 / 3)
 			numArts = maxArts + nearestPower(maxArts) / 3;
@@ -601,31 +817,50 @@ ProcessOverviewFile(const char *dirPath, const char *name)
 		     * we save a considerable amount of space.
 		     */
 
-		    if (numArts < 24)
-			numArts = 24;
+		    if (save.oe_MaxArts > 0 && numArts > save.oe_MaxArts)
+			numArts = save.oe_MaxArts;
+		    if (save.oe_MinArts >= 0 && numArts < save.oe_MinArts)
+			numArts = save.oe_MinArts;
 
-		    if (numArts != maxArts || RewriteDataOpt > 0)
-			ResizeGroup(group, fd, &oh, numArts);
+		    {
+			int needResize = RewriteDataOpt;
+			int d = abs(maxArts - numArts);
+			/*
+			 * We only rewrite if there has been a radical
+			 * change in the number of articles needed.
+			 */
+			if (numArts < 100 && d > 10)
+				needResize = 1;
+			if (numArts >= 100 && numArts < 1000 && d > 20)
+				needResize = 1;
+			if (numArts >= 1000 && d > 200)
+				needResize = 1;
+			if (needResize)
+			    ResizeGroup(group, fd, &oh, numArts);
+		    }
 		}
 
 		if (UseExpireBySpool)
 		    ExpireBySpool(group, fd, &oh);
 		else if (UseExpireFromFile)
-		    ExpireFromFile(group, fd, &oh);
-		else
-		    ExpireByDays(group, fd, &oh);
+		    ExpireFromFile(group, fd, &oh, expireSecs);
+		else if (UseExpireByDays)
+		    ExpireByDays(group, fd, &oh, expireSecs);
 
 		/*
 		 * Rewrite data files associated with over. file if -R.
 		 */
 		if (RewriteDataOpt > 0)
-		    RewriteData(group, fd, &oh, dirPath, h1, h2);
+		    RewriteData(group, fd, &oh, dirPath);
 
 	    } else {
 		printf("group %s, file \"%s\" bad file header\n",
 		    group->gr_GroupName,
 		    path
 		);
+		if (oh.oh_Version > OH_VERSION)
+		    printf("   expected version %d, got version %d\n",
+						OH_VERSION, oh.oh_Version);
 	    }
 	    close(fd);
 	}
@@ -635,6 +870,16 @@ ProcessOverviewFile(const char *dirPath, const char *name)
 	 * part of conditional is a fudge to make 100% sure we do not
 	 * delete a brand new data file.
 	 */
+	int recLen;
+	const char *rec = KPDBReadRecord(KDBActive, group->gr_GroupName,
+                                                        KP_LOCK, &recLen);
+	/*
+	 * Refresh the EndNo
+	 */
+	group->gr_EndNo = strtol(KPDBGetField(rec, recLen, "NE",
+							NULL, "-1"), NULL, 10);
+	KPDBUnlock(KDBActive, rec);
+
 	if (artBase + OD_HARTS <= group->gr_StartNo ||
 	    artBase >= group->gr_EndNo + OD_HARTS
 	) {
@@ -642,11 +887,54 @@ ProcessOverviewFile(const char *dirPath, const char *name)
 		group->gr_GroupName,
 		path
 	    );
-	    remove(path);
+	    printf("  artBase=%ld  StartNo=%d  EndNo=%d\n",
+			artBase, group->gr_StartNo, group->gr_EndNo);
+	    if (ForReal)
+		remove(path);
 	} 
     }
 }
 
+int
+getFirstArtAge(Group *group, int fd, OverHead *oh)
+{
+    const OverArt *oaBase;
+    struct stat st;
+    int n;
+    time_t t = time(NULL);
+    int dt = 0;
+
+    if (fstat(fd, &st) != 0)
+	return(-1);
+
+    /*
+     * Calculate number of overview records
+     */
+
+    n = (st.st_size - oh->oh_HeadSize) / sizeof(OverArt);
+
+    oaBase = xmap(NULL, n * sizeof(OverArt), PROT_READ, MAP_SHARED, fd,
+			oh->oh_HeadSize);
+    if (oaBase == NULL) {
+	fprintf(stderr, "Unable to xmap over.* file for group %s\n",
+			group->gr_GroupName);
+	return(-1);
+    }
+
+    {
+        int i = (group->gr_StartNo & 0x7FFFFFFF) % oh->oh_MaxArts;
+	const OverArt *oa = &oaBase[i];
+
+	/*
+	 * Get the age of the first valid article
+	 */
+	if (oa && i <= n)
+	    dt = (int)(t - oa->oa_TimeRcvd);
+    }
+    xunmap((void *)oaBase, n * sizeof(OverArt));
+    return(dt);
+}
+
 char *
 allocTmpCopy(const char *buf, int bufLen)
 {
@@ -665,9 +953,9 @@ allocTmpCopy(const char *buf, int bufLen)
 }
 
 Group *
-EnterGroup(const char *groupName, int begNo, int endNo, int lmts, int cts, const char *flags)
+EnterGroup(const char *groupName, int begNo, int endNo, int lmts, int cts, int iter, const char *flags)
 {
-    hash_t hv = hhash(groupName);
+    hash_t hv = hhash(GFHash(groupName, &DOpts.ReaderGroupHashMethod));
     Group **pgroup = &GHash[hv.h1 & GHMASK];
     Group *group;
 
@@ -680,7 +968,8 @@ EnterGroup(const char *groupName, int begNo, int endNo, int lmts, int cts, const
 	*pgroup = group = calloc(sizeof(Group) + strlen(groupName) + 1, 1);
 	group->gr_State = GRF_NEW;
 	group->gr_GroupName = (char *)(group + 1);
-	group->gr_Hv = hv;
+	group->gr_Hash = strdup(GFHash(groupName, &DOpts.ReaderGroupHashMethod));
+	group->gr_Iter = iter;
 	strcpy(group->gr_GroupName, groupName);
     }
 
@@ -716,16 +1005,14 @@ EnterGroup(const char *groupName, int begNo, int endNo, int lmts, int cts, const
 }
 
 Group *
-FindGroupByHash(hash_t *hv)
+FindGroupByHash(char *Hash, int iter)
 {
     Group *group;
+    hash_t hv = hhash(Hash);
 
-    for (group = GHash[hv->h1 & GHMASK]; group; group = group->gr_Next) {
-	if (group->gr_Hv.h1 == hv->h1 &&
-	    group->gr_Hv.h2 == hv->h2
-	) {
+    for (group = GHash[hv.h1 & GHMASK]; group; group = group->gr_Next) {
+	if (strcmp(group->gr_Hash, Hash) == 0 && group->gr_Iter == iter)
 	    break;
-	}
     }
     return(group);
 }
@@ -746,7 +1033,7 @@ SetField(char **pptr, const char *str)
  */
 
 void
-ExpireByDays(Group *group, int fd, OverHead *oh)
+ExpireByDays(Group *group, int fd, OverHead *oh, int expireSecs)
 {
     const OverArt *oaBase;
     struct stat st;
@@ -756,20 +1043,11 @@ ExpireByDays(Group *group, int fd, OverHead *oh)
     int canceled = 0;
     int stale = 0;
     int n;
-    int expireSecs = -1;
     time_t t = time(NULL);
 
     if (fstat(fd, &st) != 0)
 	return;
 
-    /*
-     * Figure out expireSecs
-     */
-    {
-	if ((expireSecs = GetOverExpire(group->gr_GroupName)) == 0)
-	    expireSecs = -1;
-    }
-
     /*
      * Calculate number of overview records
      */
@@ -792,7 +1070,7 @@ ExpireByDays(Group *group, int fd, OverHead *oh)
 	for (i = 0; i < n; ++i) {
 	    const OverArt *oa = &oaBase[i];
 
-	    if (oa->oa_ArtNo > 0) {
+	    if (oa && oa->oa_ArtNo > 0) {
 		int dt = (int)(t - oa->oa_TimeRcvd);
 
 		if (VerboseOpt > 2)
@@ -816,9 +1094,9 @@ ExpireByDays(Group *group, int fd, OverHead *oh)
 
     {
 	/*
-	 * Update history file begin sequence number
+	 * Update active file begin sequence number
 	 */
-	while (group->gr_StartNo < group->gr_EndNo) {
+	while (group->gr_StartNo <= group->gr_EndNo) {
 	    int i = (group->gr_StartNo & 0x7FFFFFFF) % n;
 	    const OverArt *oa = &oaBase[i];
 
@@ -848,7 +1126,7 @@ ExpireByDays(Group *group, int fd, OverHead *oh)
 	    count,
 	    jumped,
 	    canceled, stale, expired,
-	    group->gr_EndNo - group->gr_StartNo,
+	    group->gr_EndNo - group->gr_StartNo + 1,
 	    ((expireSecs>0) ? (double)expireSecs / (60.0 * 60.0 * 24.0) :-1.0),
 	    group->gr_GroupName
 	);
@@ -899,7 +1177,7 @@ ExpireBySpool(Group *group, int fd, OverHead *oh)
 	for (i = 0; i < n; ++i) {
 	    const OverArt *oa = &oaBase[i];
 
-	    if (oa->oa_ArtNo > 0) {
+	    if (oa && oa->oa_ArtNo > 0) {
 		hash_t dh = oa->oa_MsgHash;
 		History dh_lookup;
 
@@ -912,8 +1190,7 @@ ExpireBySpool(Group *group, int fd, OverHead *oh)
 		 * its overview entry as well.
 		 */
 		if ((HistoryLookupByHash(dh, &dh_lookup) == -1) ||
-		    (H_EXP(dh_lookup.exp) == H_EXP((unsigned short)-1))
-		    ) {
+						H_EXPIRED(dh_lookup.exp)) {
 		    OverArt copy = *oa;
 
 		    copy.oa_ArtNo = -2;		/* EXPIRED */
@@ -931,7 +1208,7 @@ ExpireBySpool(Group *group, int fd, OverHead *oh)
 	/*
 	 * Update history file begin sequence number
 	 */
-	while (group->gr_StartNo < group->gr_EndNo) {
+	while (group->gr_StartNo <= group->gr_EndNo) {
 	    int i = (group->gr_StartNo & 0x7FFFFFFF) % n;
 	    const OverArt *oa = &oaBase[i];
 
@@ -961,7 +1238,7 @@ ExpireBySpool(Group *group, int fd, OverHead *oh)
 	    count,
 	    jumped,
 	    canceled, stale, expired,
-	    group->gr_EndNo - group->gr_StartNo,
+	    group->gr_EndNo - group->gr_StartNo + 1,
 	    group->gr_GroupName
 	);
     }
@@ -978,7 +1255,7 @@ ExpireBySpool(Group *group, int fd, OverHead *oh)
  */
 
 void
-ExpireFromFile(Group *group, int fd, OverHead *oh)
+ExpireFromFile(Group *group, int fd, OverHead *oh, int expireSecs)
 {
     const OverArt *oaBase;
     struct stat st;
@@ -988,20 +1265,11 @@ ExpireFromFile(Group *group, int fd, OverHead *oh)
     int canceled = 0;
     int stale = 0;
     int n;
-    int expireSecs = -1;
     time_t t = time(NULL);
 
     if (fstat(fd, &st) != 0)
 	return;
 
-    /*
-     * Figure out expireSecs
-     */
-    {
-	if ((expireSecs = GetOverExpire(group->gr_GroupName)) == 0)
-	    expireSecs = -1;
-    }
-
     /*
      * Calculate number of overview records
      */
@@ -1024,7 +1292,7 @@ ExpireFromFile(Group *group, int fd, OverHead *oh)
 	for (i = 0; i < n; ++i) {
 	    const OverArt *oa = &oaBase[i];
 
-	    if (oa->oa_ArtNo > 0) {
+	    if (oa && oa->oa_ArtNo > 0) {
 		int dt = (int)(t - oa->oa_TimeRcvd);
 		hash_t *hv = (hash_t *)(&(oa->oa_MsgHash));
 
@@ -1052,7 +1320,7 @@ ExpireFromFile(Group *group, int fd, OverHead *oh)
 	/*
 	 * Update history file begin sequence number
 	 */
-	while (group->gr_StartNo < group->gr_EndNo) {
+	while (group->gr_StartNo <= group->gr_EndNo) {
 	    int i = (group->gr_StartNo & 0x7FFFFFFF) % n;
 	    const OverArt *oa = &oaBase[i];
 
@@ -1082,7 +1350,7 @@ ExpireFromFile(Group *group, int fd, OverHead *oh)
 	    count,
 	    jumped,
 	    canceled, stale, expired,
-	    group->gr_EndNo - group->gr_StartNo,
+	    group->gr_EndNo - group->gr_StartNo + 1,
 	    ((expireSecs>0) ? (double)expireSecs / (60.0 * 60.0 * 24.0) :-1.0),
 	    group->gr_GroupName
 	);
@@ -1096,7 +1364,7 @@ ExpireFromFile(Group *group, int fd, OverHead *oh)
 
 
 void
-RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int h2)
+RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath)
 {
     const OverArt *oaBase;
     OverArt *obBase;
@@ -1119,8 +1387,17 @@ RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int
      * another expireover.
      */
 
-    if (hflock(fd, 4, XLOCK_EX|XLOCK_NB) < 0)
+    nrsignal(SIGALRM, sigAlarm);
+    alarm(LockWaitTime);
+    if (hflock(fd, 4, XLOCK_EX) < 0) {
+	nrsignal(SIGALRM, SIG_IGN);
+	if (VerboseOpt > 2)
+	    printf("unable to rewrite data for grp=%s - file in use\n",
+							group->gr_GroupName);
 	return;
+    }
+    nrsignal(SIGALRM, SIG_IGN);
+
 
     if (fstat(fd, &st) != 0 || st.st_nlink == 0) {
 	hflock(fd, 4, XLOCK_UN);
@@ -1164,7 +1441,7 @@ RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int
 		oa->oa_ArtNo < cacheArtBase ||
 		oa->oa_ArtNo >= cacheArtBase + OD_HARTS
 	    ) {
-		char path[1024];
+		char path[PATH_MAX];
 		struct stat st;
 		int cfd;
 
@@ -1174,7 +1451,13 @@ RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int
 		    cacheBase = NULL;
 		}
 		cacheArtBase = oa->oa_ArtNo & ~OD_HMASK;
-		snprintf(path, sizeof(path), "%s/data.%d.%08x.%08x", dirPath, cacheArtBase, h1, h2);
+		snprintf(path, sizeof(path), "%s/%s", dirPath,
+					GFName(group->gr_GroupName,
+					GRPFTYPE_DATA, cacheArtBase, 0,
+					group->gr_Iter,
+					&DOpts.ReaderGroupHashMethod));
+		if (DebugOpt > 1)
+		    printf("Opening: %s\n", path);
 		if ((cfd = open(path, O_RDONLY)) >= 0 &&
 		    fstat(cfd, &st) == 0
 		) {
@@ -1184,7 +1467,8 @@ RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int
 		if (cfd >= 0)
 		    close(cfd);
 	    }
-	    rewriteDataFile(group, &rfBase, cacheBase, cacheSize, oa, &obBase[i], dirPath, h1, h2);
+	    rewriteDataFile(group, &rfBase, cacheBase, cacheSize, oa,
+							&obBase[i], dirPath);
 	}
 	if (cacheBase != NULL) {
 	    xunmap((void *)cacheBase, cacheSize);
@@ -1202,9 +1486,16 @@ RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int
 	ReplaceFile *rf = rfBase;
 
 	while (rf) {
-	    char path1[1024];
-	    snprintf(path1, sizeof(path1), "%s/data.%d.%08x.%08x", dirPath, rf->rf_ArtBase, h1, h2);
-	    remove(path1);
+	    char path1[PATH_MAX];
+	    snprintf(path1, sizeof(path1), "%s/%s", dirPath,
+					GFName(group->gr_GroupName,
+					GRPFTYPE_DATA, rf->rf_ArtBase, 0,
+					group->gr_Iter,
+					&DOpts.ReaderGroupHashMethod));
+	    if (DebugOpt > 1)
+		printf("Remove: %s\n", path1);
+	    if (ForReal)
+		remove(path1);
 	    rf = rf->rf_Next;
 	}
     }
@@ -1215,13 +1506,30 @@ RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int
      */
 
     {
-	char path1[1024];
-	char path2[1024];
+	char path1[PATH_MAX];
+	char path2[PATH_MAX];
 	int ovFd;
 
-	snprintf(path1, sizeof(path1), "%s/.over.0.%08x.%08x", dirPath, h1, h2);
-	snprintf(path2, sizeof(path2), "%s/over.0.%08x.%08x", dirPath, h1, h2);
+	snprintf(path1, sizeof(path1), "%s/.%s", dirPath,
+					GFName(group->gr_GroupName,
+					GRPFTYPE_OVER, 0, 0,
+					group->gr_Iter,
+					&DOpts.ReaderGroupHashMethod));
+	snprintf(path2, sizeof(path2), "%s/%s", dirPath,
+					GFName(group->gr_GroupName,
+					GRPFTYPE_OVER, 0, 0,
+					group->gr_Iter,
+					&DOpts.ReaderGroupHashMethod));
 
+	/* Make sure we are using the latest versions */
+	if (oh->oh_Version != OH_VERSION) {
+	    printf("Upgrading overinfo header version %d->%d in %s\n",
+						oh->oh_Version, OH_VERSION,
+						group->gr_GroupName);
+	    oh->oh_Version = OH_VERSION;
+	    strncpy(oh->oh_Gname, group->gr_GroupName, sizeof(oh->oh_Gname));
+	    oh->oh_HeadSize = sizeof(OverHead);
+	}
 	ovFd = open(path1, O_RDWR|O_CREAT|O_TRUNC, 0644);
 	if (ovFd >= 0) {
 	    if (write(ovFd, oh, oh->oh_HeadSize) != oh->oh_HeadSize)
@@ -1230,39 +1538,61 @@ RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int
 		ok = 0;
 	    close(ovFd);
 	    if (ok) {
-		if (rename(path1, path2) < 0)
+		if (DebugOpt > 1)
+		    printf("Renaming over: %s -> %s\n", path1, path2);
+		if (ForReal && rename(path1, path2) < 0) {
+		    printf("Error renaming %s -> %s (%s)\n", path1, path2,
+							strerror(errno));
 		    ok = 0;
+		}
 	    }
 	} else {
 	    ok = 0;
 	}
 	if (ok == 0) {
-	    printf("Rewrite of %s over. file failed, removing\n", group->gr_GroupName);
-	    remove(path2);
+	    printf("Rewrite of %s over. file failed, removing\n",
+							group->gr_GroupName);
+	    if (ForReal)
+		remove(path1);
 	}
     }
 
     /*
      * Rename the temporary .data files to the real ones.
      */
-
+    i = 0;
     while (rfBase) {
-	char path1[1024];
-	char path2[1024];
+	char path1[PATH_MAX];
+	char path2[PATH_MAX];
 	ReplaceFile *rf = rfBase;
 
-	snprintf(path1, sizeof(path1), "%s/.data.%d.%08x.%08x", dirPath, rf->rf_ArtBase, h1, h2);
-	snprintf(path2, sizeof(path2), "%s/data.%d.%08x.%08x", dirPath, rf->rf_ArtBase, h1, h2);
-
-	if (rename(path1, path2) < 0)
-	    remove(path2);
+	snprintf(path1, sizeof(path1), "%s/.%s", dirPath,
+					GFName(group->gr_GroupName,
+					GRPFTYPE_DATA, rf->rf_ArtBase, 0,
+					group->gr_Iter,
+					&DOpts.ReaderGroupHashMethod));
+	snprintf(path2, sizeof(path2), "%s/%s", dirPath,
+					GFName(group->gr_GroupName,
+					GRPFTYPE_DATA, rf->rf_ArtBase, 0,
+					group->gr_Iter,
+					&DOpts.ReaderGroupHashMethod));
+	if (DebugOpt > 1)
+	    printf("Renaming data: %s -> %s\n", path1, path2);
+	if (ForReal && rename(path1, path2) < 0) {
+	    printf("Error renaming %s -> %s (%s)\n", path1, path2,
+							strerror(errno));
+	    remove(path1);
+	}
 
 	if (rf->rf_Fd >= 0)
 	    close(rf->rf_Fd);
 
 	rfBase = rf->rf_Next;
 	free(rf);
+	i++;
     }
+    if (VerboseOpt)
+	printf("Rebuilt %d data file(s) for %s\n", i, group->gr_GroupName);
 
     /*
      * cleanup
@@ -1270,6 +1600,7 @@ RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int
     hflock(fd, 4, XLOCK_UN);
     xunmap((void *)oaBase, n * sizeof(OverArt));
     free(obBase);
+    return;
 }
 
 /*
@@ -1278,11 +1609,14 @@ RewriteData(Group *group, int fd, OverHead *oh, const char *dirPath, int h1, int
  */
 
 void
-rewriteDataFile(Group *group, ReplaceFile **prf, const char *cacheBase, int cacheSize, const OverArt *oa, OverArt *ob, const char *dirPath, int h1, int h2)
+rewriteDataFile(Group *group, ReplaceFile **prf, const char *cacheBase, int cacheSize, const OverArt *oa, OverArt *ob, const char *dirPath)
 {
     int artBase = oa->oa_ArtNo & ~OD_HMASK;
     ReplaceFile *rf;
 
+    if (VerboseOpt > 1)
+	printf("Rewriting: %s:%d\n", group->gr_GroupName, artBase);
+
     /*
      * Locate the artBase file in our rewrite 'cache'.  Note: it's possible to
      * negatively cache an ArtBase, where rf_Fd will be < 0.
@@ -1294,11 +1628,19 @@ rewriteDataFile(Group *group, ReplaceFile **prf, const char *cacheBase, int cach
 	prf = &rf->rf_Next;
     }
     if (rf == NULL) {
-	char path[1024];
+	char path[PATH_MAX];
 	rf = calloc(1, sizeof(ReplaceFile));
 
-	snprintf(path, sizeof(path), "%s/.data.%d.%08x.%08x", dirPath, artBase, h1, h2);
+	snprintf(path, sizeof(path), "%s/.%s", dirPath,
+					GFName(group->gr_GroupName,
+					GRPFTYPE_DATA, artBase, 0,
+					group->gr_Iter,
+					&DOpts.ReaderGroupHashMethod));
+	if (DebugOpt > 1)
+            printf("Creating: %s\n", path);
 	rf->rf_Fd = open(path, O_RDWR|O_CREAT|O_TRUNC, 0644);
+	if (rf->rf_Fd == -1)
+	    printf("Error opening %s: %s\n", path, strerror(errno));
 	rf->rf_ArtBase = artBase;
 	*prf = rf;
     }
@@ -1356,19 +1698,24 @@ ResizeGroup(Group *group, int fd, OverHead *oh, int newSize)
     int oldSize = oh->oh_MaxArts;
     struct stat st;
 
-    if (hflock(fd, 4, XLOCK_EX|XLOCK_NB) < 0) {
+    nrsignal(SIGALRM, sigAlarm);
+    alarm(LockWaitTime);
+    if (hflock(fd, 4, XLOCK_EX) < 0) {
+	nrsignal(SIGALRM, SIG_IGN);
 	++NoResizedGroups;
-	if (VerboseOpt) {
-	    printf("resize maxArts from %d to %d failed, file in use grp=%s\n", oldSize, newSize, group->gr_GroupName);
-	}
+	if (VerboseOpt)
+	    printf("resize maxArts from %d to %d failed, file in use grp=%s\n",
+					oldSize, newSize, group->gr_GroupName);
 	return;
     }
+    nrsignal(SIGALRM, SIG_IGN);
 
     if (fstat(fd, &st) < 0 || st.st_nlink == 0) {
-	++NoResizedGroups;
 	hflock(fd, 4, XLOCK_UN);
+	++NoResizedGroups;
 	if (VerboseOpt) {
-	    printf("resize maxArts from %d to %d failed, file in use grp=%s\n", oldSize, newSize, group->gr_GroupName);
+	    printf("resize maxArts from %d to %d failed, file in use grp=%s\n",
+					oldSize, newSize, group->gr_GroupName);
 	}
 	return;
     }
@@ -1388,10 +1735,22 @@ ResizeGroup(Group *group, int fd, OverHead *oh, int newSize)
 	OverArt *ob = calloc(newSize, sizeof(OverArt));
 	int n;
 	int i;
+	int recLen;
+	const char *rec = KPDBReadRecord(KDBActive, group->gr_GroupName,
+							KP_LOCK, &recLen);
 
 	lseek(fd, oh->oh_HeadSize, 0);
 	n = read(fd, oa, oldSize * sizeof(OverArt)) / sizeof(OverArt);
 
+	/*
+	 * We need to get the NE field again now that we have locked the
+	 * over.* file because it could have been adjusted by dreaderd
+	 * since we originally got it.
+	 */
+	group->gr_EndNo = strtol(KPDBGetField(rec, recLen, "NE", NULL, "-1"),
+								NULL, 10);
+	KPDBUnlock(KDBActive, rec);
+
 	if (group->gr_EndNo - group->gr_StartNo >= newSize) {
 	    group->gr_StartNo = group->gr_EndNo - newSize + 1;
 	    group->gr_State |= GRF_EDITEDBEG | GRF_MODIFIED;
@@ -1414,6 +1773,15 @@ ResizeGroup(Group *group, int fd, OverHead *oh, int newSize)
 	 */
 
 	if (ForReal) {
+	    /* Take this opportunity to set the correct version */
+	    if (oh->oh_Version != OH_VERSION) {
+		printf("Upgrading overinfo header version %d->%d\n",
+						oh->oh_Version, OH_VERSION);
+		oh->oh_Version = OH_VERSION;
+		strncpy(oh->oh_Gname, group->gr_GroupName,
+							sizeof(oh->oh_Gname));
+		oh->oh_HeadSize = sizeof(OverHead);
+	    }
 	    lseek(fd, 0L, 0);
 	    ftruncate(fd, oh->oh_HeadSize + newSize * sizeof(OverArt));
 	    oh->oh_MaxArts = newSize;
@@ -1430,6 +1798,7 @@ ResizeGroup(Group *group, int fd, OverHead *oh, int newSize)
     if (VerboseOpt && oldSize != newSize) {
 	printf("resized maxArts from %d to %d grp=%s\n", oldSize, newSize, group->gr_GroupName);
     }
+    return;
 }
 
 int
