@@ -13,7 +13,7 @@ int
 main(int ac, char **av)
 {
     LoadDiabloConfig(ac, av);
-    DumpSpamFilterCache();
+    DumpSpamFilterCache(stdout);
     return(0);
 }
 
