@@ -10,8 +10,8 @@
 
 #include "defs.h"
 
-#define TYPE_NORMAL	1
-#define TYPE_XREPLIC	2
+#define TYPE_NORMAL     1
+#define TYPE_XREPLIC    2
 
 #define SF_NOSTREAM	0x0001
 #define SF_REALTIME	0x0002
@@ -19,9 +19,13 @@
 #define SF_HEADFEED	0x0008
 #define SF_NOCHECK	0x0010
 #define SF_GENLINES	0x0020
-#define SF_LOGARTS	0x0040
+#define SF_PRESBYTES	0x0040
+#define SF_ARTSTAT	0x0080
+#define SF_NOTIFY	0x0100
 
-int MaxRun = 2;		/* default MaxRun			       */
+#define MAXLOGARTS      80   /* maximal length of the logarts string */
+
+int MaxRun = 2;
 int MinFlushSecs = 0;	/* minimum time between flshs if queue not caught up */
 
 typedef struct DNode {
@@ -31,13 +35,16 @@ typedef struct DNode {
 
 DNode *Base;
 
-int spool_file(char *spoolFile, char *hostName, int maxq, int maxr, int qskip, int port, int delay, int type, int flags, char *obip, int txbufsiz, int rxbufsiz, int pass);
+void processDnewsfeeds(const char *qFile);
+void processDspoolout(const char *fname, char *qFile);
+int spool_file(char *spoolFile, int pass, NewslinkInfo *li, int flags);
 DNode *FindNode(char *name);
 void AddNode(char *name);
 
 int VerboseOpt;
 int ForReal = 1;
 int Quiet;
+int ForceRealTime = 0;
 int  TxBufSize = 0;
 int  RxBufSize = 0;
 char *OutboundIpStr = "-nop";
@@ -45,18 +52,26 @@ char *OutboundIpStr = "-nop";
 int
 main(int ac, char **av)
 {
-    char buf[256];
-    FILE *fi;
-    int pass;
-    int sfCount = 200;
-    int *sfAry = malloc(sfCount * sizeof(int));
     char *qFile = NULL;
-    const char *cFile;
+    const char *cFile = NULL;
+    int oldFormat = 0;
 
+    openlog("dspoolout", (DebugOpt > 0 ? LOG_PERROR: 0) | LOG_NDELAY | LOG_PID, LOG_NEWS);
     LoadDiabloConfig(ac, av);
 
-    cFile = DNNTPSpoolCtlPat;
+    cFile = DNewsfeedsPat;
 
+    /*
+     * Maintain compatability with old format
+     */
+    {
+	struct stat sb;
+	if (stat(PatLibExpand(DNNTPSpoolCtlPat), &sb) == 0) {
+	    cFile = DNNTPSpoolCtlPat;
+	    oldFormat = 1;
+	}
+    }
+    
     /*
      * Shift into the out.going directory
      */
@@ -74,45 +89,53 @@ main(int ac, char **av)
 	    if (*ptr == '-') {
 		ptr += 2;
 		switch(ptr[-1]) {
+		case 'B':
+		    if (*ptr == 0)
+			ptr = av[++i];
+		    OutboundIpStr = malloc(strlen(ptr) + 8);
+		    sprintf(OutboundIpStr, "-B%s", ptr);
+		    break;
 		case 'C':
 		    if (*ptr == 0)
 			++i;
 		    break;
-		case 'q':
-		    Quiet = 1;
-		    break;
-		case 'v':
-		    VerboseOpt = 1;
-		    break;
-		case 'n':
-		    ForReal = 0;
+		case 'd':
+		    DebugOpt = 1;
+		    if (*ptr)
+			DebugOpt = strtol(ptr, NULL, 0);
 		    break;
 		case 'f':
 		    cFile = (*ptr) ? ptr : av[++i];
 		    break;
-		case 'T':
-		    {
-			TxBufSize = strtol((*ptr ? ptr : av[++i]), NULL, 0);
-		    }
+		case 'm':
+		    MaxRun = strtol((*ptr ? ptr : av[++i]), NULL, 0);
+		    break;
+		case 'n':
+		    ForReal = 0;
+		    break;
+		case 'p':
+		    oldFormat = 1;
+		    break;
+		case 'q':
+		    Quiet = 1;
 		    break;
 		case 'R':
-		    {
-			RxBufSize = strtol((*ptr ? ptr : av[++i]), NULL, 0);
-		    }
+		    RxBufSize = strtol((*ptr ? ptr : av[++i]), NULL, 0);
 		    break;
-		case 'B':
-		    {
-			if (*ptr == 0)
-			    ptr = av[++i];
-			OutboundIpStr = malloc(strlen(ptr) + 8);
-			sprintf(OutboundIpStr, "-B%s", ptr);
-		    }
+		case 'r':
+		    ForceRealTime = 1;
 		    break;
 		case 's':
 		    MinFlushSecs = strtol((*ptr ? ptr : av[++i]), NULL, 0) * 60;
 		    break;
-		case 'm':
-		    MaxRun = strtol((*ptr ? ptr : av[++i]), NULL, 0);
+		case 'T':
+		    TxBufSize = strtol((*ptr ? ptr : av[++i]), NULL, 0);
+		    break;
+		case 'V':
+		    PrintVersion();
+		    break;
+		case 'v':
+		    VerboseOpt = 1;
 		    break;
 		default:
 		    break;
@@ -122,16 +145,39 @@ main(int ac, char **av)
 	    }
 	}
     }
+    if (DebugOpt && qFile != NULL)
+	printf("Processing label: %s\n", qFile);
 
-    cFile = PatLibExpand(cFile);
-
-    if (cFile[0] != '/') {
-	char *p = malloc(strlen(cFile) + strlen(NewsHome) + 3);
-	sprintf(p, "%s/%s", NewsHome, cFile);
-	cFile = p;
+    if (oldFormat || strstr(PatLibExpand(cFile), "dspoolout") != NULL) {
+	    processDspoolout(PatLibExpand(cFile), qFile);
     } else {
-	cFile = strdup(cFile);
+	    DNewsfeedsPat = cFile;
+	    processDnewsfeeds(qFile);
     }
+    return(0);
+}
+
+void
+processDnewsfeeds(const char *qFile)
+{
+    int sfCount = 200;
+    int *sfAry = malloc(sfCount * sizeof(int));
+    int pass;
+    int hour;
+
+    {
+	struct tm *tp;
+	time_t t;
+
+	t=time(0);
+	tp=localtime(&t);
+	hour=tp->tm_hour;
+    }
+
+    if (DebugOpt > 1)
+	printf("Using dnewsfeeds for feed config\n");
+
+    LoadNewsFeed(0, 1, NULL);
 
     /*
      * Get list of spoolfile names and NNTP hosts, one pair per line,
@@ -139,7 +185,123 @@ main(int ac, char **av)
      */
 
     for (pass = 1; pass <= 2; ++pass) {
-	if ((fi = fopen(cFile, "r")) != NULL) {
+	LabelList *ll;
+        NewslinkInfo *li;
+	int count = 0;
+
+	for (ll = FeedLinkLabelList(); ll != NULL && ll->label != NULL;
+							ll = ll->next) {
+	    int flags = 0;
+
+	    li = FeedLinkInfo(ll->label);
+	    if (li == NULL)
+		continue;
+	    if (qFile != NULL && strcmp(ll->label, qFile) != 0)
+		continue;
+
+	    if (li->li_Hours != NULL) {
+		int found = 0;
+		int hourbegin;
+		int hourend;
+		char *hourinterval;
+		char *minus;
+
+		if ((hourinterval=strtok(li->li_Hours,",")) != NULL) {
+		    while (hourinterval != NULL) {
+			hourbegin=strtol(hourinterval,NULL,0);
+			if ((minus=strstr(hourinterval,"-")) != NULL)
+			    hourend=strtol(minus+1,NULL,0);
+			else
+			    hourend=hourbegin;
+			if (hourbegin <= hourend) {
+			    if (hour >= hourbegin && hour <= hourend)
+				found=1;
+			} else {
+			    if (hour >= hourbegin || hour <= hourend)
+				found=1;
+			}
+			hourinterval=strtok(NULL,",");
+		    }
+		}
+		if (found == 0)
+		    continue;
+	    }
+	    flags = flags | (li->li_NoStream ? SF_NOSTREAM : 0);
+	    flags = flags | (li->li_RealTime ? SF_REALTIME : 0);
+	    flags = flags | (li->li_Notify ? SF_NOTIFY : 0);
+	    flags = flags | (li->li_HeadFeed ? SF_HEADFEED : 0);
+	    flags = flags | (li->li_NoCheck ? SF_NOCHECK : 0);
+	    flags = flags | (li->li_GenLines ? SF_GENLINES : 0);
+	    flags = flags | (li->li_PreserveBytes ? SF_PRESBYTES : 0);
+	    flags = flags | (li->li_ArticleStat ? SF_ARTSTAT : 0);
+
+	    if (flags & SF_NOTIFY && !(flags & SF_REALTIME))
+		flags &= ~SF_NOTIFY;
+
+	    /*
+	     * run spool file, with flags modifier from pass 1.  Pass1
+	     * may clear the SF_REALTIME flag.
+	     */
+	    if (count >= sfCount) {
+		sfCount += 200;
+		sfAry = realloc(sfAry, sfCount * sizeof(int));
+		memset(sfAry+count, 0, (sfCount - count) * sizeof(int));
+	    }
+	    if (pass == 2) {
+		flags = (flags & ~SF_REALTIME) | 
+			(sfAry[count] & SF_REALTIME);
+	    }
+
+	    flags = spool_file(ll->label, pass, li, flags);
+
+	    if (pass == 1)
+		sfAry[count] = flags;
+	    ++count;
+	}
+	if (pass == 1) {
+	    char sysline[256];
+
+	    snprintf(sysline, sizeof(sysline), "%s/dbin/dicmd flush", NewsHome);
+	    if (ForReal)
+		system(sysline);
+	}
+    }
+}
+
+void
+processDspoolout(const char *fname, char *qFile)
+{
+    int sfCount = 200;
+    int *sfAry = malloc(sfCount * sizeof(int));
+    int pass;
+    int hour;
+    FILE *fi;
+    char buf[256];
+
+    {
+	struct tm *tp;
+	time_t t;
+
+	t=time(0);
+	tp=localtime(&t);
+	hour=tp->tm_hour;
+    }
+
+    if (*fname != '/') {
+	char *p = malloc(strlen(fname) + strlen(NewsHome) + 3);
+	sprintf(p, "%s/%s", NewsHome, fname);
+	fname = p;
+    }
+
+    printf("Using legacy dnntpspool.ctl format from %s\n", fname);
+
+    /*
+     * Get list of spoolfile names and NNTP hosts, one pair per line,
+     * blank lines and lines beginning with '#' excepted.
+     */
+
+    for (pass = 1; pass <= 2; ++pass) {
+	if ((fi = fopen(fname, "r")) != NULL) {
 	    int count = 0;
 
 	    while (fgets(buf, sizeof(buf), fi) != NULL) {
@@ -159,9 +321,13 @@ main(int ac, char **av)
 		    int delay = 0;
 		    int flags = 0;
 		    int qskip = 0;
+		    int compress = 0;
 		    int txbufsiz = TxBufSize;
 		    int rxbufsiz = RxBufSize;
+		    int maxstream = 0;
 		    char *obip = NULL;
+		    char *hours = NULL;
+		    char *logarts = NULL;
 
 		    {
 			char *maxqStr;
@@ -175,6 +341,10 @@ main(int ac, char **av)
 				flags |= SF_NOSTREAM;
 			    } else if (strcmp(maxqStr, "headfeed") == 0) {
 				flags |= SF_HEADFEED;
+			    } else if (strcmp(maxqStr, "preservebytes") == 0) {
+				flags |= SF_PRESBYTES;
+			    } else if (strcmp(maxqStr, "ignorebytes") == 0) {
+				flags |= SF_PRESBYTES;
 			    } else if (strcmp(maxqStr, "nocheck") == 0) {
 				flags |= SF_NOCHECK;
 			    } else if (strcmp(maxqStr, "realtime") == 0) {
@@ -183,10 +353,19 @@ main(int ac, char **av)
 				flags |= SF_NOBATCH;
 			    } else if (strcmp(maxqStr, "genlines") == 0) {
 				flags |= SF_GENLINES;
+			    } else if (strcmp(maxqStr, "notify") == 0) {
+				flags |= SF_NOTIFY;
+			    } else if (strcmp(maxqStr, "compress") == 0) {
+				compress = 1;
 			    } else if (strcmp(maxqStr, "logarts") == 0) {
-				flags |= SF_LOGARTS;
+				logarts = "all";
+			    } else if (strncmp(maxqStr, "logarts=", 8) == 0) {
+			        /* we should check for validity of the parameter here */
+				logarts = maxqStr + 8;
 			    } else if (strncmp(maxqStr, "bind=", 5) == 0) {
 				obip = maxqStr + 5;
+			    } else if (strncmp(maxqStr, "hours=", 6) == 0) {
+				hours = maxqStr + 6;
 			    } else if (maxqStr[0] == 'q') {
 				qskip = strtol(maxqStr + 1, NULL, 0);
 			    } else if (maxqStr[0] == 'n') {
@@ -199,6 +378,8 @@ main(int ac, char **av)
 				txbufsiz = strtol(maxqStr + 1, NULL, 0);
 			    } else if (maxqStr[0] == 'R') {
 				rxbufsiz = strtol(maxqStr + 1, NULL, 0);
+			    } else if (maxqStr[0] == 'M') {
+				maxstream = strtol(maxqStr + 1, NULL, 0);
 			    } else {
 				if (pass == 1)
 				    fprintf(stderr, "bad keyword in control file: %s\n", maxqStr);
@@ -206,6 +387,34 @@ main(int ac, char **av)
 			}
 		    }
 
+		    if (hours != NULL) {
+			int found = 0;
+			int hourbegin;
+			int hourend;
+			char *hourinterval;
+			char *minus;
+
+			if ((hourinterval=strtok(hours,",")) != NULL) {
+			    while (hourinterval != NULL) {
+				hourbegin=strtol(hourinterval,NULL,0);
+				if ((minus=strstr(hourinterval,"-")) != NULL)
+				    hourend=strtol(minus+1,NULL,0);
+				else
+				    hourend=hourbegin;
+				if (hourbegin <= hourend) {
+				    if (hour >= hourbegin && hour <= hourend)
+					found=1;
+				} else {
+				    if (hour >= hourbegin || hour <= hourend)
+					found=1;
+				}
+				hourinterval=strtok(NULL,",");
+			    }
+			}
+			if (found == 0)
+			    continue;
+		    }
+			
 		    /*
 		     * run spool file, with flags modifier from pass 1.  Pass1
 		     * may clear the SF_REALTIME flag.
@@ -220,6 +429,25 @@ main(int ac, char **av)
 				(sfAry[count] & SF_REALTIME);
 		    }
 
+		    {
+			NewslinkInfo li;
+
+			li.li_HostName = hostName;
+			li.li_MaxQueueFile = maxq;
+			li.li_MaxParallel = maxr;
+			li.li_Port = port;
+			li.li_StartDelay = delay;
+			li.li_TransmitBuf = txbufsiz;
+			li.li_ReceiveBuf = rxbufsiz;
+			li.li_LogArts = logarts;
+			li.li_BindAddress = obip;
+			li.li_MaxStream = 0;
+			li.li_Compress = compress;
+			li.li_QueueSkip = qskip;
+
+			flags = spool_file(spoolFile, pass, &li, flags);
+
+/*
 		    flags = spool_file(
 			spoolFile, 
 			hostName,
@@ -233,8 +461,12 @@ main(int ac, char **av)
 			obip,
 			txbufsiz,
 			rxbufsiz,
-			pass
+			logarts,
+			pass,
+			maxstream
 		    );
+*/
+		    }
 
 		    if (pass == 1)
 			sfAry[count] = flags;
@@ -243,7 +475,7 @@ main(int ac, char **av)
 	    }
 	    fclose(fi);
 	} else if (pass == 1) {
-	    fprintf(stderr, "Unable to open %s\n", cFile);
+	    fprintf(stderr, "Unable to open %s\n", fname);
 	}
 	if (pass == 1) {
 	    char sysline[256];
@@ -252,7 +484,6 @@ main(int ac, char **av)
 	    system(sysline);
 	}
     }
-    return(0);
 }
 
 /*
@@ -262,7 +493,7 @@ main(int ac, char **av)
  */
 
 int
-spool_file(char *spoolFile, char *hostName, int maxq, int maxr, int qskip, int port, int delay, int type, int flags, char *obip, int txbufsiz, int rxbufsiz, int pass)
+spool_file(char *spoolFile, int pass, NewslinkInfo *li, int flags)
 {
     char seqFile[256];
     char newFile[256];
@@ -272,35 +503,57 @@ spool_file(char *spoolFile, char *hostName, int maxq, int maxr, int qskip, int p
     int endSeq = 0;
     long newTime = 0;
     char portBuf[32];
+    char logartsStr[MAXLOGARTS+2];
     int fd;
     struct stat st;
     char txBufSizeStr[32];
     char rxBufSizeStr[32];
+    char maxStreamStr[32];
+    char compressStr[32];
 
+    if (VerboseOpt)
+	printf("spool_file(%s, %d, %s, %d)\n",
+			spoolFile, pass, li->li_HostName, flags);
     /*
      * Initialize rx/txBufSizeStr
      */
 
     sprintf(txBufSizeStr, "-nop");
     sprintf(rxBufSizeStr, "-nop");
+    sprintf(maxStreamStr, "-nop");
+    sprintf(compressStr, "-nop");
 
-    if (txbufsiz > 0 && txbufsiz < 16 * 1024 * 1024)
-	sprintf(txBufSizeStr, "-T%d", txbufsiz);
-    if (rxbufsiz > 0 && rxbufsiz < 16 * 1024 * 1024)
-	sprintf(rxBufSizeStr, "-R%d", rxbufsiz);
+    if (li->li_TransmitBuf > 0 && li->li_TransmitBuf < 16 * 1024 * 1024)
+	sprintf(txBufSizeStr, "-T%d", li->li_TransmitBuf);
+    if (li->li_ReceiveBuf > 0 && li->li_ReceiveBuf < 16 * 1024 * 1024)
+	sprintf(rxBufSizeStr, "-R%d", li->li_ReceiveBuf);
+    if (li->li_MaxStream > 0)
+	sprintf(maxStreamStr, "-M%d", li->li_MaxStream);
+    if (li->li_Compress > 0)
+	sprintf(compressStr, "-Z%d", li->li_Compress);
+
+    /*
+     * Outbound article logging.
+     */
+    sprintf(logartsStr, "-nop");
+    if (li->li_LogArts && *li->li_LogArts) {
+	strcpy(logartsStr, "-A");
+	StrnCpyNull(logartsStr+2, li->li_LogArts, MAXLOGARTS);
+    }
 
     /*
      * Override outbound IP
      */
 
-    if (obip != NULL && strlen(obip) < sizeof(obIpBuf) - 8) {
-	sprintf(obIpBuf, "-B%s", obip);
+    if (li->li_BindAddress != NULL &&
+			strlen(li->li_BindAddress) < sizeof(obIpBuf) - 8) {
+	sprintf(obIpBuf, "-B%s", li->li_BindAddress);
 	outboundIpStr = obIpBuf;
     }
 
-    if (port < 0)
-	port = 119;
-    sprintf(portBuf, "%d", port);
+    if (li->li_Port <= 0)
+	li->li_Port = 119;
+    sprintf(portBuf, "%d", li->li_Port);
 
     sprintf(seqFile, ".%s.seq", spoolFile);
 
@@ -326,7 +579,7 @@ spool_file(char *spoolFile, char *hostName, int maxq, int maxr, int qskip, int p
 
     while (begSeq < endSeq) {
 	sprintf(newFile, "%s.S%05d", spoolFile, begSeq);
-	if (maxq > 0 && begSeq < endSeq - maxq)
+	if (li->li_MaxQueueFile > 0 && begSeq < endSeq - li->li_MaxQueueFile)
 	    remove(newFile);
 	if (stat(newFile, &st) == 0)
 	    break;
@@ -363,6 +616,10 @@ spool_file(char *spoolFile, char *hostName, int maxq, int maxr, int qskip, int p
 
 	    dt = time(NULL) - ((int32)newTime + MinFlushSecs);
 
+	    if (VerboseOpt)
+		printf("PASS %d: dt=%d tries=%d begSeq=%d endSeq=%d\n",
+			pass, dt, tries, begSeq, endSeq);
+
 	    if (endSeq <= begSeq + 4 || dt > 0 || dt < -MinFlushSecs) {
 		/*
 		 * If a dnewslink has a lock on the unsequenced spool file,
@@ -428,31 +685,38 @@ spool_file(char *spoolFile, char *hostName, int maxq, int maxr, int qskip, int p
      */
 
     if (pass == 2 && FindNode(spoolFile) == NULL) {
-	int tries = maxr;
+	int tries = li->li_MaxParallel;
 	int look;
 
 	/*
-	 * Start up to maxr dnewslinks.  If the nobatch
+	 * Start up to MaxParallel dnewslinks.  If the nobatch
 	 * flag is set, do not start any.
 	 */
 
 	if (flags & SF_NOBATCH)
 	    tries = 0;
 
-	for (look = 0; look < 10 && tries && begSeq + look < endSeq - qskip; ++look) {
+	for (look = 0; look < 10 && tries &&
+			begSeq + look < endSeq - li->li_QueueSkip; ++look) {
 	    int use = begSeq + look;
 	    int fd;
 	    char numBuf[32];
 
 	    {
-		int left = endSeq - use - qskip + 5;
+		int left = endSeq - use - li->li_QueueSkip;
 		if (left > 32)
 		    left = 32;
-		if (left < 3)
+		if (left < 3 && li->li_QueueSkip == 0)
 		    left = 3;
+		else if (left < 1)
+		    left = 1;
 		sprintf(numBuf, "%d", left);
+	    if (VerboseOpt)
+		printf("PASS %d: look=%d tries=%d begSeq=%d endSeq=%d use=%d left=%d\n",
+			pass, look, tries, begSeq, endSeq, use, left);
 	    }
 
+
 	    sprintf(newFile, "%s.S%05d", spoolFile, use);
 
 	    if ((fd = open(newFile, O_RDWR)) >= 0) {
@@ -465,49 +729,64 @@ spool_file(char *spoolFile, char *hostName, int maxq, int maxr, int qskip, int p
 		    if (!Quiet)
 			printf("dspoolout: run %s.S%05d\n", spoolFile, use);
 
-		    if (ForReal && fork() == 0) {
-			char binPath[128];
-
-			if (delay > 0)
-			    sleep(delay);
-
-			sprintf(templateFile, "%s.S%%05d", spoolFile);
-			sprintf(seqName, "%d", use);
-			snprintf(binPath, sizeof(binPath), "%s/dbin/dnewslink", NewsHome);
-
-			switch(type) {
-			case TYPE_NORMAL:
-			    execl(binPath, "dnewslink", 
-				"-h", hostName,
+		    sprintf(templateFile, "%s.S%%05d", spoolFile);
+		    sprintf(seqName, "%d", use);
+		    if (VerboseOpt)
+			    printf("%s %s%s %s%s %s%s %s%s %s%s %s %s %s %s %s %s %s %s %s %s %s %s\n",
+				"dnewslink",
+				"-h", li->li_HostName,
 				"-b", templateFile,
 				"-S", seqName,
 				"-N", numBuf,
 				"-P", portBuf,
 				"-D",
 				((flags & SF_NOSTREAM) ? "-i" : "-nop"),
-				((flags & SF_HEADFEED) ? "-H" : "-nop"),
+				((flags & SF_HEADFEED) ?
+				    ((flags & SF_PRESBYTES) ? "-HB" : "-H") :
+				    "-nop"),
 				((flags & SF_NOCHECK) ? "-I" : "-nop"),
 				((flags & SF_GENLINES) ? "-L" : "-nop"),
-				((flags & SF_LOGARTS) ? "-A" : "-nop"),
+				((flags & SF_ARTSTAT) ? "-x" : "-nop"),
+				compressStr,
+				logartsStr,
 				outboundIpStr,
 				txBufSizeStr,
 				rxBufSizeStr,
+				maxStreamStr
+			    );
+		    if (ForReal && fork() == 0) {
+			char binPath[128];
+
+			if (li->li_StartDelay > 0)
+			    sleep(li->li_StartDelay);
+
+			if (li->li_Priority > 0)
+			    nice(li->li_Priority);
+
+			snprintf(binPath, sizeof(binPath), "%s/dbin/dnewslink", NewsHome);
+
+			    execl(binPath, "dnewslink",
+				"-h", li->li_HostName,
+				"-b", templateFile,
+				"-S", seqName,
+				"-N", numBuf,
+				"-P", portBuf,
+				"-D",
+				((flags & SF_NOSTREAM) ? "-i" : "-nop"),
+				((flags & SF_HEADFEED) ?
+				    ((flags & SF_PRESBYTES) ? "-HB" : "-H") :
+				    "-nop"),
+				((flags & SF_NOCHECK) ? "-I" : "-nop"),
+				((flags & SF_GENLINES) ? "-L" : "-nop"),
+				((flags & SF_ARTSTAT) ? "-x" : "-nop"),
+				compressStr,
+				logartsStr,
+				outboundIpStr,
+				txBufSizeStr,
+				rxBufSizeStr,
+				maxStreamStr,
 				NULL
 			    );
-			    break;
-			case TYPE_XREPLIC:
-#ifdef NOTDEF
-		/* not supported with new format */
-			    execlp("innxmit", "innxmit",
-				"-a",
-				"-r",
-				"-S", hostName,
-				newFile,
-				NULL
-			    );
-#endif
-			    break;
-			}
 			exit(0);
 		    }
 		}
@@ -526,7 +805,8 @@ spool_file(char *spoolFile, char *hostName, int maxq, int maxr, int qskip, int p
 	if (VerboseOpt)
 	    printf("%s: flags %02x nseq %d\n", spoolFile, flags & SF_REALTIME, endSeq-begSeq);
 
-	if ((flags & SF_REALTIME) && endSeq - begSeq <= 2) {
+	if ((flags & SF_REALTIME) &&
+				(ForceRealTime || (endSeq - begSeq <= 2))) {
 	    int fd;
 
 	    if ((fd = open(spoolFile, O_RDWR|O_CREAT, 0600)) >= 0) {
@@ -536,41 +816,66 @@ spool_file(char *spoolFile, char *hostName, int maxq, int maxr, int qskip, int p
 		    if (!Quiet)
 			printf("dspoolout: run realtime %s\n", spoolFile);
 
+		    if (VerboseOpt)
+			printf("%s %s%s %s%s %s%s %s%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s\n",
+				"dnewslink",
+				"-h", li->li_HostName,
+				"-b", spoolFile,
+				"-P", portBuf,
+				"-N", "100",	/* 1000 spool cycles	*/
+				((li->li_StartDelay < 0) ? "-r-1" : "-r"),
+				((flags & SF_NOTIFY) ? "-o" : "-nop"),
+				((flags & SF_NOTIFY) ? spoolFile : "-nop"),
+				((flags & SF_NOSTREAM) ? "-i" : "-nop"),
+				((flags & SF_HEADFEED) ?
+				    ((flags & SF_PRESBYTES) ? "-HB" : "-H") :
+				    "-nop"),
+				((flags & SF_NOCHECK) ? "-I" : "-nop"),
+				((flags & SF_GENLINES) ? "-L" : "-nop"),
+				((flags & SF_ARTSTAT) ? "-x" : "-nop"),
+				compressStr,
+				logartsStr,
+				outboundIpStr,
+				txBufSizeStr,
+				rxBufSizeStr,
+				maxStreamStr
+			    );
 		    if (ForReal && fork() == 0) {
 			char binPath[128];
 
-			if (delay > 0)
-			    sleep(delay);
+			if (li->li_StartDelay > 0)
+			    sleep(li->li_StartDelay);
 
 			snprintf(binPath, sizeof(binPath), "%s/dbin/dnewslink", NewsHome);
 
-			switch(type) {
-			case TYPE_NORMAL:
 			    /*
 			     * realtime option is '-r'.  If a delay of -1
 			     * is specified, we pass -r-1 to dnewslink which
 			     * reduces the tail latency from 1 second to 1ms.
 			     */
 			    execl(binPath, "dnewslink", 
-				"-h", hostName,
+				"-h", li->li_HostName,
 				"-b", spoolFile,
 				"-P", portBuf,
 				"-N", "100",	/* 1000 spool cycles	*/
-				((delay < 0) ? "-r-1" : "-r"),
+				((li->li_StartDelay < 0) ? "-r-1" : "-r"),
+				((flags & SF_NOTIFY) ? "-o" : "-nop"),
+				((flags & SF_NOTIFY) ? spoolFile : "-nop"),
 				((flags & SF_NOSTREAM) ? "-i" : "-nop"),
-				((flags & SF_HEADFEED) ? "-H" : "-nop"),
+				((flags & SF_HEADFEED) ?
+				    ((flags & SF_PRESBYTES) ? "-HB" : "-H") :
+				    "-nop"),
 				((flags & SF_NOCHECK) ? "-I" : "-nop"),
 				((flags & SF_GENLINES) ? "-L" : "-nop"),
-				((flags & SF_LOGARTS) ? "-A" : "-nop"),
+				((flags & SF_ARTSTAT) ? "-x" : "-nop"),
+				compressStr,
+				logartsStr,
 				outboundIpStr,
 				txBufSizeStr,
 				rxBufSizeStr,
+				maxStreamStr,
 				NULL
 			    );
-			    break;
-			case TYPE_XREPLIC:
-			    break;
-			}
 			exit(0);
 		    }
 		}
