@@ -15,8 +15,11 @@ typedef struct Group {
     struct Group *gr_Next;
     int		gr_State;
     int		gr_StartNo;
+    int		gr_StartDiff;
     int		gr_EndNo;
+    int		gr_EndDiff;
     int		gr_SynNo;
+    int		gr_SynDiff;
     int		gr_NoCTS;
     int		gr_NoLMTS;
     char	*gr_GroupName;
@@ -44,23 +47,62 @@ FILE *Fo;
 Group *GHash[GHSIZE];
 KPDB  *KDBActive;
 
-int SyncOverwriteOpt = 0;	/* allow remote info to replace existing info	*/
+int SyncOverwriteOpt = 0; /* allow remote info to replace existing info	*/
 int SyncDeleteOpt = 0;	/* delete local info not present on remote	*/
 int SyncGroupsOpt = 0;	/* synchronize newsgroup names 			*/
 int SyncDescrOpt = 0;	/* synchronize newsgroup descriptions		*/
-int SyncModStatusOpt = 0;	/* synchronize newsgroup moderation status	*/
+int SyncModStatusOpt = 0; /* synchronize newsgroup moderation status	*/
 int SyncArtNumbersBegOpt = 0; /* synchronize article numbers	*/
 int SyncArtNumbersEndOpt = 0; /* synchronize article numbers	*/
 int SyncArtNumbersSynOpt = 0; /* synchronize article numbers	*/
 int SyncArtNumbersRange = 0;  /* relative sync beginning based on range	*/
+int SyncArtNumbersAddition = 0;  /* add this when syncing ending article no. */
+int SyncArtNumbersMaxPerc = 0;	/* max % value to increase upwards	*/
+int VerboseOpt = 0;
+int ForReal = 1;
+char *SyncModStatusFlag = NULL;
+char *BindHost = NULL;
+char *BindPort = NULL;
 
+void Usage(void);
+void processLine(char *cmd, char *wild, int *count, int *ok);
 char *allocTmpCopy(const char *buf, int bufLen);
-Group *EnterGroup(const char *, int, int, int, int, int, const char *, const char *);
+Group *EnterGroup(int pass, const char *, int, int, int, int, int, const char *, const char *);
 int ConnectToHost(const char *host, int port);
 int CommandResponse(const char *cmd, char **pres);
 int hash(const char *str);
 int SetField(char **pptr, const char *str);
 
+void
+Usage(void)
+{
+    fprintf(stderr, "dsyncgroups -h remotehost | -G filename [-p port] [-B hostname] [-w wild] [-f/F file.kp] [-a] [-o] [-g] [-i] [-m] [-M c] [-D] [-N[B,E,X,R]] [-X[EX]] [-n#] [-v] [-z]\n");
+    fprintf(stderr, "  -h remotehost = the host that we connect to\n");
+    fprintf(stderr, "  -G filename   = specify an active file to use as input\n");
+    fprintf(stderr, "  -p port       = port on remote host (default: 119)\n");
+    fprintf(stderr, "  -B bindhost   = interface to bind to\n");
+    fprintf(stderr, "  -w wild       = only update groups matching wildmat (default: *)\n");
+    fprintf(stderr, "  -f file.kp    = specify the active to update (Default: dactive.kp)\n");
+    fprintf(stderr, "  -F file.kp    = create active file if it doesn't exist\n");
+    fprintf(stderr, "  -a            = shortcut for -o, -g, -i and -m\n");
+    fprintf(stderr, "  -o            = overwrite description\n");
+    fprintf(stderr, "  -g            = add missing groups\n");
+    fprintf(stderr, "  -i            = sync newsgroup description\n");
+    fprintf(stderr, "  -m            = update moderation status\n");
+    fprintf(stderr, "  -M c          = update moderation status to the specified flag\n");
+    fprintf(stderr, "  -D            = delete local groups in not on remote\n");
+    fprintf(stderr, "  -N            = sync begin and end numbers\n");
+    fprintf(stderr, "  -NB           = sync begin numbers\n");
+    fprintf(stderr, "  -NE           = sync end numbers\n");
+    fprintf(stderr, "  -NX           = sync end numbers but only upwards\n");
+    fprintf(stderr, "  -XE           = sync NX field - feeder sync\n");
+    fprintf(stderr, "  -XX           = sync NX but only upwards\n");
+    fprintf(stderr, "  -n#           = if we sync end number then increase local value by remote+#\n");
+    fprintf(stderr, "  -N#           = don't sync end number  more than #\n");
+    fprintf(stderr, "  -v            = be more verbose about what is being done\n");
+    fprintf(stderr, "  -z            = don't actually make changes\n");
+    exit(1);
+}
 int
 main(int ac, char **av)
 {
@@ -69,32 +111,73 @@ main(int ac, char **av)
     char *host = NULL;
     char *wild = NULL;
     char *dbfile = NULL;
+    char *grpfile = NULL;
     int port = 119;
 
     LoadDiabloConfig(ac, av);
 
     for (i = 1; i < ac; ++i) {
+	char *p;
 	char *ptr = av[i];
+
 	if (*ptr != '-') {
 	    fprintf(stderr, "Unexpected argument: %s\n", ptr);
 	    exit(1);
 	}
 	ptr += 2;
 	switch(ptr[-1]) {
+	case 'a':
+	    SyncOverwriteOpt = 1;
+	    SyncGroupsOpt = 1;
+	    SyncDescrOpt = 1;
+	    SyncModStatusOpt = 1;
+	    break;
+	case 'B':
+	    if (*ptr == 0)
+		ptr = av[++i];
+	    if ((p = strrchr(ptr, ':')) != NULL) {
+		*p++ = 0;
+		BindPort = p;
+		BindHost = ptr;
+	    } else {  
+		BindHost = ptr;
+	    }
+	    break;
 	case 'C':
 	    if (*ptr == 0)
 		++i;
 	    break;
+	case 'D':
+	    SyncDeleteOpt = 1;
+	    break;
 	case 'd':
 	    DebugOpt = (*ptr) ? strtol(ptr, NULL, 0) : 1;
 	    break;
-	case 'p':
-	    ptr = (*ptr) ? ptr : av[++i];
-	    port = strtol(ptr, NULL, 0);
+	case 'F':
+	    ocreate = O_CREAT;
+	    /* fall through */
+	case 'f':
+	    dbfile = (*ptr) ? ptr : av[++i];
+	    break;
+	case 'G':
+	    grpfile = (*ptr) ? ptr : av[++i];
+	    break;
+	case 'g':
+	    SyncGroupsOpt = 1;
+	    break;
+	case 'i':
+	    SyncDescrOpt = 1;
 	    break;
 	case 'h':
 	    host = (*ptr) ? ptr : av[++i];
 	    break;
+	case 'm':
+	    SyncModStatusOpt = 1;
+	    break;
+	case 'M':
+	    SyncModStatusOpt = 1;
+	    SyncModStatusFlag = (*ptr) ? ptr : av[++i];
+	    break;
 	case 'N':
 	    if (*ptr == 0) {
 		SyncArtNumbersBegOpt = 1;
@@ -112,6 +195,22 @@ main(int ac, char **av)
 		++ptr;
 	    }
 	    break;
+	case 'n':
+	    SyncArtNumbersAddition = strtol((*ptr) ? ptr : av[++i], NULL, 0);
+	    break;
+	case 'o':
+	    SyncOverwriteOpt = 1;
+	    break;
+	case 'P':
+	    SyncArtNumbersMaxPerc = strtol((*ptr) ? ptr : av[++i], NULL, 0);
+	    break;
+	case 'p':
+	    ptr = (*ptr) ? ptr : av[++i];
+	    port = strtol(ptr, NULL, 0);
+	    break;
+	case 'w':
+	    wild = (*ptr) ? ptr : av[++i];
+	    break;
 	case 'X':
 	    if (*ptr == 0) {
 		SyncArtNumbersSynOpt = 1;
@@ -124,47 +223,24 @@ main(int ac, char **av)
 		++ptr;
 	    }
 	    break;
-	case 'F':
-	    ocreate = O_CREAT;
-	    /* fall through */
-	case 'f':
-	    dbfile = (*ptr) ? ptr : av[++i];
+	case 'V':
+	    PrintVersion();
 	    break;
-	case 'a':
-	    SyncOverwriteOpt = 1;
-	    SyncGroupsOpt = 1;
-	    SyncDescrOpt = 1;
-	    SyncModStatusOpt = 1;
+	case 'v':
+	    VerboseOpt = 1;
 	    break;
-	case 'o':
-	    SyncOverwriteOpt = 1;
-	    break;
-	case 'D':
-	    SyncDeleteOpt = 1;
-	    break;
-	case 'g':
-	    SyncGroupsOpt = 1;
-	    break;
-	case 'i':
-	    SyncDescrOpt = 1;
-	    break;
-	case 'm':
-	    SyncModStatusOpt = 1;
-	    break;
-	case 'w':
-	    wild = (*ptr) ? ptr : av[++i];
+	case 'z':
+	    ForReal = 0;
 	    break;
 	default:
 	    fprintf(stderr, "Unknown option: %s\n", ptr - 2);
 	    break;
 	}
     }
-    if (host == NULL) {
-	if (ac == 1) {
-	    puts("dsyncgroups -h remotehost [-p port] [-w wild] [-f/F file.kp] [-a] [-o] [-D] [-N[B,E,X,R]] [-g] [-i] [-m]");
-	    exit(0);
-	}
-	fprintf(stderr, "Missing host option\n");
+    if (host == NULL && grpfile == NULL) {
+	if (ac == 1)
+	    Usage();
+	fprintf(stderr, "Missing host/filename option\n");
 	exit(1);
     }
     if (port <= 0) {
@@ -180,6 +256,10 @@ main(int ac, char **av)
      * open dactive.kp file
      */
 
+    if (VerboseOpt)
+	printf("%s local active file\n",
+				ocreate ? "Loading/creating" : "Loading");
+
     if (dbfile) {
 	KDBActive = KPDBOpen(dbfile, O_RDWR|ocreate);
     } else {
@@ -235,6 +315,7 @@ main(int ac, char **av)
 		continue;
 
 	    grp = EnterGroup(
+		1,
 		group,
 		begNo,
 		endNo,
@@ -256,8 +337,14 @@ main(int ac, char **av)
      * Connect to remote host
      */
 
+    if (host)
     {
-	int fd = ConnectToHost(host, port);
+	int fd;
+
+	if (VerboseOpt)
+	    printf("Connecting to %s:%d\n", host, port);
+
+	fd = ConnectToHost(host, port);
 
 	if (fd < 0) {
 	    fprintf(
@@ -271,12 +358,22 @@ main(int ac, char **av)
 	}
 	Fi = fdopen(fd, "r");
 	Fo = fdopen(dup(fd), "w");
+    } else {
+	if (VerboseOpt)
+	    printf("Syncing with %s\n", grpfile);
+	Fi = fopen(grpfile, "r");
+	if (Fi == NULL) {
+	    printf("Unable to open %s\n", grpfile);
+	    exit(1);
+	}
+	Fo = NULL;
     }
 
     /*
      * startup
      */
 
+    if (host)
     {
 	char *line;
 
@@ -294,6 +391,7 @@ main(int ac, char **av)
      * mode reader
      */
 
+    if (host)
     {
 	char *line;
 
@@ -312,6 +410,7 @@ main(int ac, char **av)
      * list active [wild]
      */
 
+    if (host)
     {
 	char *line;
 	char cmd[4096];
@@ -319,50 +418,12 @@ main(int ac, char **av)
 	int ok = 0;
 
 	sprintf(cmd, "list active%s%s", ((wild) ? " " : ""), ((wild) ? wild : ""));
-
 	switch(CommandResponse(cmd, &line)) {
 	case 215:
 	    while (fgets(cmd, sizeof(cmd), Fi) != NULL) {
-		char *group;
-		char *gbeg;
-		char *gend;
-		char *flags;
-		Group *grp;
-
-		if (strcmp(cmd, ".\r\n") == 0) {
-		    ok = 1;
+		processLine(cmd, wild, &count, &ok);
+		if (ok)
 		    break;
-		}
-		group = strtok(cmd, " \r\n");
-		gend  = strtok(NULL, " \r\n");
-		gbeg  = strtok(NULL, " \r\n");
-		flags = strtok(NULL, " \r\n");
-
-		/*
-		 * ignore group that does not match the wildcard, even though
-		 * the remote end is supposed to do this for us.
-		 */
-
-		if (wild && WildCmp(wild, group) != 0)
-		    continue;
-
-		grp = EnterGroup(
-		    group, 
-		    strtol(gbeg, NULL, 10), 
-		    strtol(gend, NULL, 10), 
-		    strtol(gend, NULL, 10), /* end repeated */
-		    -1,
-		    -1,
-		    flags, 
-		    NULL
-		);
-
-		/*
-		 * mark group as coming from remote
-		 */
-		if (grp)
-		    grp->gr_State |= GRF_FROMREMOTE;
-		++count;
 	    }
 	    break;
 	default:
@@ -374,12 +435,19 @@ main(int ac, char **av)
 	    printf("Did not receive . terminator to list active command\n");
 	    exit(1);
 	}
+    } else {
+	char cmd[4096];
+	int count = 0;
+	while (fgets(cmd, sizeof(cmd), Fi) != NULL)
+	    processLine(cmd, wild, &count, NULL);
+	printf("%d groups found\n", count);
     }
 
     /*
      * list newsgroups [wild]
      */
 
+   if (host && SyncDescrOpt)
     {
 	char *line;
 	char cmd[4096];
@@ -387,7 +455,6 @@ main(int ac, char **av)
 	int ok = 0;
 
 	sprintf(cmd, "list newsgroups%s%s", ((wild) ? " " : ""), ((wild) ? wild : ""));
-
 	switch(CommandResponse(cmd, &line)) {
 	case 215:
 	    while (fgets(cmd, sizeof(cmd), Fi) != NULL) {
@@ -412,6 +479,7 @@ main(int ac, char **av)
 
 		if (group)
 		grp = EnterGroup(
+		    2,
 		    group, 
 		    -1,
 		    -1,
@@ -441,10 +509,11 @@ main(int ac, char **av)
      * close remote connections
      */
 
-    printf("All remote commands completed, sending quit\n");
+    if (host)
     {
 	char *line;
 
+	printf("All remote commands completed, sending quit\n");
 	switch(CommandResponse("quit", &line)) {
 	case 205:
 	    printf("quit response ok\n");
@@ -456,7 +525,8 @@ main(int ac, char **av)
     }
 
     fclose(Fi);
-    fclose(Fo);
+    if (Fo)
+	fclose(Fo);
 
     printf("Writing to .kp file\n");
     fflush(stdout);
@@ -491,42 +561,68 @@ main(int ac, char **av)
 		    SyncDeleteOpt
 		) {
 		    printf("Deleting %s\n", group->gr_GroupName);
-		    KPDBDelete(KDBActive, group->gr_GroupName);
-		} else if (group->gr_State & GRF_MODIFIED) {
+		    if (ForReal)
+			KPDBDelete(KDBActive, group->gr_GroupName);
+		} else if (group->gr_State & (GRF_MODIFIED | GRF_NEW)) {
 		    if (group->gr_State & GRF_NEW) {
-			printf("Creating %s %s\n", group->gr_GroupName, (group->gr_Flags ? group->gr_Flags : "y"));
+			printf("Creating %s", group->gr_GroupName);
 		    } else {
-			printf("Updating %s %s\n", group->gr_GroupName, (group->gr_Flags ? group->gr_Flags : "y"));
+			printf("Updating %s", group->gr_GroupName);
 		    }
-		    if (group->gr_State & GRF_DESCRIPTION)
-			KPDBWriteEncode(KDBActive, group->gr_GroupName, "GD", group->gr_Description, 0);
-		    if (group->gr_State & GRF_STARTNO) {
+		    if (group->gr_State & GRF_DESCRIPTION) {
+			if (ForReal)
+			    KPDBWriteEncode(KDBActive, group->gr_GroupName, "GD", group->gr_Description, 0);
+			if (VerboseOpt)
+			    printf(" desc='%s'", group->gr_Description);
+		    }
+		    if (group->gr_State & (GRF_STARTNO | GRF_NEW)) {
 			char startBuf[16];
 			sprintf(startBuf, "%010d", group->gr_StartNo);
-			KPDBWriteEncode(KDBActive, group->gr_GroupName, "NB", startBuf, 0);
+			if (ForReal)
+			    KPDBWriteEncode(KDBActive, group->gr_GroupName, "NB", startBuf, 0);
+			if (VerboseOpt)
+			    printf(" NB=%s(%+d)", startBuf, group->gr_StartDiff);
 		    }
-		    if (group->gr_State & GRF_ENDNO) {
+		    if (group->gr_State & (GRF_ENDNO | GRF_NEW)) {
 			char endBuf[16];
 			sprintf(endBuf, "%010d", group->gr_EndNo);
-			KPDBWriteEncode(KDBActive, group->gr_GroupName, "NE", endBuf, 0);
+			if (ForReal)
+			    KPDBWriteEncode(KDBActive, group->gr_GroupName, "NE", endBuf, 0);
+			if (VerboseOpt)
+			    printf(" NE=%s(%+d)", endBuf, group->gr_EndDiff);
 		    }
-		    if (group->gr_State & GRF_SYNNO) {
+		    if (group->gr_State & (GRF_SYNNO | GRF_NEW)) {
 			char synBuf[16];
 			sprintf(synBuf, "%010d", group->gr_SynNo);
-			KPDBWriteEncode(KDBActive, group->gr_GroupName, "NX", synBuf, 0);
+			if (ForReal)
+			    KPDBWriteEncode(KDBActive, group->gr_GroupName, "NX", synBuf, 0);
+			if (VerboseOpt)
+			    printf(" NX=%s(%+d)", synBuf, group->gr_SynDiff);
 		    }
 		    if (group->gr_State & GRF_CTS) {
 			char ctsBuf[16];
 			sprintf(ctsBuf, "%08x", nowt);
-			KPDBWriteEncode(KDBActive, group->gr_GroupName, "CTS", ctsBuf, 0);
+			if (ForReal)
+			    KPDBWriteEncode(KDBActive, group->gr_GroupName, "CTS", ctsBuf, 0);
+			if (VerboseOpt)
+			    printf(" CTS=%s", ctsBuf);
 		    }
 		    if (group->gr_State & GRF_LMTS) {
 			char lmtsBuf[16];
 			sprintf(lmtsBuf, "%08x", nowt);
-			KPDBWriteEncode(KDBActive, group->gr_GroupName, "LMTS", lmtsBuf, 0);
+			if (ForReal)
+			    KPDBWriteEncode(KDBActive, group->gr_GroupName, "LMTS", lmtsBuf, 0);
+			if (VerboseOpt)
+			    printf(" LMTS=%s", lmtsBuf);
 		    }
-		    if (group->gr_State & GRF_FLAGS)
-			KPDBWriteEncode(KDBActive, group->gr_GroupName, "S", group->gr_Flags, 0);
+		    if (group->gr_Flags && group->gr_State & GRF_FLAGS) {
+			if (ForReal)
+			    KPDBWriteEncode(KDBActive, group->gr_GroupName, "S", group->gr_Flags, 0);
+			if (VerboseOpt)
+			    printf(" S=%s",
+				(group->gr_Flags ? group->gr_Flags : "y"));
+		    }
+		    printf("\n");
 		}
 	    }
 	}
@@ -535,6 +631,58 @@ main(int ac, char **av)
     return(0);
 }
 
+void
+processLine(char *cmd, char *wild, int *count, int *ok)
+{
+    char *group;
+	char *gbeg;
+	char *gend;
+	char *flags;
+	Group *grp;
+
+	if (strcmp(cmd, ".\r\n") == 0) {
+	    if (ok)
+		*ok = 1;
+	    return;
+	}
+	group = strtok(cmd, " \r\n");
+	gend  = strtok(NULL, " \r\n");
+	gbeg  = strtok(NULL, " \r\n");
+	flags = strtok(NULL, " \r\n");
+
+	/*
+	 * ignore group that does not match the wildcard, even though
+	 * the remote end is supposed to do this for us.
+	 */
+
+	if (wild && WildCmp(wild, group) != 0)
+	    return;
+
+	if (*gend == '-' || *gbeg == '-') {
+	    printf("Invalid group begin or end number - skipping : %s\n", group);
+	    return;
+	}
+
+	grp = EnterGroup(
+	    2,
+	    group, 
+	    strtol(gbeg, NULL, 10), 
+	    strtol(gend, NULL, 10), 
+	    strtol(gend, NULL, 10), /* end repeated */
+	    -1,
+	    -1,
+	    flags, 
+	    NULL
+	);
+
+	/*
+	 * mark group as coming from remote
+	 */
+	if (grp)
+	    grp->gr_State |= GRF_FROMREMOTE;
+	++*count;
+}
+
 char *
 allocTmpCopy(const char *buf, int bufLen)
 {
@@ -553,7 +701,7 @@ allocTmpCopy(const char *buf, int bufLen)
 }
 
 Group *
-EnterGroup(const char *groupName, int begNo, int endNo, int synNo, int noCTS, int noLMTS, const char *flags, const char *desc)
+EnterGroup(int pass, const char *groupName, int begNo, int endNo, int synNo, int noCTS, int noLMTS, const char *flags, const char *desc)
 {
     Group **pgroup = &GHash[hash(groupName)];
     Group *group;
@@ -575,6 +723,7 @@ EnterGroup(const char *groupName, int begNo, int endNo, int synNo, int noCTS, in
 	group->gr_GroupName = (char *)(group + 1);
 	group->gr_NoCTS = 1;
 	group->gr_NoLMTS = 1;
+	group->gr_StartNo = 1;
 	strcpy(group->gr_GroupName, groupName);
     }
 
@@ -583,13 +732,26 @@ EnterGroup(const char *groupName, int begNo, int endNo, int synNo, int noCTS, in
      */
 
     if (synNo >= 0) {
-	if ((group->gr_State & GRF_SYNNO) == 0 ||
-	    SyncArtNumbersSynOpt
-	) {
+	if (pass == 1) {
+	    group->gr_SynNo = synNo;
+	} else if (SyncArtNumbersSynOpt) {
 	    group->gr_State |= GRF_SYNNO;
 	    if (group->gr_SynNo != synNo) {
 		if (SyncArtNumbersSynOpt != 2 || synNo > group->gr_SynNo) {
-		    group->gr_SynNo = synNo;
+		    int perc;
+		    if (group->gr_SynNo == 0)
+			group->gr_SynNo = 1;
+		    group->gr_SynDiff = (synNo + SyncArtNumbersAddition) -
+							group->gr_SynNo;
+		    perc = group->gr_SynDiff * 100 / group->gr_SynNo;
+		    if (SyncArtNumbersMaxPerc && group->gr_SynNo &&
+				(group->gr_SynDiff > 0) &&
+				(perc > SyncArtNumbersMaxPerc))
+			group->gr_SynDiff = SyncArtNumbersMaxPerc *
+							group->gr_SynNo / 100;
+		    if (group->gr_SynDiff == 0)
+			group->gr_SynDiff = 1;
+		    group->gr_SynNo += group->gr_SynDiff;
 		    group->gr_State |= GRF_MODIFIED;
 		}
 	    }
@@ -597,13 +759,26 @@ EnterGroup(const char *groupName, int begNo, int endNo, int synNo, int noCTS, in
     }
 
     if (endNo >= 0) {
-	if ((group->gr_State & GRF_ENDNO) == 0 ||
-	    SyncArtNumbersEndOpt
-	) {
+	if (pass == 1) {
+	    group->gr_EndNo = endNo;
+	} else if (SyncArtNumbersEndOpt) {
 	    group->gr_State |= GRF_ENDNO;
 	    if (group->gr_EndNo != endNo) {
 		if (SyncArtNumbersEndOpt != 2 || endNo > group->gr_EndNo) {
-		    group->gr_EndNo = endNo;
+		    int perc;
+		    if (group->gr_EndNo == 0)
+			group->gr_EndNo = 1;
+		    group->gr_EndDiff = (synNo + SyncArtNumbersAddition) -
+							group->gr_EndNo;
+		    perc = group->gr_EndDiff * 100 / group->gr_EndNo;
+		    if (SyncArtNumbersMaxPerc && group->gr_EndNo &&
+				(group->gr_EndDiff > 0) &&
+				(perc > SyncArtNumbersMaxPerc))
+			group->gr_EndDiff = SyncArtNumbersMaxPerc *
+							group->gr_EndNo / 100;
+		    if (group->gr_EndDiff == 0)
+			group->gr_EndDiff = 1;
+		    group->gr_EndNo += group->gr_EndDiff;
 		    group->gr_NoLMTS = 1;
 		    group->gr_State |= GRF_MODIFIED;
 		}
@@ -612,11 +787,11 @@ EnterGroup(const char *groupName, int begNo, int endNo, int synNo, int noCTS, in
     }
 
     if (begNo >= 0) {
-	if ((group->gr_State & GRF_STARTNO) == 0 ||
-	    SyncArtNumbersBegOpt ||
+	if (pass == 1) {
+	    group->gr_StartNo = begNo;
+	} else if (SyncArtNumbersBegOpt ||
 	    (SyncArtNumbersRange && (group->gr_State & GRF_ENDNO) && endNo > 0)
 	) {
-	    group->gr_State |= GRF_STARTNO;
 	    if (SyncArtNumbersRange) {
 		/*
 		 * Set beginning based on ending minus the difference
@@ -628,25 +803,31 @@ EnterGroup(const char *groupName, int begNo, int endNo, int synNo, int noCTS, in
 		if (begNo > group->gr_EndNo)
 		    begNo = group->gr_EndNo;
 	    }
-	    if (group->gr_StartNo != begNo)
+	    if (group->gr_StartNo != begNo) {
+		group->gr_State |= GRF_STARTNO;
 		group->gr_State |= GRF_MODIFIED;
+	    }
+	    group->gr_StartDiff = begNo - group->gr_StartNo;
 	    group->gr_StartNo = begNo;
 	}
     }
 
     if (flags) {
-	if ((group->gr_State & GRF_FLAGS) == 0 ||
-	    (SyncOverwriteOpt && SyncModStatusOpt)
-	) {
+	if (pass == 1) {
+	    SetField(&group->gr_Flags, flags);
+	} else if ((SyncOverwriteOpt && SyncModStatusOpt) ||
+					(group->gr_State & GRF_NEW)) {
 	    group->gr_State |= GRF_FLAGS;
-	    if (SetField(&group->gr_Flags, flags))
+	    if (SetField(&group->gr_Flags,
+				SyncModStatusFlag ? SyncModStatusFlag : flags))
 		group->gr_State |= GRF_MODIFIED;
 	}
     }
     if (desc) {
-	if ((group->gr_State & GRF_DESCRIPTION) == 0 ||
-	    (SyncDescrOpt && SyncOverwriteOpt)
-	) {
+	if (pass == 1) {
+	    SetField(&group->gr_Description, desc);
+	} else if ((SyncOverwriteOpt && SyncDescrOpt) ||
+			(SyncDescrOpt && (group->gr_State & GRF_NEW))) {
 	    group->gr_State |= GRF_DESCRIPTION;
 	    if (SetField(&group->gr_Description, desc))
 		group->gr_State |= GRF_MODIFIED;
@@ -655,12 +836,14 @@ EnterGroup(const char *groupName, int begNo, int endNo, int synNo, int noCTS, in
 
     if (noCTS >= 0 && group->gr_NoCTS != noCTS) {
 	group->gr_NoCTS = noCTS;
-	group->gr_State |= GRF_MODIFIED | GRF_CTS;
+	if (pass > 1)
+	    group->gr_State |= GRF_MODIFIED | GRF_CTS;
     }
 
     if (noLMTS >= 0 && group->gr_NoLMTS != noLMTS) {
 	group->gr_NoLMTS = noLMTS;
-	group->gr_State |= GRF_MODIFIED | GRF_LMTS;
+	if (pass > 1)
+	    group->gr_State |= GRF_MODIFIED | GRF_LMTS;
     }
 
     return(group);
@@ -676,7 +859,37 @@ ConnectToHost(const char *host, int port)
     bzero(&lsin, sizeof(lsin));
     bzero(&rsin, sizeof(rsin));
     lsin.sin_port = 0;
+    if (BindPort != NULL) {
+	struct servent *sen;
+	int port;
+	if ((port = strtol(BindPort, NULL, 0)) != 0) {
+	    lsin.sin_port = htons(port);
+	} else { 
+	    if ((sen = getservbyname(BindPort, "tcp")) != NULL) {
+		lsin.sin_port = sen->s_port;
+	    } else {
+		fprintf(stderr, "Unknown service: %s\n", BindPort);
+		exit(1);
+	    }
+	}
+    }
     lsin.sin_family = AF_INET;
+    if (BindHost == NULL) {
+	lsin.sin_addr.s_addr = INADDR_ANY;
+    } else {
+	if (strtol(BindHost, NULL, 0) > 0) {
+	    lsin.sin_addr.s_addr = inet_addr(BindHost);
+	} else {
+	    struct hostent *he;
+                
+	    if ((he = gethostbyname(BindHost)) != NULL) {
+		lsin.sin_addr = *(struct in_addr *)he->h_addr;
+	    } else { 
+		fprintf(stderr, "Unknown bind host: %s\n", BindHost);
+		exit(1);
+	    }
+	}
+    }
 
     errno = 0;
     if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
@@ -755,7 +968,7 @@ CommandResponse(const char *cmd, char **pres)
 int
 hash(const char *str)
 {
-    int hv = 0xA432BCDD;
+    unsigned int hv = 0xA432BCDD;
 
     while (*str) {
 	hv = (hv << 5) ^ *str ^ (hv >> 23);
